#! /usr/bin/env python

""" 
This script is responsible for making the necessary changes to Amber11
so that it will compile without issue after combining it with AmberTools 1.5
"""

import os, sys

from os import path

def printusage():
   """ Prints usage and description """
   print >> sys.stderr, "Usage: %s" % sys.argv[0]
   print >> sys.stderr, "\nThis script patches Amber11 to work with AmberTools 1.5. Note that"
   print >> sys.stderr, "the building process will no longer work with AmberTools 1.4, and must"
   print >> sys.stderr, "be unpacked and reinstalled from a fresh copy if you wish to revert back"
   print >> sys.stderr, "to AmberTools 1.4 + Amber11. %s takes no arguments" % sys.argv[0]
   sys.exit(1)

try:
   amberhome = os.environ["AMBERHOME"]
except KeyError:
   print >> sys.stderr, "Error: AMBERHOME must be set!"
   printusage()

if len(sys.argv) > 1:
   printusage()

# First thing we need to do is patch config.h
if not path.exists(path.join(amberhome, 'src', 'config.h')):
   print >> sys.stderr, "Error: You must run configure in %s/AmberTools/src first!" % amberhome
   sys.exit(1)

# Open config.h and load the lines, then close it
configh = open(os.path.join(amberhome, 'src', 'config.h'), 'r')
lines = configh.readlines()
configh.close()

# Determine whether we're doing a serial or parallel build
buildtype = "serial"
for line in lines:
   if "configure" in line and "-mpi" in line:
      buildtype = "parallel"

print "Patching %s/src/config.h" % amberhome

# Get the first line to see if it's already been modified
if lines[0].strip() != "#MODIFIED FOR AMBERTOOLS 1.5":
   os.system('rm %s/src/config.h' % amberhome)

   # Now open config.h in write mode to overwrite it
   configh = open(os.path.join(amberhome, 'src', 'config.h'), 'w')
   configh.write('#MODIFIED FOR AMBERTOOLS 1.5\n')
   for line in lines:
      if line[:6] == "CFLAGS":
         line = line.strip() + " $(AMBERCFLAGS)\n"
         configh.write(line)
         configh.write("OCFLAGS= $(COPTFLAGS) $(AMBERCFLAGS)\n")
      elif line[:6] == "LDFLAG":
         line = line.strip() + " $(AMBERLDFLAGS)\n"
         configh.write(line)
      elif line[:6] == "FFLAGS":
         line = line.strip() + " $(FNOOPTFLAGS)\n"
         configh.write(line)
      elif line[:9] == "FOPTFLAGS":
         line = line.strip() + " $(LOCALFLAGS) $(CUSTOMBUILDFLAGS)\n"
         configh.write(line)
      elif line[:4] == "FPP=":
         line = line.strip() + " $(FPPFLAGS) $(AMBERFPPFLAGS)\n"
         configh.write(line)
      elif line[:9] == "PMEMD_FPP":
         configh.write(line)
         configh.write("PMEMD_NETCDFLIB= $(NETCDFLIB)\n")
      else:
         configh.write(line)
   
   configh.close()
   print "%s/src/config.h patched!\n" % amberhome
else:
   print "%s/src/config.h already patched.\n" % amberhome

# Patch sander Makefile so it doesn't copy ncdump to bin.
# AT 1.5 NetCDF already dumps it there by default
print "Patching sander's Makefile"
makefile = open(path.join(amberhome, 'src', 'sander', 'Makefile'), 'r')
lines = makefile.readlines()
makefile.close()

makefile = open(path.join(amberhome, 'src', 'sander', 'Makefile'), 'w')
for line in lines:
   if line.strip() == 'cp ../../AmberTools/src/netcdf/bin/ncdump$(SFX) $(BINDIR)':
      continue
   elif line.strip() == 'cp ../../AmberTools/src/netcdf/include/*.mod .':
      makefile.write(line.replace('AmberTools/src/netcdf/', ''))
      continue
   makefile.write(line)
makefile.close()
print "Done patching sander's Makefile!\n"

# Now get rid of mmpbsa_py since it's now in AmberTools
print "Patching %s/src/mmpbsa_py" % amberhome
os.system('rm -fr %s/src/mmpbsa_py/*' % amberhome)
makefile = open(path.join(amberhome, 'src', 'mmpbsa_py', 'Makefile'), 'w')
makefile.write("""install:
\t@echo "MMPBSA.py has been moved to %s/AmberTools/src"
parallel: install
""" % amberhome)
makefile.close()
readme = open(path.join(amberhome, 'src', 'mmpbsa_py', 'README'), 'w')
readme.write("""MMPBSA.py has been upgraded and is released with AmberTools.
The source code is now found in $AMBERHOME/AmberTools/src/mmpbsa_py,
the test cases are found in $AMBERHOME/AmberTools/test/mmpbsa_py,
and the updated documentation describing all of the new functionality
can be found in the AmberTools 1.5 manual.
""")
print "Done patching %s/src/mmpbsa_py\n" % amberhome

# Get rid of mmpbsa_py tests
print "Patching %s/test/mmpbsa_py" % amberhome
os.system("rm -fr %s/test/mmpbsa_py/*" % amberhome)
runfile = open(path.join(amberhome, 'test', 'mmpbsa_py', 'Run'),'w')
runfile.write('#!/bin/sh\necho "MMPBSA.py tests have been moved to %s/AmberTools/test"' % amberhome)
runfile.close()
os.chmod(path.join(amberhome, 'test', 'mmpbsa_py', 'Run'), 0755)
print "Done patching %s/test/mmpbsa_py\n" % amberhome

# A lot of the Amber tests still need the exe link
print "Creating exe link for test cases..."
os.system("if [ ! -x %s/exe ]; then ln -s bin %s/exe; fi" % (amberhome, amberhome))
print "Finished creating link...\n"

# Modify pmemd's Makefile for netcdf changes
print "Patching pmemd's Makefile"
makefile = open(path.join(amberhome, 'src', 'pmemd', 'Makefile'), 'r')
lines = makefile.readlines()
makefile.close()

makefile = open(path.join(amberhome, 'src', 'pmemd', 'Makefile'), 'w')
for line in lines:
   if line.strip() == 'cp ../netcdf/include/*.mod ./src':
      makefile.write(line.replace('netcdf', '..'))
      continue
   makefile.write(line)
makefile.close()
      
# Now we are done
print "\nAmber 11 patched for AmberTools 1.5. Run\nmake %s\nto build amber11\n" % buildtype
