#!/bin/sh

# require AMBERHOME
if [ -z $AMBERHOME ]; then
   AMBERHOME=`dirname \`dirname $PWD\``
   export AMBERHOME=`dirname\`dirname $AMBERHOME\``
   echo "AMBERHOME is not set.  Assuming it is $AMBERHOME"
fi

# Find the proper python
if [ -x $AMBERHOME/bin/python ]; then
   echo " Using AmberTools' python"
   pythonexe=$AMBERHOME/bin/python
   pythonstring="\$AMBERHOME/bin/python"
   amberpy='yes'
else
   echo " Using system python"
   pythonexe=`which python`
   pythonstring=`which python`
   amberpy='no'
fi

# create the wrapper script
cat > cpinutil << EOF
#!/bin/sh
if [ -z \$AMBERHOME ]; then
   echo "Error: AMBERHOME must be set for MMPBSA!"
   exit 1
fi

if [ ! -f \$AMBERHOME/AmberTools/src/etc/cpinutil.py ]; then
   echo "Error: cpinutil.py cannot be found, which means it was removed."
   echo "       Unpack a fresh copy of AmberTools, or at least this file."
   exit 1
fi

$pythonstring \$AMBERHOME/AmberTools/src/etc/cpinutil.py \$*
EOF

/bin/chmod +x cpinutil

$pythonexe -c "import chemistry.amber.readparm, cpin_data, cpin_utilities" > /dev/null 2>&1

if [ $? -gt 0 ]; then
   echo "Error importing cpinutil.py python modules! cpinutil will not work."
   exit 1
fi

if [ "$amberpy" = 'yes' ]; then
   if [ "`uname -s`" = "Darwin" ]; then
      cp -r chemistry cpin_utilities.py cpin_data.py *.pyc $AMBERHOME/Python.framework/Versions/2.6/lib/python2.6/
   else
      cp -r chemistry cpin_utilities.py cpin_data.py *.pyc $AMBERHOME/lib/python2.6
   fi
else
   cp -r chemistry *.pyc ../../../bin
fi

# write the frcmod extractor
cat > extractFrcmod.py << EOF
#! $pythonexe
from chemistry.amber.readparm import amberParm
import sys

def usage():
   print 'Usage: extractFrcmod.py <prmtop> <frcmod>'
   sys.exit()

if len(sys.argv) != 3:
   usage()

if sys.argv[1] == '--help' or sys.argv[1] == '-h' or \
      sys.argv[1] == '-help':
   usage()

print 'Reading the prmtop...'

prmtop = amberParm(sys.argv[1])

if not prmtop.exists:
   print 'Error: %s does not exist!' % sys.argv[1]
   usage()

print 'Extracting your frcmod...'
prmtop.frcmod(sys.argv[2])

print 'Your frcmod, %s, is ready.' % sys.argv[2]
EOF

/bin/chmod +x extractFrcmod.py
