#include <object.hpp>
#include <capbox.hpp>
#include "addions.hpp"

namespace amber
{
    int get_numion( double dstchg, double ionchg )
    {
        if( dstchg * ionchg >= 0 )
        {
            throw std::runtime_error( "Error: molecule and ion are both positive (or negative), can not make neutral" );
        }
            
        return int( -dstchg/ionchg+0.5);
    }


    
    addions_command::addions_command( const char* cmd )
        :command_i( cmd )
    {
    }
    
    addions_command::addions_command( const string& dest, const string& ion1, int numion1 )
        :m_dest( dest ), m_ion1( ion1 ), m_numion1( numion1 )
    {
        m_runtyp = 1;
    }

    addions_command::addions_command( const string& dest, const string& ion1, int numion1, const string& ion2, int numion2 )
        :m_dest( dest ), m_ion1( ion1 ), m_numion1( numion1 ), m_ion2(ion2), m_numion2(numion2)
    {
        m_runtyp = 2;
    }
    
    addions_command::~addions_command()
    {
    }
    
    bool addions_command::exec()
    {
        double shell_extent = 4.0;
        double resolution   = 1.0;
        molecule_ptr pdst = content().get_mol( m_dest );

        if( m_runtyp==1 )
        {
            molecule_ptr pion = content().get_mol( m_ion1 );
            if( m_numion1==0 )
            {
                m_numion1 = get_numion( charge(*pdst), charge(*pion) );
            }

            assert( pdst != NULL && pion != NULL );
            addions( *pdst, *pion, m_numion1, shell_extent, resolution);
        }
        else if( m_runtyp==2 )
        {
            molecule_ptr pion1 = content().get_mol( m_ion1 );
            molecule_ptr pion2 = content().get_mol( m_ion2 );
            if( m_numion1==0 && m_numion2==0 )
            {
                throw std::runtime_error( "Error: while addions2, numion1 and numion2 cannot both be zero" );
            }

            if( m_numion1==0 )
            {
                m_numion1 = get_numion( charge(*pdst)+m_numion2*charge(*pion2), charge(*pion1) );
            }
            else if( m_numion2==0 )
            {
                m_numion2 = get_numion( charge(*pdst)+m_numion1*charge(*pion1), charge(*pion2) );
            }

            addions( *pdst, *pion1, m_numion1, *pion2, m_numion2, shell_extent, resolution );
        }

	return true;
    }
    
    void addions_command::undo()
    {
        throw std::runtime_error( "Sorry: not implemented yet" );
    }

    const char* addions_command::info( ) const
    {
        return "  usage: addions mol ion1 numIon1 ";
    }
    
    shared_ptr< command_i > addions_command::clone( const vector< string >& args ) const
    {
        if( args.size()==3 || args.size()==4 )
        {
            int numb = args.size()==3 ? 0 : atoi(args[3].c_str());
            return shared_ptr< command_i >( new addions_command(args[1], args[2], numb) );
        }

        if( args.size()==6 )
        {
            int nion1 = atoi( args[3].c_str() );
            int nion2 = atoi( args[5].c_str() );
            return shared_ptr< command_i >( new addions_command(args[1], args[2], nion1, args[4], nion2 ) );
        }

        throw std::runtime_error( "Error: Wrong number of arguments!" );
    }

    addions_command g_addions_command( "addions" );
    addions_command g_addions2_command( "addions2" );
    
} // namespace amber



