#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "Constants.h"
#include "Macros.h"
#include "SpecialMath.h"
#include "mdgxVector.h"

#include "CellManipDS.h"
#include "pmeDirectDS.h"
#include "CompFrcDS.h"
#include "TopologyDS.h"
#include "TrajectoryDS.h"

/***=======================================================================***/
/*** TestBondedExclusion: this function tests whether two atoms share a    ***/
/***                      bonded exclusion by checking the topology.  This ***/
/***                      is a fairly time-consuming process, but it only  ***/
/***                      happens if the distance between two atoms is     ***/
/***                      extremely small.                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   a1:  the first atom                                                 ***/
/***   a2:  the second atom                                                ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
static int TestBondedExclusion(int a1, int a2, prmtop *tp)
{
  int i, llim, hlim, isexcl;
  int* exatm;

  /*** Make sure that a1 < a2 for reference in the exclusion list ***/
  if (a1 > a2) {
    SWAP(a1, a2, i);
  }

  /*** Determine limits and loop ***/
  llim = tp->ConExcl[a1];
  hlim = tp->ConExcl[a1+1];
  exatm = &tp->ExclList[llim];
  isexcl = 0;
  for (i = 0; i < hlim-llim; i++) {
    if (exatm[i] == a2) {
      isexcl = 1;
      break;
    }
  }
  
  return isexcl;
}

/*** BLOCKS 1 and 2: the forces are required ***/
#define NEEDFORCE 1

/*** BLOCK 1: energy (and virial) are not needed ***/
#define NEEDENERGY 0
#define NEEDVIRIAL 0
#define VGTE 1
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 1
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE
#undef NEEDVIRIAL
#undef NEEDENERGY
/*** END BLOCK 1 ***/

/*** BLOCK 2a: energy (but not virial) is needed, ***/
/***           in addition to forces              ***/
#define NEEDENERGY 1
#define NEEDVIRIAL 0
#define VGTE 1
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 1
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE
#undef NEEDVIRIAL
/*** END BLOCK 2a ***/

/*** BLOCK 2b: energy, virial, and forces are all needed ***/
#define NEEDVIRIAL 1
#define VGTE 1
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 1
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE
#undef NEEDVIRIAL
#undef NEEDENERGY
/*** END BLOCK 2b ***/

#undef NEEDFORCE
/*** END BLOCKS 1 and 2 ***/

/*** BLOCK 3: energy alone, but not forces or virials, are needed ***/
#define NEEDFORCE 0
#define NEEDENERGY 1
#define NEEDVIRIAL 0
#define VGTE 1
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 1
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 1
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE

#define VGTE 0
#define CHKRANGE 0
#include "NBLoop.c"
#undef VGTE
#undef CHKRANGE
#undef NEEDVIRIAL
#undef NEEDENERGY
#undef NEEDFORCE
/*** END BLOCK 3 ***/
