#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Matrix.h"
#include "Constants.h"

/***=======================================================================***/
/*** CountWords: this function counts the number of words appearing on a   ***/
/***             line.  To store all words, sequentially and individually, ***/
/***             use the ParseWords function in this same library.         ***/
/***=======================================================================***/
int CountWords(char* line)
{
  int i, nword, slen, onword, inquotes;

  nword = 0;
  slen = strlen(line);
  onword = 0;
  inquotes = 0;
  for (i = 0; i < slen; i++) {
    if (line[i] == 34 || line[i] == 39) {
      inquotes = 1 - inquotes;
    }
    if (line[i] != ' ' && onword == 0 && line[i] != '\n') {
      onword = 1;
      nword++;
    }
    else if (inquotes == 0 && line[i] == ' ' && onword == 1) {
      onword = 0;
    }
  }

  return nword;
}

/***=======================================================================***/
/*** ParseWords: this function stores all words appearing on a line.       ***/
/***=======================================================================***/
cmat ParseWords(char* line)
{
  int i, j, onword, inquotes, nc;
  cmat CW;

  CW = CreateCmat(CountWords(line), MAXNAME);
  i = 0;
  j = 0;
  while (i < CW.row) {

    /*** Advance to the next word ***/
    while (line[j] == ' ') {
      j++;
    }

    /*** Read this word ***/
    onword = 1;
    if (line[j] == 34 || line[j] == 39) {
      inquotes = 1;
      j++;
    }
    else {
      inquotes = 0;
    }
    nc = 0;
    while (onword == 1) {

      if (inquotes == 0 && (line[j] == ' ' || line[j] == '\n')) {
	onword = 0;
	CW.map[i][nc] = '\0';
      }
      else if (inquotes == 1 && (line[j] == 34 || line[j] == 39)) {
	onword = 0;
	inquotes = 0;
	CW.map[i][nc] = '\0';
      }
      else {
	CW.map[i][nc] = line[j];
	nc++;
      }
      j++;
    }
    i++;
  }

  return CW;
}

/***=======================================================================***/
/*** RemoveWhiteSpace: function for removing white space at the beginning  ***/
/***                   of a string.                                        ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   a:     the string                                                   ***/
/***   asize: the length of the string                                     ***/
/***=======================================================================***/
void RemoveWhiteSpace(char* a, int asize)
{
  int i, j, num_rem;

  num_rem = 0;
  while (a[num_rem] == ' ' && num_rem < asize) {
    num_rem++;
  }
  j = 0;
  for (i = num_rem; i < asize; i++) {
    a[j] = a[i];
    j++;
  }
  for (i = asize-num_rem; i < asize; i++) {
    a[i] = ' ';
  }
}

/***=======================================================================***/
/*** EqualSpace: add white space before and after any equal symbols in a   ***/
/***             line UNLESS they occur within ('') or ("") quotes.        ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   line:    the line (from an input command file)                      ***/
/***=======================================================================***/
void EqualSpace(char* line)
{
  int i, j, slen, inquotes;
  slen = strlen(line);

  inquotes = 0;
  for (i = 0; i < slen; i++) {
    if (line[i] == '=' && inquotes == 0) {
      line[slen+2] = '\0';
      for (j = slen-1; j > i; j--) {
        line[j+2] = line[j];
      }
      line[i] = ' ';
      line[i+1] = '=';
      line[i+2] = ' ';
      slen += 2;
      i += 2;
    }

    /*** ASCII standard character set used to identify quotation symbols ***/
    if (line[i] == 34 || line[i] == 39) {
      inquotes = 1 - inquotes;
    }
  }
}

/***=======================================================================***/
/*** RemoveComments: this function removes commented text from a line.     ***/
/***                 Comments are denoted by the appearance of '%', '#',   ***/
/***                 or '$' symbols.                                       ***/
/***=======================================================================***/
void RemoveComments(char* line)
{
  int i;
  const int slen = strlen(line);

  for (i = 0; i < slen; i++) {
    if (line[i] == '$' || line[i] == '%' || line[i] == '#') {
      line[i] = '\0';
      break;
    }
  }
}

/***=======================================================================***/
/*** NixCommaCarriage: replaces all commas and carriage returns on a line  ***/
/***                   with white space.                                   ***/
/***=======================================================================***/
void NixCommaCarriage(char* line)
{
  int i, slen, inquotes;

  slen = strlen(line);
  inquotes = 0;
  for (i = 0; i < slen; i++) {
    if (inquotes == 0 && (line[i] == ',' || line[i] == '\n')) {
      line[i] = ' ';
    }

    /*** ASCII standard character set used to identify quotation symbols ***/
    if (line[i] == 34 || line[i] == 39) {
      inquotes = 1 - inquotes;
    }
  }
}

/***=======================================================================***/
/*** AdvanceToSegment: this function searches a command file from the      ***/
/***                   beginning and stops after finding the heading for a ***/
/***                   section of interest.  It is very similar to the     ***/
/***                   function ReadToFlag in the Topology library, but    ***/
/***                   there are enough differences between topology and   ***/
/***                   command files to warrant a new function.            ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   inp:     the input file                                             ***/
/***   segname: the name of the target segment                             ***/
/***=======================================================================***/
int AdvanceToSegment(FILE *inp, char* segname)
{
  int slen, collect;
  char line[MAXLINE];

  /*** Start by rewinding the file ***/
  slen = strlen(segname);
  rewind(inp);
  collect = 0;
  while (collect == 0) {
    if (fgets(line, MAXLINE, inp) == NULL) {
      break;
    }
    RemoveWhiteSpace(line, MAXLINE);
    if (line[0] == '&' && strncmp(&line[1], segname, slen) == 0) {
      collect = 1;
    }
  }

  return collect;
}

/***=======================================================================***/
/*** DetectNamelistEnd: detect the end of a namelist, signified by a line  ***/
/***                    containing "&end."                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   line:    the line (from an input command file)                      ***/
/***   errmsg:  the error message to display (the name of the function in  ***/
/***              which the call to DetectNamelistEnd originated)          ***/
/***=======================================================================***/
int DetectNamelistEnd(char* line, char* errmsg)
{
  if (line[0] == '&') {
    if (strncmp(&line[1], "end", 3) == 0) {
      return 0;
    }
    else {
      printf("%s >> Error.  New segment encountered before termination of "
             "%s >> current segment.\n", errmsg, errmsg);
      exit(1);
    }
  }

  return 1;
}

/***=======================================================================***/
/*** SeekString: find a tag for a particular string.                       ***/
/***=======================================================================***/
void SeekString(cmat L, char* val, char* sname, char* salias)
{
  int i;

  for (i = 0; i < L.row; i++) {
    if (strcmp(L.map[i], sname) == 0 || strcmp(L.map[i], salias) == 0) {
      if (i+1 < L.row && strcmp(L.map[i+1], "=") != 0) {
        strcpy(val, L.map[i+1]);
      }
      else if (i+2 < L.row) {
        strcpy(val, L.map[i+2]);
      }
      else {
        printf("SeekString >> Error.  No value specified for identifier %s/%s"
               ".\n", sname, salias);
      }
    }
  }
}

/***=======================================================================***/
/*** SeekReal: find a tag for a particular real value.                     ***/
/***=======================================================================***/
void SeekReal(cmat L, double *val, char* sname, char* salias)
{
  int i;

  for (i = 0; i < L.row; i++) {
    if (strcmp(L.map[i], sname) == 0 || strcmp(L.map[i], salias) == 0) {
      if (i+1 < L.row && strcmp(L.map[i+1], "=") != 0) {
        *val = atof(L.map[i+1]);
      }
      else if (i+2 < L.row) {
        *val = atof(L.map[i+2]);
      }
      else {
        printf("SeekReal >> Error.  No value specified for identifier %s/%s"
               ".\n", sname, salias);
      }
    }
  }
}

/***=======================================================================***/
/*** SeekInt: find a tag for a particular integer value.                   ***/
/***=======================================================================***/
void SeekInt(cmat L, int *val, char* sname, char* salias)
{
  int i;

  for (i = 0; i < L.row; i++) {
    if (strcmp(L.map[i], sname) == 0 || strcmp(L.map[i], salias) == 0) {
      if (i+1 < L.row && strcmp(L.map[i+1], "=") != 0) {
        *val = atoi(L.map[i+1]);
      }
      else if (i+2 < L.row) {
        *val = atoi(L.map[i+2]);
      }
      else {
        printf("SeekInt >> Error.  No value specified for identifier %s/%s"
               ".\n", sname, salias);
      }
    }
  }
}

/***=======================================================================***/
/*** SeekLLInt: find a tag for a particular long long integer value.       ***/
/***=======================================================================***/
void SeekLLInt(cmat L, long long int *val, char* sname, char* salias)
{
  int i;

  for (i = 0; i < L.row; i++) {
    if (strcmp(L.map[i], sname) == 0 || strcmp(L.map[i], salias) == 0) {
      if (i+1 < L.row && strcmp(L.map[i+1], "=") != 0) {
        sscanf(L.map[i+1], "%lld", val);
      }
      else if (i+2 < L.row) {
        sscanf(L.map[i+2], "%lld", val);
      }
      else {
        printf("SeekLLInt >> Error.  No value specified for identifier %s/%s"
               ".\n", sname, salias);
      }
    }
  }
}
