#define _REAL_ double precision
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     AMG1R5                                        MAIN SUBROUTINE

!     RELEASE 1.5, OCTOBER 1990

!     CHANGES AGAINST VERSION 1.1, JULY 1985:

! 1.  A BUG WAS DETECTED WHICH UNDER CERTAIN CIRCUMSTANCES INFLUENCED
!     SLIGHTLY THE CONVERGENCE RATE OF AMG1R1. FOR THAT REASON, THE
!     FOLLOWING LINE IN SUBROUTINE RESC:
!     IW(IMAXW(KC-1)+1) = IA(IMIN(KC))
!     HAS BEEN CHANGED TO:
!     IW(IMAXW(KC-1)+1) = IAUX

! 2.  A BUG WAS DETECTED IN SUBROUTINE PWINT. UNDER CERTAIN CIRCUM-
!     STANCES AN UNDEFINED VARIABLE WAS USED. ALTHOUGH THIS DID NOT
!     AFFECT THE NUMERICAL RESULTS, PROBLEMS CAN OCCUR IF CHECKING
!     FOR UNDEFINED VARIABLES IS USED. TO FIX THIS ERROR, IN PWINT
!     THE LABEL 1000 WAS MOVED TO THE STATEMENT
!     IBLCK1 = IMINW(K).

! 3.  A PARAMETER LRATIO HAS BEEN INTRODUCED, DENOTING THE RATIO
!     OF SPACE OCCUPIED BY A REAL REAL VARIABLE AND
!     THAT OF AN INTEGER. FOR THE IBM-VERSION LRATIO HAS BEEN SET
!     TO 2. CHANGE THIS VALUE IF NECESSARY. (IF, FOR EXAMPLE, YOU
!     WANT TO CHANGE THE REAL VECTORS TO SINGLE PRE-
!     CISION, LRATIO HAS TO BE SET TO 1. IN THE YALE SMP - ROUTINE
!     NDRV THERE IS A PARAMETER LRATIO, TOO.

! 4.  TYPE DECLARATIONS REAL*4 AND REAL*8 HAVE BEEN CHANGED TO THE
!     STANDARD-CONFORMING KEYWORDS REAL AND REAL, RESPEC-
!     TIVELY.

! 5.  CALLS TO THE FOLLOWING INTRINSIC FUNCTIONS HAVE BEEN REPLACED BY
!     CALLS USING GENERIC NAMES: DSQRT, MIN0, MAX0, IABS, DABS, FLOAT,
!     DFLOAT, DMAX1, ISIGN, IDINT, DLOG10.

! 6.  A SAVE STATEMENT HAS BEEN INSERTED IN ALL SUBROUTINES.

! 7.  EXTERNAL DECLARATION STATEMENTS HAVE BEEN INSERTED IN ALL SUB-
!     ROUTINES FOR ALL EXTERNAL REFERENCES.

!-----------------------------------------------------------------------

!     CHANGE AGAINST VERSION 1.3, APRIL 1986:

! 1.  A BUG IN SUBROUTINE CHECK HAS BEEN REMOVED. IF THE ORIGINAL MATRIX
!     WAS STORED IN AN UNSYMMETRIC WAY, THE SYMMETRIZATION BY AMG1R3
!     COULD FAIL UNDER CERTAIN CIRCUMSTANCES. FOR A FIX, THE FOLLOWING
!     STATEMENTS IN SUBROUTINE CHECK HAVE BEEN CHANGED:

!     DO 450 J=IA(I)+1,IA(I+1)-1 WAS CHANGED TO
!     DO 450 J=IA(I)+1,ICG(I)-1

!     DO 430 J1=IA(I1)+1,IA(I1+1)-1 WAS CHANGED TO
!     DO 430 J1=IA(I1)+1,ICG(I1)-1

!     DO 550 J=IA(I)+1,IA(I+1)-1 WAS CHANGED TO
!     DO 550 J=IA(I)+1,ICG(I)-1

!     DO 530 J1=IA(I1)+1,IA(I1+1)-1 WAS CHANGED TO
!     DO 530 J1=IA(I1)+1,ICG(I1)-1

! 2.  THE EXPLANATORY PART IN SUBROUTINE AMG1R5 HAS BEEN ENLARGED TO
!     AVOID MISUNDERSTANDINGS IN THE DEFINITION OF THE ARGUMENT LIST.

!-----------------------------------------------------------------------

!     CHANGE AGAINST VERSION 1.4, OCTOBER, 1990 (BY JOHN W. RUGE)

! 1.  A BUG IN SUBROUTINE CHECK HAS BEEN REMOVED. IF THE ORIGINAL MATRIX
!     WAS STORED IN AN UNSYMMETRIC WAY, THE SYMMETRIZATION BY AMG1R3
!     COULD STILL FAIL UNDER CERTAIN CIRCUMSTANCES, AND WAS NOT FIXED
!     IN THE PREVIOUS VERSION. IN ADDITION, THE ROUTINE WAS CHANGED
!     IN ORDER TO AVOID SOME UNNECESSARY ROW SEARCHES FOR TRANSOSE
!     ENTRIES.

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine amg1r5 here]
subroutine amg1r5(a,ia,ja,u,f,ig, &
      nda,ndia,ndja,ndu,ndf,ndig,nnu,matrix, &
      iswtch,iout,iprint, &
      levelx,ifirst,ncyc,eps,madapt,nrd,nsolco,nru, &
      ecg1,ecg2,ewt2,nwt,ntr, &
      ierr)
   
   !         -----------------------------------------------
   !         | AMG-MODULE FOR SOLVING LINEAR SYSTEMS L*U=F |
   !         -----------------------------------------------
   
   !         --------------------------------------------------------------
   
   !     ASSUMPTIONS ON L:
   
   !         THE PROGRAM REQUIRES:
   
   !             - DIAGONAL ENTRIES ARE ALWAYS POSITIVE (ON ALL GRIDS);
   !             - L IS A SQUARE MATRIX WHICH IS EITHER REGULAR OR SINGULAR
   !               WITH ROWSUMS=0.
   
   !         FOR THEORETICAL REASONS THE FOLLOWING SHOULD HOLD:
   
   !             - L POSITIVE DEFINITE (OR SEMI-DEFINITE WITH ROWSUM=0)
   !             - L "ESSENTIALLY" POSITIVE TYPE, I.E.,
   
   !                  -- DIAGONAL ENTRIES MUST BE > 0 ;
   !                  -- MOST OF THE OFF-DIAGONAL ENTRIES <= 0 ;
   !                  -- ROWSUMS SHOULD BE >= 0 .
   
   
   !     THE USER HAS TO PROVIDE THE MATRIX L, THE RIGHT HAND SIDE F AND
   !     CERTAIN POINTER VECTORS IA AND JA.
   
   !         --------------------------------------------------------------
   
   !     STORAGE OF L:
   
   !         THE NON-ZERO ENTRIES OF THE MATRIX L ARE STORED IN
   !         "COMPRESSED" SKY-LINE FASHION IN A 1-D VECTOR A, I.E., ROW
   !         AFTER ROW, EACH ROW STARTING WITH ITS DIAGONAL ELEMENT. THE
   !         OTHER NON-ZERO ROW ENTRIES FOLLOW THEIR DIAGONAL ENTRY IN ANY
   !         ORDER.
   
   
   !         IN ORDER TO IDENTIFY EACH ELEMENT IN A, THE USER HAS TO
   !         PROVIDE TWO POINTER ARRAYS IA AND JA. IF NNU DENOTES THE TOTAL
   !         NUMBER OF UNKNOWNS, THE NON-ZERO ENTRIES OF ANY ROW I OF L
   !         (1.LE.I.LE.NNU) ARE STORED IN A(J) WHERE THE RANGE OF J
   !         IS GIVEN BY
   
   !                     IA(I) .LE. J .LE. IA(I+1)-1.
   
   !         THUS, IA(I) POINTS TO THE POSITION OF THE DIAGONAL ENTRY OF
   !         ROW I WITHIN THE VECTOR A. IN PARTICULAR,
   
   !                     IA(1) = 1 ,  IA(NNU+1) = 1 + NNA
   
   !         WHERE NNA DENOTES THE TOTAL NUMBER OF MATRIX ENTRIES STORED.
   !         THE POINTER VECTOR JA HAS TO BE DEFINED SUCH THAT
   !         ANY ENTRY A(J) CORRESPONDS TO THE UNKNOWN U(JA(J)), I.E.,
   !         JA(J) POINTS TO THE COLUMN INDEX OF A(J).
   !         IN PARTICULAR, A(IA(I)) IS THE DIAGONAL ENTRY OF ROW I
   !         AND CORRESPONDS TO THE UNKNOWN U(I): JA(IA(I))=I.
   
   !         IN THIS TERMINOLOGY, THE I-TH EQUATION READS AS FOLLOWS
   !         (FOR ANY I WITH  1.LE.I.LE.NNU):
   
   !                  F(I) =        SUM      A(J) * U(JA(J))
   !                           J1.LE.J.LE.J2
   
   !         WHERE F(I) DENOTES THE I-TH COMPONENT OF THE RIGHT HAND
   !         SIDE AND
   
   !                     J1 = IA(I) ,  J2 = IA(I+1)-1.
   
   !         NOTES: THE ENTRY IA(NNU+1) HAS TO POINT TO THE FIRST FREE
   !                ENTRY IN VECTORS A AND JA, RESPECTIVELY. OTHERWISE,
   !                AMG CANNOT KNOW THE LENGTH OF THE LAST MATRIX ROW.
   
   !                THE INPUT VECTORS A, IA AND JA ARE CHANGED BY AMG1R5.
   !                SO, AFTER RETURN FROM AMG1R5, THE PACKAGE MUST NOT
   !                BE CALLED A SECOND TIME WITHOUT HAVING NEWLY DEFINED
   !                THE INPUT VECTORS AND USING ISWTCH=4. OTHERWISE, THE
   !                SETUP PHASE WILL FAIL.
   !                  ON THE OTHER HAND, RUNNING AMG A SECOND TIME ON THE
   !                SAME INPUT DATA WITH ISWTCH=4 HAS NO SENSE, BECAUSE
   !                THE RESULTS OF THE FIRST SETUP PHASE ARE STILL STORED
   !                AND THUS THIS PHASE CAN BE SKIPPED IN A SECOND CALL.
   !                IN ORDER TO DO THIS, SET ISWTCH TO 1, 2 OR 3.
   
   !-----------------------------------------------------------------------
   
   !         THE FORM OF THE CALLING PROGRAM HAS TO BE AS FOLLOWS:
   
   !               PROGRAM DRIVER
   !         C
   !               REAL A(#NDA),U(#NDU),F(#NDF)
   !               INTEGER IA(#NDIA),JA(#NDJA),IG(#NDIG)
   !         C
   !               NDA  = #NDA
   !               NDU  = #NDU
   !               NDF  = #NDF
   !               NDIA = #NDIA
   !               NDJA = #NDJA
   !               NDIG = #NDIG
   !         C
   !         C     SET UP A, F, IA, JA AND SPECIFY NECESSARY PARAMETERS
   !         C
   !               ....
   !               ....
   !         C
   !               CALL AMG1R5(A,IA,JA,U,F,IG,
   !        +                  NDA,NDIA,NDJA,NDU,NDF,NDIG,NNU,MATRIX,
   !        +                  ISWTCH,IOUT,IPRINT,
   !        +                LEVELX,IFIRST,NCYC,EPS,MADAPT,NRD,NSOLCO,NRU,
   !        +                  ECG1,ECG2,EWT2,NWT,NTR,
   !        +                  IERR)
   !         C
   !               ....
   !               ....
   !         C
   !               STOP
   !               END
   
   !-----------------------------------------------------------------------
   
   !     INPUT VIA ARRAYS (SEE ABOVE):
   
   !     A        -   MATRIX L
   
   !     IA       -   POINTER VECTOR
   
   !     JA       -   POINTER VECTOR
   
   !     U        -   FIRST APPROXIMATION TO SOLUTION
   
   !     F        -   RIGHT HAND SIDE
   
   
   !-----------------------------------------------------------------------
   
   
   !     SCALAR INPUT PARAMETERS OF AMG1R5:
   
   !     THE INPUT PARAMETERS OF AMG1R5 IN THE LIST BELOW ARE ARRANGED
   !     ACCORDING TO THEIR IMPORTANCE TO THE GENERAL USER. THE PARAMETERS
   !     PRECEEDED BY A * MUST BE SPECIFIED EXPLICITELY. ALL THE OTHER
   !     PARAMETERS ARE SET TO STANDARD VALUES IF ZERO ON INPUT.
   
   !     THERE ARE FOUR CLASSES OF INPUT PARAMETERS WITH DECREASING PRI-
   !     ORITY:
   
   !     1. PARAMETERS DESCRIBING THE USER-DEFINED PROBLEM AND DIMENSIONING
   !        OF VECTORS IN THE CALLING PROGRAM
   
   !     2. PARAMETERS SPECIFYING SOME GENERAL ALGORITHMIC ALTERNATIVES AND
   !        THE AMOUNT OF OUTPUT DURING SOLUTION
   
   !     3. PARAMETERS CONTROLLING THE MULTIGRID CYCLING DURING THE SOLU-
   !        TION PHASE
   
   !     4. PARAMETERS CONTROLLING THE CREATION OF COARSER GRIDS AND INTER-
   !        POLATION FORMULAS.
   
   !     ONLY THE CLASS 1 - PARAMETERS MUST BE SPECIFIED EXPLICITELY BY
   !     THE USER. CLASS 2 - PARAMETERS CONTROL THE GENERAL PERFORMANCE OF
   !     AMG1R5. CHANGING THEM DOESN'T REQUIRE UNDERSTANDING THE AMG -
   !     ALGORITHM. SPECIFYING NON-STANDARD-VALUES FOR CLASS 3 - PARAMETERS
   !     PRESUPPOSES A GENERAL KNOWLEDGE OF MULTIGRID METHODS, WHEREAS THE
   !     FUNCTION OF CLASS 4 - PARAMETERS IS ONLY UNDERSTANDABLE AFTER
   !     STUDYING THE AMG-ALGORITHM IN DETAIL. FORTUNATELY IN MOST CASES
   !     THE CHOICE OF CLASS 3 AND 4 - PARAMETERS ISN'T CRITICAL AND USING
   !     THE AMG1R5 - SUPPLIED STANDARD VALUES SHOULD GIVE SATISFACTORY
   !     RESULTS.
   
   !         --------------------------------------------------------------
   
   !     CLASS 1 - PARAMETERS:
   
   !  *  NDA      -   DIMENSIONING OF VECTOR A IN CALLING PROGRAM
   
   !  *  NDIA     -   DIMENSIONING OF VECTOR IA IN CALLING PROGRAM
   
   !  *  NDJA     -   DIMENSIONING OF VECTOR JA IN CALLING PROGRAM
   
   !  *  NDU      -   DIMENSIONING OF VECTOR U IN CALLING PROGRAM
   
   !  *  NDF      -   DIMENSIONING OF VECTOR F IN CALLING PROGRAM
   
   !  *  NDIG     -   DIMENSIONING OF VECTOR IG IN CALLING PROGRAM
   
   !  *  NNU      -   NUMBER OF UNKNOWNS
   
   !  *  MATRIX   -   INTEGER VALUE CONTAINING INFO ABOUT THE MATRIX L.
   
   !                  1ST DIGIT OF MATRIX  --  ISYM:
   !                    =1: L IS SYMMETRIC;
   !                    =2: L IS NOT SYMMETRIC.
   
   !                  2ND DIGIT OF MATRIX  --  IROW0:
   !                    =1: L HAS ROWSUM ZERO;
   !                    =2: L DOES NOT HAVE ROWSUM ZERO.
   
   !         --------------------------------------------------------------
   
   !     CLASS 2 - PARAMETERS:
   
   !     ISWTCH   -   PARAMETER CONTROLLING WHICH MODULES OF AMG1R5 ARE TO
   !                  BE USED.
   !                    =1:   CALL FOR -----, -----, -----, WRKCNT.
   !                    =2:   CALL FOR -----, -----, SOLVE, WRKCNT.
   !                    =3:   CALL FOR -----, FIRST, SOLVE, WRKCNT.
   !                    =4:   CALL FOR SETUP, FIRST, SOLVE, WRKCNT.
   !                  SETUP DEFINES THE OPERATORS NEEDED IN THE SOLUTION
   !                         PHASE.
   !                  FIRST INITIALIZES THE SOLUTION VECTOR (SEE PARAMETER
   !                         IFIRST).
   !                  SOLVE COMPUTES THE SOLUTION BY AMG CYCLING (SEE
   !                         PARAMETER NCYC).
   !                  WRKCNT PROVIDES THE USER WITH INFORMATION ABOUT
   !                         RESIDUALS, STORAGE REQUIREMENTS AND CP-TIMES
   !                         (SEE PARAMETER IOUT).
   !                  IF AMG1R5 IS CALLED THE FIRST TIME, ISWTCH HAS TO
   !                  BE =4. INDEPENDENT OF ISWTCH, SINGLE MODULES CAN BE
   !                  BYPASSED BY A PROPER CHOICE OF THE CORRESPONDING
   !                  PARAMETER.
   
   !     IOUT     -   PARAMETER CONTROLLING THE AMOUNT OF OUTPUT DURING
   !                  SOLUTION PHASE:
   
   !                  1ST DIGIT: NOT USED; HAS TO BE NON-ZERO.
   
   !                  2ND DIGIT:
   !                    =0: NO OUTPUT (EXCEPT FOR MESSAGES)
   !                    =1: RESIDUAL BEFORE AND AFTER SOLUTION PROCESS
   !                    =2: ADD.: STATISTICS ON CP-TIMES AND STORAGE REQUI-
   !                        REMENTS
   !                    =3: ADD.: RESIDUAL AFTER EACH AMG-CYCLE
   
   !     IPRINT   -   PARAMETER SPECIFYING THE FORTRAN UNIT NUMBERS FOR
   !                  OUTPUT:
   
   !                  1ST DIGIT: NOT USED; HAS TO BE NON-ZERO
   
   !                  2ND AND 3RD DIGIT  --  IUP: UNIT NUMBER FOR RESULTS
   
   !                  4TH AND 5TH DIGIT  --  IUM: UNIT NUMBER FOR MESSAGES
   
   !         --------------------------------------------------------------
   
   !     CLASS 3 - PARAMETERS:
   
   !     LEVELX   -   MAXIMUM NUMBER OF MG-LEVELS TO BE CREATED (>=1).
   
   !     IFIRST   -   PARAMETER FOR FIRST APPROXIMATION.
   
   !                  1ST DIGIT OF IFIRST: NOT USED; HAS TO BE NON-ZERO.
   
   !                  2ND DIGIT OF IFIRST  --  ITYPU:
   !                    =0: NO SETTING OF FIRST APPROXIMATION,
   !                    =1: FIRST APPROXIMATION CONSTANT TO ZERO,
   !                    =2: FIRST APPROXIMATION CONSTANT TO ONE,
   !                    =3: FIRST APPROXIMATION IS RANDOM FUNCTION WITH
   !                        THE CONCRETE RANDOM SEQUENCE BEING DETERMINED
   !                        BY THE FOLLWING DIGITS.
   
   !                  REST OF IFIRST  --  RNDU:
   !                    DETERMINES THE CONCRETE RANDOM SEQUENCE USED IN
   !                    THE CASE ITYPU=3. (IFIRST=13 IS EQUIVALENT TO
   !                    IFIRST=1372815)
   
   !     NCYC     -   INTEGER PARAMETER DESCRIBING THE TYPE OF CYCLE TO BE
   !                  USED AND THE NUMBER OF CYCLES TO BE PERFORMED.
   
   !                  1ST DIGIT OF NCYC  --  IGAM:
   !                    =1: V -CYCLE,
   !                    =2: V*-CYCLE,
   !                    =3: F -CYCLE,
   !                    =4: W -CYCLE.
   !                  IF NCYC IS NEGATIV, THEN THE APPROXIMATION OF THE
   !                  PROBLEM ON THE SECOND FINEST GRID IS COMPUTED BY
   !                  IGAM V-CYCLES ON THAT PARTICULAR GRID.
   
   !                  2ND DIGIT OF NCYC  --  ICGR:
   !                    =0: NO CONJUGATE GRADIENT,
   !                    =1: CONJUGATE GRADIENT (ONLY FIRST STEP OF CG),
   !                    =2: CONJUGATE GRADIENT (FULL CG).
   
   !                  3RD DIGIT OF NCYC  --  ICONV:
   !                    CONVERGENCE CRITERION FOR THE USER-DEFINED PROBLEM
   !                    (FINEST GRID):
   !                    =1: PERFORM A FIXED NUMBER OF CYCLES AS GIVEN BY
   !                        NCYCLE (SEE BELOW)
   !                    =2: STOP, IF  ||RES|| < EPS
   !                    =3: STOP, IF  ||RES|| < EPS * |F|
   !                    =4: STOP, IF  ||RES|| < EPS * |U| * |DIAG|
   !                    WITH ||RES|| = L2-NORM OF RESIDUAL,
   !                           EPS     (SEE INPUT PARAMETER EPS)
   !                           |F|   = SUPREMUM NORM OF RIGHT HAND SIDE
   !                           |U|   = SUPREMUM NORM OF SOLUTION
   !                         |DIAG|  = MAXIMAL DIAGONAL ENTRY IN MATRIX L
   !                    NOTE THAT IN ANY CASE THE SOLUTION PROCESS STOPS
   !                    AFTER AT MOST NCYCLE CYCLES.
   
   !                  REST OF NCYC  --  NCYCLE:
   !                    MAXIMAL NUMBER OF CYCLES TO BE PERFORMED (>0) OR
   !                    NCYCLE=0: NO CYCLING.
   
   !     EPS      -   CONVERGENCE CRITERION FOR SOLUTION PROCESS: (SEE
   !                  PARAMETER NCYC). NOTE THAT NO MORE THAN NCYCLE CYCLES
   !                  ARE PERFORMED, REGARDLESS OF EPS.
   
   !     MADAPT   -   INTEGER VALUE SPECIFYING THE CHOICE OF COARSEST
   !                  GRID IN CYCLING:
   
   !                  1ST DIGIT OF MADAPT  --  MSEL:
   !                    =1: IN CYCLING, ALL GRIDS CONSTRUCTED IN THE SETUP
   !                        PHASE ARE USED WITHOUT CHECK.
   !                    =2: THE NUMBER OF GRIDS IS AUTOMATICALLY REDUCED
   !                        IF THE CONVERGENCE FACTOR ON THE COARSER GRIDS
   !                        IS FOUND TO BE LARGER THAN A GIVEN VALUE FAC
   !                        (SEE BELOW).
   
   !                  REST OF MADAPT  --  FAC
   !                        THE REST OF MADAPT DEFINES THE FRACTIONAL PART
   !                        OF A REAL NUMBER FAC BETWEEN 0.1 AND 0.99, E.G.
   !                        MADAPT=258 MEANS MSEL=2 AND FAC=0.58. IF MADAPT
   !                        CONSISTS OF ONLY ONE DIGIT, FAC IS SET TO 0.7
   !                        BY DEFAULT.
   
   
   !     NRD      -   PARAMETER DESCRIBING RELAXATION (DOWNWARDS):
   
   !                  1ST DIGIT OF NRD: NOT USED; HAS TO BE NON-ZERO.
   
   !                  2ND DIGIT OF NRD  --  NRDX:
   !                    ACTUAL NUMBER OF SMOOTHING STEPS TO BE PERFORMED
   !                    THE TYPE OF WHICH IS GIVEN BY THE FOLLOWING DIGITS
   
   !                  FOLLOWING DIGITS  --  ARRAY NRDTYP:
   !                    =1: RELAXATION OVER THE F-POINTS ONLY
   !                    =2: FULL GS SWEEP
   !                    =3: RELAXATION OVER THE C-POINTS ONLY
   !                    =4: FULL MORE COLOR SWEEP, HIGHEST COLOR FIRST
   
   !     NSOLCO   -   PARAMETER CONTROLLING THE SOLUTION ON COARSEST GRID:
   
   !                  1ST DIGIT  --  NSC:
   !                    =1: GAUSS-SEIDEL METHOD
   !                    =2: DIRECT SOLVER (YALE SMP)
   
   !                  REST OF NSOLCO  --  NRCX: (ONLY IF NSC=1)
   !                  NUMBER OF GS SWEEPS ON COARSEST GRID (>=0).
   !                  IF NRCX=0, THEN AS MANY GS SWEEPS ARE PERFORMED
   !                  AS ARE NEEDED TO REDUCE THE RESIDUAL BY TWO ORDERS
   !                  OF MAGNITUDE. (MAXIMAL 100 RELAXATION SWEEPS)
   
   !     NRU      -   PARAMETER FOR RELAXATION (UPWARDS), ANALOGOUS TO NRD.
   
   !         --------------------------------------------------------------
   
   !     CLASS 4 - PARAMETERS:
   
   !     ECG1,ECG2-   REAL PARAMETERS AFFECTING THE CREATION OF COARSER
   !     EWT2     -   GRIDS AND/OR THE DEFINITION OF THE INTERPOLATION.
   !                  THE CHOICE OF THESE PARAMETERS DEPENDS ON
   !                  THE ACTUAL AMG VERSION (SEE SUBROUTINE CRSNG)
   
   !     NWT      -   INTEGER PARAMETER AFFECTING THE CREATION OF COARSER
   !                  GRIDS AND/OR THE DEFINITION OF THE INTERPOLATION.
   !                  THE CHOICE OF THIS PARAMETER DEPENDS ON
   !                  THE ACTUAL AMG VERSION (SEE SUBROUTINE CRSNG)
   
   !     NTR      -   PARAMETER CONTROLLING COARSE-GRID OPERATOR TRUNCATION
   !                    =0: PAIRS OF ZEROES ARE REMOVED FROM COARSE GRID
   !                        OPERATORS
   !                    =1: NO COARSE-GRID OPERATOR TRUNCATION
   
   
   !-----------------------------------------------------------------------
   
   !     OUTPUT:
   
   !     U        -   CONTAINS THE COMPUTED SOLUTION
   
   
   !     IERR     -   ERROR PARAMETER:
   
   !                    >0: FATAL ERROR (ABNORMAL TERMINATION OF AMG1R5)
   !                    <0: NON-FATAL ERROR (EXECUTION OF AMG1R5 CONTINUES)
   
   !                  ERROR CODES IN DETAIL:
   
   !                  1. DIMENSIONING TOO SMALL FOR VECTOR
   !                        A      (IERR = 1)
   !                        IA     (IERR = 2)
   !                        JA     (IERR = 3)
   !                        U      (IERR = 4)
   !                        F      (IERR = 5)
   !                        IG     (IERR = 6)
   
   !                     NO YALE-SMP BECAUSE OF STORAGE (NDA TOO SMALL):
   !                               (IERR = -1)
   !                     NO YALE-SMP BECAUSE OF STORAGE (NDJA TOO SMALL):
   !                               (IERR = -3)
   !                     NO CG BECAUSE OF STORAGE (NDU TOO SMALL):
   !                               (IERR = -4)
   !                     NO SPACE FOR TRANSPOSE OF INTERPOLATION (NDA OR
   !                                                     NDJA TOO SMALL):
   !                               (IERR = -1)
   
   !                  2. INPUT DATA ERRONEOUS:
   
   !                     A-ENTRY MISSING, ISYM = 1:           (IERR = -11)
   !                     PARAMETER MATRIX MAY BE ERRONEOUS:   (IERR = -12)
   !                     DIAGONAL ELEMENT NOT STORED FIRST:   (IERR =  13)
   !                     DIAGONAL ELEMENT NOT POSITIV:        (IERR =  14)
   !                     POINTER IA ERRONEOUS:                (IERR =  15)
   !                     POINTER JA ERRONEOUS:                (IERR =  16)
   !                     PARAMETER ISWTCH ERRONEOUS:          (IERR =  17)
   !                     PARAMETER LEVELX ERRONEOUS:          (IERR =  18)
   
   !                  3. ERRORS OF THE AMG1R5-SYSTEM (SHOULD NOT OCCUR):
   
   !                     TRANSPOSE A-ENTRY MISSING:           (IERR =  21)
   !                     INTERPOLATION ENTRY MISSING:         (IERR =  22)
   
   !                  4. ALGORITHMIC ERRORS:
   
   !                     CG-CORRECTION NOT DEFINED:           (IERR =  31)
   !                     NO YALE-SMP BECAUSE OF ERROR IN
   !                     FACTORIZATION:                       (IERR = -32)
   
   !-----------------------------------------------------------------------
   
   !     WORK SPACE:
   
   !     THE INTEGER VECTOR IG HAS TO BE PASSED TO AMG1R5 AS WORK SPACE.
   
   !-----------------------------------------------------------------------
   
   !     DIMENSIONING OF INPUT VECTORS AND WORK SPACE:
   
   !     IT'S IMPOSSIBLE TO TELL IN ADVANCE THE EXACT STORAGE REQUIREMENTS
   !     OF AMG. THUS, THE FOLLOWING FORMULAS GIVE ONLY REASONABLE GUESSES
   !     FOR THE VECTOR LENGTHS WHICH HAVE TO BE DECLARED IN THE CALLING
   !     PROGRAM. IN THESE FORMULAS NNA DENOTES THE NUMBER OF NON-ZERO
   !     ENTRIES IN THE INPUT-MATRIX L AND NNU IS THE NUMBER OF UNKNOWNS.
   
   !     VECTOR         NEEDED LENGTH (GUESS)
   !       A               3*NNA + 5*NNU
   !       JA              3*NNA + 5*NNU
   !       IA              2.2*NNU
   !       U               2.2*NNU
   !       F               2.2*NNU
   !       IG              5.4*NNU
   
   !-----------------------------------------------------------------------
   
   
   !     STANDARD CHOICES OF PARAMETERS (AS FAR AS MEANINGFUL):
   
   !          ISWTCH = 4
   !          IOUT   = 12
   !          IPRINT = 10606
   
   !          LEVELX = 25
   !          IFIRST = 13
   !          NCYC   = 10110
   !          EPS    = 1.D-12
   !          MADAPT = 27
   !          NRD    = 1131
   !          NSOLCO = 110
   !          NRU    = 1131
   
   !          ECG1   = 0.
   !          ECG2   = 0.25
   !          EWT2   = 0.35
   !          NWT    = 2
   !          NTR    = 0
   
   !     IF ANY ONE OF THESE PARAMETERS IS 0 ON INPUT, ITS CORRESPONDING
   !     STANDARD VALUE IS USED BY AMG1R5.
   
   !-----------------------------------------------------------------------
   
   !     PORTABILITY RESTRICTIONS:
   
   !     1. ROUTINE CTIME IS MACHINE DEPENDENT AND HAS TO BE ADAPTED TO
   !        YOUR COMPUTER INSTALLATION OR REPLACED BY A DUMMY ROUTINE.
   
   !     2. MOST INPUT PARAMETERS ARE COMPOSED OF SEVERAL DIGITS, THEIR
   !        SIGNIFICANCE HAVING BEEN DESCRIBED ABOVE. BE SURE NOT TO ENTER
   !        MORE DIGITS THAN YOUR COMPUTER CAN STORE ON AN INTEGER VARI-
   !        ABLE.
   
   !     3. APART FROM FORTRAN INTRINSIC FUNCTIONS AND SERVICE ROUTINES,
   !        THERE IS ONLY ONE EXTERNAL REFERENCE TO A PROGRAM NOT CONTAINED
   !        IN THE AMG1R5 - SYSTEM, I.E. THE LINEAR SYSTEM SOLVER NDRV OF
   !        THE YALE SPARSE MATRIX PACKAGE. IF YOU HAVN'T ACCESS TO THIS
   !        PACKAGE, ENTER A DUMMY ROUTINE NDRV AND AVOID CHOOSING NSC=2
   !        (SUBPARAMETER OF NSOLCO). THEN NDRV ISN'T CALLED BY AMG1R5.
   !        IN THIS CASE, HOWEVER, INDEFINITE PROBLEMS WILL NOT BE SOLV-
   !        ABLE.
   !          THE YALE SPARSE MATRIX PACKAGE IS FREELY AVAILABLE FOR NON-
   !        PROFIT PURPOSES. CONTACT THE DEPARTMENT OF COMPUTER SCIENCE,
   !        YALE UNITVERSITY.
   
   !     4. IN AMG1R5 THERE IS THE PARAMETER LRATIO, DENOTING THE RATIO
   !        OF SPACE OCCUPIED BY A REAL REAL VARIABLE AND
   !        THAT OF AN INTEGER. FOR THE IBM-VERSION LRATIO HAS BEEN SET
   !        TO 2. CHANGE THIS VALUE IF NECESSARY. (THE SAME HAS TO BE
   !        DONE WITH THE YALE SMP-ROUTINE NDRV.)
   
   
   !-----------------------------------------------------------------------
   
   !     AUTHORS:
   
   !          JOHN RUGE, FORT COLLINS (USA),
   !              INSTITUTE FOR COMPUTATIONAL STUDIES AT CSU;
   
   !          KLAUS STUEBEN, D-5205 ST. AUGUSTIN (W.-GERMANY),
   !              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG (GMD).
   
   !          ROLF HEMPEL, D-5205 ST. AUGUSTIN (W.-GERMANY),
   !              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG (GMD).
   
   !-----------------------------------------------------------------------
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ time(20)
   integer iarr(25),ia(*),ja(*),ig(*)
   
   !===> LRATIO HAS TO BE SET TO THE NUMBER OF INTEGERS OCCUPYING THE SAME
   !     AMOUNT OF STORAGE AS ONE REAL REAL.
   
   parameter (lratio=2)
   
   !===> MAXGR IS THE MAXIMAL NUMBER OF GRIDS. CHANGING THIS UPPER LIMIT
   !     JUST REQUIRES CHANGING THE PARAMETER STATEMENT.
   
   parameter (maxgr=25)
   _REAL_ resi(maxgr)
   integer imin(maxgr),imax(maxgr),iminw(maxgr),imaxw(maxgr), &
         nstcol(maxgr)
   external setup,first,solve,wrkcnt
   save
   
   ierr = 0
   
   !===> SET PARAMETERS TO STANDARD VALUES, IF NECCESSARY
   
   
   if (iout /= 0) then
      call idec(iout,2,ndigit,iarr)
      kout = iarr(2)
   else
      kout = 2
   end if
   
   if (iswtch /= 0) then
      kswtch = iswtch
   else
      kswtch = 4
   end if
   
   if (levelx > 0) then
      kevelx = min(levelx,maxgr)
   else if (levelx < 0) then
      goto 70
   else
      kevelx = maxgr
   end if
   
   if (iprint /= 0) then
      call idec(iprint,4,ndigit,iarr)
      iup = 10*iarr(2)+iarr(3)
      ium = iarr(4)
   else
      iup = 6
      ium = 6
   end if
   icgst = nnu+3
   ndicg = (ndig-icgst+1)/2
   if (ndicg <= 0) goto 60
   
   goto (10,20,30,40), kswtch
   write (ium,9920)
   ierr = 17
   return
   
   40 call setup (nnu,matrix,kevelx,ecg1,ecg2,ewt2,nwt,ntr,ierr, &
         a,u,ia,ja,ig,imin,imax,iminw,imaxw,ig(icgst), &
         ig(icgst+ndicg),nstcol,iarr,time,levels,irow0, &
         nda,ndia,ndja,ndu,ndf,ndicg,ium,mda,mdia, &
         mdja,mdu,mdf,mdig,maxgr,lratio)
   if (ierr > 0) return
   30 call first (ifirst,u,imin,imax,iarr,irow0)
   20 call solve (madapt,ncyc,nrd,nsolco,nru,kout,ierr,a,u,f,ia,ja,ig, &
         eps,imin,imax,iminw,imaxw,ig(icgst),ig(icgst+ndicg), &
         nstcol,iarr,time,ncyc0,irow0,levels,nda,ndja, &
         ndu,ndf,mda,mdja,mdu,mdf,iup,ium,resi,res0,res)
   if (ierr > 0) return
   10 call wrkcnt(kout,ia,ig,imin,imax,iminw,levels,time,ncyc0,iup, &
         mda,mdia,mdja,mdu,mdf,mdig,res0,res)
   return
   60 write (ium,9910)
   ierr = 6
   return
   70 write (ium,9930)
   ierr = 18
   return
   9910 format (' *** ERROR IN AMG1R5: NDIG TOO SMALL ***')
   9920 format (' *** ERROR IN AMG1R5: ILLEGAL PARAMETER ISWTCH ***')
   9930 format (' *** ERROR IN AMG1R5: ILLEGAL PARAMETER LEVELX ***')
end subroutine amg1r5 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     GENERAL AMG1R5 SETUP-SUBROUTINES

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!.......................................................................

!     SETUP                                                  SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine setup here]
subroutine setup(nnu,matrix,levelx,ecg1,ecg2,ewt2,nwt,ntr,ierr, &
      a,u,ia,ja,iw,imin,imax,iminw,imaxw,icg,ifg, &
      nstcol,iarr,time,levels,irow0,nda,ndia,ndja, &
      ndu,ndf,ndicg,ium,mda,mdia,mdja,mdu,mdf,mdig, &
      maxgr,lratio)
   
   !     PREPARATION PHASE OF AMG1R5 (GENERAL PART)
   
   _REAL_ a(*),u(*),ecg1,ecg2,ewt2
   _REAL_ time(*)
   integer iarr(*),ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*), &
         imaxw(*),icg(*),ifg(*),nstcol(*)
   external idec,check,crsng
   save
   
   !===> DECOMPOSE "MATRIX"
   
   call idec(matrix,2,ndigit,iarr)
   isym  = iarr(1)
   irow0 = iarr(2)
   
   !===> PREPARATION (IGNORED IN TIMING)
   
   imin(1) = 1
   imax(1) = nnu
   call check(ierr,a,ia,ja,imin,imax,icg,ifg,time,isym,irow0,nda, &
         ndicg,ndja,ium)
   if (ierr > 0) return
   
   !===> RESET TIME COUNTERS
   
   do i=1,20
      time(i) = 0.0
   end do
   30 continue
   
   !===> DEFINE COARSER GRIDS + OPERATORS. RESET LEVELS IF NECESSARY.
   
   call crsng(levelx,ecg1,ecg2,ewt2,nwt,ntr,ierr,a,u,ia,ja,iw,imin, &
         imax,iminw,imaxw,icg,ifg,nstcol,levels,irow0,nda, &
         ndja,ndia,ndu,ndf,ndicg,time,ium,mda,mdia,mdja,mdu,mdf, &
         mdig,maxgr,lratio)
   return
end subroutine setup 

!.......................................................................

!     CHECK                                                 SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine check here]
subroutine check(ierr,a,ia,ja,imin,imax,icg,ifg,time,isym,irow0, &
      nda,ndicg,ndja,ium)
   
   !     CHECKS FOR SEVERAL PROPERTIES OF A, IA, JA. IN PARTICULAR,
   !     CHECKS FOR SYMMETRIC STORAGE OF GIVEN MATRIX (I.E. L(I,J) IS
   !     STORED IN A IFF L(J,I) IS STORED IN A).
   
   !     IF STORAGE IS SYMMETRIC, PAIRS OF ZEROES ARE REMOVED (IF THERE
   !     ARE ANY) AND PROGRAM EXECUTION CONTINUES.
   
   !     IF, HOWEVER, STORAGE OF A IS NOT SYMMETRIC, IT IS SYMMETRIZED.
   !     IF ISYM.EQ.1, THE MISSING TRANSPOSE CONNECTIONS ARE COPIED.
   !     IF ISYM.NE.1, THE STORAGE OF THE MATRIX A IS SYMMETRIZED (BY
   !     ADDING CERTAIN ZERO ELEMENTS). THEN PAIRS OF ZEROES ARE REMOVED
   !     (IF THERE ARE ANY) AND PROGRAM EXECUTION CONTINUES.
   
   !     ARRAYS USED FOR TEMPORARY STORAGE: ICG, IFG
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*)
   _REAL_ time(*)
   integer ia(*),ja(*),imin(*),imax(*),icg(*),ifg(*)
   external trunc
   save
   
   !===> CHECK IF POINTERS IA, JA ARE REASONABLE
   
   if (ia(1) /= 1) goto 3300
   nnu = imax(1)
   if (nnu >= ndicg) goto 3200
   do i=1,nnu
      icg(i) = 0
   end do
   do i=1,nnu
      j1 = ia(i)
      j2 = ia(i+1)-1
      if (j2 < j1 .or. j2 > nda) goto 3300
      if (ja(j1) /= i)             goto 3320
      do j=j1,j2
         i1 = ja(j)
         if (i1 < 1 .or. i1 > nnu .or. icg(i1) == 1) goto 3310
         icg(i1) = 1
      end do
      do j=j1,j2
         icg(ja(j)) = 0
      end do
   end do
   10 continue
   nna = ia(nnu+1)-1
   
   !===> CHECK FOR PROPERTIES OF A. IN PARTICULAR, COUNT
   !===> MISSING STORAGE PLACES ("NEW"). RETURN IF NEW=0
   
   anormm = 0.0d0
   anormp = 0.0d0
   naoff  = 0
   napos  = 0
   naneg  = 0
   nazer  = 0
   new    = 0
   do i=1,nnu
      icg(i) = ia(i+1)-ia(i)
   end do
   
   do i=1,nnu
      d = a(ia(i))
      if (d <= 0.0d0) goto 3330
      deps = d*1.0d-12
      rowsum = d
      anormp = anormp+2.0d0*d**2
      do j=ia(i)+1,ia(i+1)-1
         rowsum = rowsum+a(j)
         if (a(j) >= deps) naoff = naoff+1
         i1 = ja(j)
         if (i1 > 0) goto 140
         ja(j) = -i1
         goto 150
         140 if(i1 < i) goto 135
         do j1=ia(i1)+1,ia(i1+1)-1
            if (ja(j1) /= i) cycle
            ja(j1) = -ja(j1)
            anormm = anormm+(a(j)-a(j1))**2
            anormp = anormp+(a(j)+a(j1))**2
            goto 150
         end do
         135 ja(j)=-ja(j)
         anormm = anormm+a(j)**2
         anormp = anormp+a(j)**2
         new = new+1
         icg(i1) = icg(i1)+1
         150 continue
      end do
      if (rowsum > deps) then
         napos = napos+1
      else if (rowsum < -deps) then
         naneg = naneg+1
      else
         nazer = nazer+1
      end if
   end do  ! i=1,nnu
   anormm = sqrt(anormm)
   anormp = sqrt(anormp)
   asym = anormm/anormp
   if (asym <= 1.0d-12) asym = 0.0d0
   
   !===> MESSAGES ON A
   
!  if (asym == 0.0d0) write (ium,9000)
!  if (asym /= 0.0d0) write (ium,9005) asym
!  if (naoff > 0)    write (ium,9010) naoff
!  if (naneg > 0)    write (ium,9020) naneg
!  if (nazer == nnu)  write (ium,9025)
!  if (naoff == 0 .and. naneg == 0) write (ium,9030)
   
   !===> WARNINGS
   
   if (isym == 1  .and. asym /= 0.0d0 .or. &
         isym > 1  .and. asym == 0.0d0 .or. &
         irow0 == 1 .and. nazer /= nnu  .or. &
         irow0 > 1 .and. nazer == nnu) then
      write (ium,9100)
      ierr = -12
   end if
   
   if (new > 0) goto 220
   write (ium,9600)
   goto 600
   
   !===> REPLACE A BY SYMMETRIZED VERSION
   
   220 if (nna+new >= nda) goto 3210
   if (nna+new >= ndja) goto 3220
   
   !===> EXTEND MATRIX A IN SITU
   
   ifg(1) = 1
   do i=2,nnu+1
      ifg(i) = ifg(i-1)+icg(i-1)
   end do
   do i=nnu,1,-1
      ishift = ifg(i)-ia(i)
      do j=ia(i+1)-1,ia(i),-1
         a(j+ishift)  =  a(j)
         ja(j+ishift) = ja(j)
      end do
      icg(i)  = ia(i+1)+ishift
      ia(i+1) = ifg(i+1)
   end do
   250 continue
   
   !===> SYMMETRIZE MATRIX A:  ISYM=1: COPY MISSING TRANSPOSE ENTRIES
   !                           ISYM=2: FILL IN ZEROES
   
   if (isym /= 1) then
      do i=1,nnu
         j2 = icg(i)-1
         do j=ia(i)+1,j2
            i1 = ja(j)
            if (i1 > 0) goto 450
            ja(j) = -i1
            jnew=icg(ja(j))
            a(jnew)= 0.0d0
            ja(jnew) = i
            icg(ja(j))=jnew+1
         450 continue
         end do
      end do
!     write (ium,9500) new
   else
      write (ium,9220) new
      ierr = -11
      do i=1,nnu
         j2 = icg(i)-1
         do j=ia(i)+1,j2
            i1 = ja(j)
            if (i1 > 0) goto 550
            ja(j) = -i1
            jnew=icg(ja(j))
            a(jnew)= 0.0d0
            ja(jnew) = i
            icg(ja(j))=jnew+1
         550 continue
         end do
      end do
   end if
   
   !===> REMOVE PAIRS OF ZEROES
   
   600 call trunc(1,0,a,ia,ja,imin,imax,time,ierr,ium)
   return
   
   !===> ERROR MESSAGES
   
   3200 write (ium,9200)
   ierr = 6
   return
   
   3210 write (ium,9210)
   ierr = 1
   return
   
   3220 write (ium,9230)
   ierr = 3
   return
   
   3300 write (ium,9300)
   ierr = 15
   return
   
   3310 write (ium,9310)
   ierr = 16
   return
   
   3320 write (ium,9320)
   ierr = 13
   return
   
   3330 write (ium,9330)
   ierr = 14
   return
   
   9000 format (' CHECK: A PROBABLY SYMMETRIC')
   9005 format (' CHECK: A PROBABLY NOT SYMMETRIC. MEASURE:',d11.3)
   9010 format (' CHECK: A PROBABLY NOT POS. TYPE:',i6,' OFF-DIAGONAL', &
         ' ELEMENTS POSITIVE')
   9020 format (' CHECK: A PROBABLY NOT POS. TYPE:',i6,' ROWSUMS', &
         ' NEGATIVE')
   9025 format (' CHECK: A PROBABLY SINGULAR - ROWSUMS ARE ZERO')
   9030 format (' CHECK: A PROBABLY POSITIVE TYPE')
   9100 format (' --- WARNG IN CHECK: PARAM MATRIX MAY BE BAD ---')
   
   9200 format (' *** ERROR IN CHECK: NDIG TOO SMALL ***')
   9210 format (' *** ERROR IN CHECK: NDA TOO SMALL ***')
   9220 format (' *** WARNG IN CHECK:',i5,' A-ENTRIES MISSING ***'/ &
         '     MISSING TRANSPOSE CONNECTIONS WILL BE FILLED IN')
   9230 format (' *** ERROR IN CHECK: NDJA TOO SMALL ***')
   9300 format (' *** ERROR IN CHECK: POINTER IA ERRONEOUS ***')
   9310 format (' *** ERROR IN CHECK: POINTER JA ERRONEOUS ***')
   9320 format (' *** ERROR IN CHECK: DIAGONAL IS NOT STORED FIRST ***')
   9330 format (' *** ERROR IN CHECK: DIAGONAL IS NON-POSITIVE ***')
   9500 format (' CHECK: STORAGE OF A HAS BEEN SYMMETRIZED BY', &
         ' INTRODUCING',i6,' ZEROES')
   9600 format (' CHECK: MATRIX A WAS SYMMETRICALLY STORED')
end subroutine check 

!.......................................................................

!     TRUNC                                                 SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine trunc here]
subroutine trunc(k,ntr,a,ia,ja,imin,imax,time,ierr,ium)
   
   !     TRUNCATES OPERATOR ON GRID K CORRESPONDING TO THE VALUE OF NTR.
   !     NTR HAS TO BE 0 OR 1:
   
   !       =0:    PAIRS OF ZEROES ARE REMOVED FROM COARSE GRID OPERATORS;
   !       =1:    COARSE GRID OPERATORS REMAIN UNCHANGED.
   
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),imin(*),imax(*)
   external ctime
   save
   
   if (ntr == 1) return
   
   call ctime(told)
   imn = imin(k)
   imx = imax(k)
   nna = ia(imx+1)-ia(imn)
   jpos = ia(imn)
   
   do i=imn,imx
      j1 = ia(i)+1
      j2 = ia(i+1)-1
      a(jpos)  = a(ia(i))
      ja(jpos) = i
      ia(i) = jpos
      jpos = jpos+1
      do j=j1,j2
         i1 = ja(j)
         if (i1 < 0) goto 250
         if (i1 < i) goto 230
         if (a(j) /= 0.0d0) goto 230
         do j3=ia(i1),ia(i1+1)-1
            if (ja(j3) /= i) goto 210
            jt = j3
            at = a(j3)
            goto 215
            210 continue
         end do
         goto 1000
         215 if (at /= 0.0d0) goto 230
         ja(jt) = -ja(jt)
         goto 250
         230 a(jpos)  = a(j)
         ja(jpos) = ja(j)
         jpos = jpos+1
         250 continue
      end do
   end do
   ia(imx+1) = jpos
   
   !===> EXIT
   
   call ctime(tnew)
   time(7) = time(7)+tnew-told
   return
   
   !===> ERROR MESSAGE
   
   1000 write (ium,9000) k
   ierr = 21
   return
   
   9000 format (' *** ERROR IN TRUNC: TRANSPOSE A-ENTRY MISSING ON GRID' &
         ,i3,' ***')
end subroutine trunc 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     AMG1R5 SETUP ROUTINES (VERSION 3)

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!.......................................................................

!     CRSNG                                           SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine crsng here]
subroutine crsng(levelx,eecg1,eecg2,eewt2,nnwt,ntr,ierr,a,u, &
      ia,ja,iw,imin,imax,iminw,imaxw,icg,ifg,nstcol, &
      levels,irow0,nda,ndja,ndia,ndu,ndf,ndicg,time, &
      ium,mda,mdia,mdja,mdu,mdf,mdig,maxgr,lratio)
   
   !     PERFORMS COARSENING, DEFINES INTERPOLATIONS AND CGC-OPERATORS
   
   !     ============== STANDARD VALUES OF PARAMETERS =====================
   
   !             ECG1=0.,    ECG2=0.25,    EWT2=0.35,   NWT=2
   
   !     ================ DESCRIPTION OF PARAMETERS =======================
   
   !     ECG1 --    DEFINES CRITERION FOR DETERMINING DIAGONAL
   !                DOMINANCE IN RWSRT. I.E., IF THE ABSOLUTE
   !                VALUE OF THE SUM OF THE OFF-DIAGONALS OF ROW
   !                I IS SMALLER THAN ECG1 TIMES THE ABSOLUTE
   !                VALUE OF THE DIAGONAL ENTRY, THEN POINT I IS
   !                IMMEDIATELY FORCED TO BE AN F-POINT IN THE
   !                PRE-COLORING ALGORITHM PCOL.
   !                IN THE SECOND PART (WINT), NO INTERPOLATION
   !                IS DEFINED FOR POINT I, AND NO POINTS USE
   !                I FOR INTERPOLATION. IN ADDITION, THE WEIGHT
   !                FOR POINT I IS NOT DISTRIBUTED TO OTHER POINTS
   !                WHEN DEFINING THE INTERPOLATION WEIGHTS FOR
   !                POINTS WHICH DEPEND ON POINT I. (THIS IS
   !                EQUIVALENT TO COMPLETELY IGNORING SUCH POINTS
   !                IN DETERMINING THE COARSE GRID AND INTERPOLATION
   !                WEIGHTS.
   
   !     ECG2 --    DEFINES STRONG CONNECTIONS (ALPHA IN THE PAPER)
   
   !     EWT2 --    DEFINES STRONG DEPENDENCE ON A SET (BETA IN THE PAPER)
   
   !     NWT  --    PARAMETER CONTROLLING THE DEFINITION OF INTERPOLATION
   !                FORMULAS:
   !                  =1 - CHECKING OF INT-FORMULA: OFF
   !                  =2 - CHECKING OF INT-FORMULA: ON
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*)
   _REAL_ time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),imaxw(*), &
         icg(*),ifg(*),nstcol(*)
   external rwsrt,pcol,pwint,opdfn,trunc,setifg
   save
   
   !===> ASSIGN DEFAULT VALUES IF ZERO
   
   nwt = nnwt
   ecg1 = eecg1
   ecg2 = eecg2
   ewt2 = eewt2
   if (nwt == 0)     nwt = 2
   if (ecg2 == 0.0d0) ecg2 = 0.25d0
   if (ewt2 == 0.0d0) ewt2 = 0.35d0
   
   !===> DECODE PARAMETER NWT
   
   5 ichk = nwt
   
   !===> COARSENING
   
   10 levels = min(levelx,maxgr)
   mmax   = maxgr
   
   !===> INITIALIZE PARAMETERS MDA - MDIW, LATER SET TO ACTUAL STORAGE
   !     REQUIREMENTS OF CORRESPONDING VECTORS
   
   mda   = 0
   mdja  = 0
   mdicg = 0
   mdir  = 0
   mdiw  = imax(1)
   
   !===> KFIRST IS THE NUMBER OF THE FIRST STORED GRID, IAJAS, IIAS AND
   !     IIRS ARE THE SHIFTS IN VECTORS A, JA, IA AND IR, RESPECTIVELY, AS
   !     COMPARED TO FULL STORAGE OF ALL GRIDS
   
   kerr   = 0
   kfirst = 1
   iajas  = 0
   iias   = 0
   iirs   = 0
   
   do k=2,levels
      
      !===> JTRST: INITIAL POINTER FOR WORK SPACE IN VECTOR A, TO CONTAIN
      !            THE STRONG TRANSPOSE CONNECTIONS
      !     NDJTR: AVAILABLE WORK SPACE
      
      20 jtrst = ia(imax(k-1)+1)
      ndjtr = lratio*(nda-jtrst+1)
      call rwsrt(k-1,ecg1,ecg2,ierr,a,ia,ja,iw,imin,imax,imaxw, &
            ifg,a(jtrst),time,ndjtr,ium,mdjtr)
      if (ierr > 0) then
         if (ierr >= 1.and.ierr <= 6) goto 30
         return
      end if
      mda = max(mda,jtrst+(mdjtr-1)/lratio+iajas)
      
      !===>   IRST: INITIAL POINTER FOR WORK SPACE IN VECTOR U, TO CONTAIN
      !             RESET STACK
      !       NDIR: AVAILABLE WORK SPACE
      
      irst = mdiw+1-iirs
      ndir = lratio*(ndu-irst+1)
      call pcol(k-1,ierr,ia,ja,iw,imin,imax,imaxw,icg,ifg,u(irst), &
            a(jtrst),time,ndicg,ndir,ium,mdicg,mdir,iias)
      if (ierr > 0) then
         if (ierr >= 1.and.ierr <= 6) goto 30
         return
      end if
      call pwint(k-1,ewt2,ichk,mmax,a,ia,ja,iw,imin,imax,iminw, &
            imaxw,ifg,icg,u(irst),time,ierr,irow0,ncolx, &
            nda,ndja,ndicg,ndia,ium,mda,mdja,iajas)
      if (ierr > 0) then
         if (ierr >= 1.and.ierr <= 6) goto 30
         return
      end if
      
      call opdfn(k,ierr,mmax,a,ia,ja,iw,imin,imax,iminw,imaxw,icg, &
            ifg,u(irst),nstcol,ncolx,time,nda,ndja,ium, &
            mda,mdja,iajas)
      if (ierr > 0) then
         if (ierr >= 1.and.ierr <= 6) goto 30
         return
      end if
      if (k > mmax) goto 100
      call trunc(k,ntr,a,ia,ja,imin,imax,time,ierr,ium)
      if (ierr > 0) return
      iw(iminw(k)) = ia(imax(k)+1)
      if (k >= mmax) goto 100
      goto 50
      30 if (k <= kfirst+1.and.iirs /= 0) return
      kerr = ierr
      ierr = 0
      kfirst = k-1
      iirs = mdiw
      isia = imin(k-1)-1
      iias = iias+isia
      isaja = ia(imin(k-1))-1
      iajas = iajas+isaja
      do i=ia(imin(k-1)),ia(imax(k-1)+1)-1
         ja(i-isaja) = ja(i)-isia
         a(i-isaja)  = a(i)
      end do
      do i=imin(k-1),imax(k-1)+1
         ia(i-isia) = ia(i)-isaja
      end do
      imin(k-1) = 1
      imax(k-1) = imax(k-1)-isia
      iw(iminw(k-1)) = ia(imax(k-1)+1)
      goto 20
      50 continue
   end do  ! k=2,levels
   goto 200
   100 levels=mmax
   200 mdia = imax(levels)+1+iias
   mdu  = max(imax(levels)+iias,mdiw+1+(mdir-1)/lratio, &
         mdiw+1+(mdiw-1)/lratio)
   mdf  = imax(levels)+iias
   mdig = mdiw+2+2*mdicg
   if (kerr /= 0.or.mdu > ndu.or.mdf > ndf) then
      write (ium,9000) mda,mdja,mdia,mdu,mdf,mdig
      write (ium,9010) mdiw
      ierr = kerr
      return
   end if
   call setifg(imin,imax,icg,ifg,nstcol,levels,time)
   return
   9000 format (/' **************** SPACE REQUIREMENTS ****************'// &
         ' VECTOR          NEEDED                              '/ &
         ' ----------------------------------------------------'/ &
         '    A ',       i16,   '   ADJUST THE DIMENSIONING OF '/ &
         '    JA',       i16,   '   VECTORS A - IG IN THE      '/ &
         '    IA',       i16,   '   CALLING PROGRAM ACCORDING  '/ &
         '    U ',       i16,   '   TO THE CALCULATED SPACE RE-'/ &
         '    F ',       i16,   '   QUIREMENTS AND RERUN THE   '/ &
         '    IG',       i16,   '   PROGRAM.                   '/ &
         ' ----------------------------------------------------')
   9010 format (/' NOTE: IF YOU WANT TO USE CG-CORRECTIONS IN THE SOLU-'/ &
         '       TION PROCESS (NCYC-SUBPARAMETER ICGR=1 OR =2),'/ &
         '       PROVIDE FOR ADDITIONAL',i6,' STORAGE LOCATIONS'/ &
         '       IN VECTORS U AND F.                           '/ &
         '         SIMILARLY, USAGE OF THE YALE-SMP SOLVER ON  '/ &
         '       THE COARSEST GRID (NSOLCO=2) WILL REQUIRE     '/ &
         '       ADDITIONAL SPACE IN VECTOR A DURING THE SOLU- '/ &
         '       TION PHASE. IN THIS CASE, HOWEVER, ITS EXACT  '/ &
         '       AMOUNT ISN''T PREDICTABLE.                    '/)
end subroutine crsng 

!.......................................................................

!     RWSRT                                             SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine rwsrt here]
subroutine rwsrt(k,ecg1,ecg2,ierr,a,ia,ja,iw,imin,imax, &
      imaxw,ifg,jtr,time,ndjtr,ium,mdjtr)
   
   !     ROW-SORT ALGORITHM FOR ROWS OF A(K). IN DETAIL:
   
   !     - SORTS THE ELEMENTS OF EACH ROW OF A(K) SUCH THAT THE STRONG
   !       CONNECTIONS JUST FOLLOW THE DIAGONAL ELEMENT. ONE OF THE STRONG-
   !       EST IS ALWAYS FIRST (EVEN IF PARTIAL SORTING IS PERFORMED!).
   !       NON-NEGATIVE CONNECTIONS ARE ALWAYS DEFINED TO BE WEAK.
   !       JA(IA(I)) IS RE-DEFINED TO POINT TO THE LAST STRONG CONNECTION
   !       OF POINT I (OR TO IA(I) IF THERE IS NO STRONG CONNECTION).
   
   !     - THE STRONG TRANSPOSE CONNECTIONS ARE LOADED LOGICALLY
   !       INTO JTR, I.E., ITS POINTERS ARE STORED IN JTR.
   
   !     ============== COMMENTS ON WORK SPACE USED =======================
   
   !     - JA(IA(I)) ----- I=IMIN(K),...,IMAX(K)
   
   !     IS DEFINED TO POINT TO THE LAST STRONG CONNECTION OF POINT I.
   !     NOTE THAT THE ORIGINAL CONTENTS OF JA(IA(I)) IS OVERWRITTEN.
   !     (IT IS PUT BACK IN SUBROUTINE WINT.)
   
   !     - JTR(J)--------- J=1,IW(IMAX(K)+IWS+1)-1
   !     - IW(I) --------- I=IMIN(K)+IWS,...,IMAX(K)+IWS
   !     - IFG(I)--------- I=IMIN(K),...,IMAX(K)
   
   !     (IWS=0, IF K=1; IWS=IMAXW(K-1)+2-IMIN(K) OTHERWISE)
   
   !     JTR IS INITIALIZED TO HAVE SAME FORM AS JA. JTR(J)
   !     CONTAINS INFORMATION ON STRONG TRANSPOSE CONNECTIONS:
   !     JTR(J) WITH IW(I+IWS)<=J<=IW(I+IWS+1)-1 POINTS TO THE STRONG
   !     TRANSPOSE CONNECTIONS OF I.
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),jtr(*),imin(*),imax(*),imaxw(*),ifg(*)
   external ctime
   save
   
   !===> INITIALIZATION OF WORK SPACE
   
   call ctime(told)
   ilo = imin(k)
   ihi = imax(k)
   if (k /= 1) then
      iws = imaxw(k-1)+2-ilo
   else
      iws = 0
   end if
   do i=ilo,ihi
      ifg(i) = 0
      ja(ia(i)) = ia(i)
   end do
   call ctime(tnew)
   time(8) = time(8)+tnew-told
   told = tnew
   
   !     *************************
   !     * PARTIAL STANDARD SORT *
   !     *************************
   
   !     NOTE: NON-NEGATIVE CONNECTIONS ARE ALWAYS DEFINED TO BE WEAK! THE
   !           STRONGEST CONNECTION IS ALWAYS FOLLOWING THE DIAGONAL.
   
   do i=ilo,ihi
      jlo = ia(i)+1
      jhi = ia(i+1)-1
      if (jhi < jlo) goto 590
      
      !===>   FIND STRONGEST CONNECTION AND SUM OF |OFF-DIAGONALS| OF ROW I
      
      amx = a(jlo)
      amn = a(jlo)
      jmx = jlo
      if (ecg1 /= 0.d0) then
         rs = 0.0d0
         do j=jlo+1,jhi
            rs = rs+abs(a(j))
            if (a(j) < amx) then
               amx = a(j)
               jmx = j
            else if (a(j) > amn) then
               amn = a(j)
            end if
         end do
         
         !===>     TEST FOR POSITIVE OFF-DIAGONALS / DIAGONAL DOMINANCE
         
         if (amx >= 0.0d0 .or. rs <= ecg1*a(ia(i))) goto 590
      else
         do j=jlo+1,jhi
            if (a(j) < amx) then
               amx = a(j)
               jmx = j
            else if (a(j) > amn) then
               amn = a(j)
            end if
         end do
         
         !===>   TEST FOR POSITIVE OFF-DIAGONALS
         
         if (amx >= 0.d0) goto 590
      end if
      
      !===>   PUT STRONGEST CONNECTION IN FIRST POSITION
      
      ast = ecg2*amx
      imx = ja(jmx)
      a(jmx)  = a(jlo)
      ja(jmx) = ja(jlo)
      a(jlo)  = amx
      ja(jlo) = imx
      if (amn <= ast) goto 580
      jhi = jhi+1
      
      !===>   DECREASE JHI UNTIL A STRONG CONNECTION IS FOUND
      !       (IF JLO >= JHI STOP: ALL CONNECTIONS ARE SORTED)
      
      560 jhi = jhi-1
      if (jlo >= jhi) goto 580
      if (a(jhi) > ast) goto 560
      
      !===>   INCREASE JLO UNTIL A WEAK CONNECTION IS FOUND
      !       (IF JLO >= JHI STOP: ALL CONNECTIONS ARE SORTED)
      
      570 jlo = jlo+1
      if (jlo >= jhi) goto 580
      if (a(jlo) <= ast) goto 570
      
      !===>   INTERCHANGE A(JHI) AND A(JLO)
      
      atmp = a(jhi)
      itmp = ja(jhi)
      a(jhi)  = a(jlo)
      ja(jhi) = ja(jlo)
      a(jlo)  = atmp
      ja(jlo) = itmp
      goto 560
      
      !===>   ROW SORTED --  SET JA(IA(I)) TO LAST STRONG CONNECTION
      
      580 ja(ia(i)) = jhi
      
      !===>   COUNT STRONG TRANSPOSE CONNECTIONS IN ROW I
      
      do j=ia(i)+1,jhi
         ifg(ja(j)) = ifg(ja(j))+1
      end do
      590 continue
   end do  ! i=ilo,ihi
   
   !===> INITIALIZATION OF WORK SPACE FOR STRONG TRANSPOSE CONNECTIONS
   
   iw(ilo+iws) = 1
   do i=ilo,ihi
      iw(i+iws+1) = iw(i+iws)+ifg(i)
      ifg(i) = iw(i+iws)
   end do
   mdjtr = iw(ihi+iws+1)-1
   if (mdjtr > ndjtr) goto 9901
   
   !===> LOAD POINTERS TO STRONG TRANSPOSE CONNECTIONS INTO JTR
   
   do i=ilo,ihi
      do j=ia(i)+1,ja(ia(i))
         ii = ja(j)
         jtr(ifg(ii)) = i
         ifg(ii) = ifg(ii)+1
      end do
   end do
   call ctime(tnew)
   time(1) = time(1)+tnew-told
   return
   
   !===> ERROR MESSAGES
   
   9901 write (ium,9910)
   ierr = 1
   return
   
   9910 format (' *** ERROR IN RWSRT: NDA TOO SMALL ***')
end subroutine rwsrt 

!.......................................................................

!     PCOL                                            SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine pcol here]
subroutine pcol(k,ierr,ia,ja,iw,imin,imax,imaxw,icg,ifg,ir,jtr, &
      time,ndicg,ndir,ium,mdicg,mdir,iias)
   
   !     PRE-COLORING ALGORITHM FOR GRID K. THIS IS THE VERSION AS
   !     DESCRIBED IN RUGE/STUEBEN (BRISTOL). THE GOAL IS TO OBTAIN QUICKLY
   !     A TENTATIVE SET OF C-POINTS WITH THE FOLLOWING PROPERTIES:
   
   !       - THE C-POINTS ARE ONLY WEAKLY CONNECTED AMONG EACH OTHER;
   !       - EACH F-POINT HAS (AT LEAST) ONE STRONG CONNECTION TO A C-POINT
   !         (EXCEPT FOR SOME EXCEPTIONAL F-POINTS, E.G., THOSE WHICH DO
   !         NOT HAVE A STRONG CONNECTION AT ALL: THE "FORCED F-POINTS").
   
   !     ON EXIT, ICG(I) (IMIN(K)<=I<=IMAX(K)) CONTAINS ALL THE INFORMATION
   !     ON THE COLORING. IN DETAIL:
   
   !        ICG(I) = 1: I IS C-POINT;
   !        ICG(I) = 0: I IS FORCED F-POINT, I.E., IT IS AN F-POINT WHICH
   !                    HAS NO STRONG CONNECTION AT ALL;
   !        ICG(I) =-1: I IS "REGULAR" F-POINT, I.E., I HAS AT LEAST ONE
   !                    STRONG CONNECTION TO A C-POINT;
   !        ICG(I) =-2: I IS "EXCEPTIONAL" F-POINT, I.E., I HAS NO STRONG
   !                    CONNECTION TO A C-POINT. FURTHERMORE, NO REGULAR
   !                    F-POINT IS STRONGLY CONNECTED TO I AND ALL POINTS
   !                    WITH ICG=-2 HAVE NO STRONG CONNECTION AMONG EACH
   !                    OTHER. (NOTE: THESE POINTS DON'T CONTRIBUTE FROM
   !                    ANY C-POINT. ALSO, IT DOES NOT MAKE SENSE TO MAKE
   !                    THESE POINTS C-POINTS AS NO F-POINT CONTRIBUTES
   !                    FROM THEM.)
   
   !     ============ COMMENTS ON INPUT ===================================
   
   !     - JA(IA(I)) ----- I=IMIN(K),...,IMAX(K)
   
   !     USED AS DEFINED IN RWSRT. NOT CHANGED.
   
   !     - JTR(J)--------- J=1,IW(IMAX(K)+IWS+1)-1
   !     - IW(I) --------- I=IMIN(K)+IWS,...,IMAX(K)+IWS
   
   !     (IWS=0, IF K=1; IWS=IMAXW(K-1)+2-IMIN(K) OTHERWISE)
   
   !     USED AS DEFINED IN RWSRT. NOT CHANGED.
   
   !     ============== COMMENTS ON WORK SPACE USED =======================
   
   !     - ICG(I) -------- I=IMIN(K),...,IMAX(K)
   
   !     USED TO DISTINGUISH F-, C- AND U- (UNDECIDED) POINTS.
   !     ON ENTRY, ICG IS DEFINED TO BE 0 FOR FORCED F-POINTS AND -2
   !     FOR U-PNTS. ON EXIT: SEE ABOVE.
   
   !     - IFG(I) -------- I=IMIN(K),...,IMAX(K)
   
   !     DEFINED TO BE A MEASURE FOR IMPORTANCE OF MAKING THE U-PNT I
   !     A C-POINT. THIS MEASURE IS A VALUE BETWEEN JVAL0 AND JVALX WITH
   
   !       JVAL0=IMAX(K)+NPTS+2,   NPTS=IMAX(K)-IMIN(K)+1;
   !       JVALX=JVAL0+4*NTRAV+1,  NTRAV=AVVERAGE NUMBER OF STRONG TRANS-
   !                                     POSE CONNECTIONS
   
   !     THE HIGHER THIS VALUE, THE MORE IMPORTANT IS IT TO MAKE U-POINT I
   !     A C-POINT. ALSO, THE VALUE JV=IFG(I) IS JUST THE "ORIGIN" OF
   !     A LIST WHICH CONNECTS ALL POINTS HAVING THE SAME MEASURE JV OF
   !     IMPORTANCE.
   
   !     - ICG(II) -------- II=IMAX(K)+2,...,JVAL0-1
   !     - IFG(II) -------- II=IMAX(K)+2,...,JVAL0-1
   
   !     LEFT AND RIGHT STACK POINTERS IN A DOUBLY LINKED LIST. THERE
   !     ARE SEVERAL SUCH LISTS, EACH OF THEM LINKING U-POINTS WHICH
   !     HAVE THE SAME MEASURE OF IMPORTANCE JV TO BE MADE C-POINTS.
   !     EACH OF THE VALUES JV MAY BE REGARDED AS THE "ORIGIN" OF SUCH
   !     A LIST DESCRIBED IN THE FOLLOWING.
   
   !     - ICG(JV) -------- JV=JVAL0,...,JVALX
   !     - IFG(JV) -------- JV=JVAL0,...,JVALX
   
   !     USED AS POINTERS INTO THE LISTS TO INDICATE ITS BEGINNING AND ITS
   !     END: IFG(JV) POINTS TO THE FIRST POINT IN THE LIST AND ICG(JV)
   !     POINTS TO THE LAST ONE. THE ROUGH PICTURE OF THE LIST WHICH
   !     CORRESPONDS TO A VALUE JVAL0 <= JV <= JVALX LOOKS AS  FOLLOWS:
   
   
   !                                  IFG
   !       ----------------------------------------------------------
   !       |                                                        |
   !       |            IFG           IFG           IFG             |
   !       ---> ------ ----> ------- ----> ------- ----> ------- ----
   !            |    |       |     |       |     |       |     |
   !            | JV |       | II1 |       | II2 |       | II3 |
   !            |    |  ICG  |     |  ICG  |     |  ICG  |     |
   !       ---- ------ <---- ------- <---- ------- <---- ------- <---
   !       |                                                        |
   !       |                          ICG                           |
   !       ----------------------------------------------------------
   
   
   !     HERE, II1, II2,... LOGICALLY REPRESENT THE "PHYSICAL" POINTS
   !     I1, I2,... (WHICH ARE VALUES BETWEEN IMIN(K) AND IMAX(K)).
   !     THE INTERCONNECTION BETWEEN THESE TWO REPRESENTATIONS OF THE
   !     SAME POINTS IS GIVEN BY THE RELATION
   
   !                          II := I+NPTS+1.
   
   !     AN EMPTY LIST IS CHARACTERIZED BY ICG(JV)=JV, IFG(JV)=JV.
   !     OBVIOUSLY, IT IS QUITE EASY TO REMOVE OR ADD POINTS TO THE LIST.
   !     SUCH RE-ARRANGEMENTS ARE NECESSARY AS (IN THE COLORING PART OF
   !     THE COARSENING ALGORITHM) UNDECIDED POINTS BECOME C- OR F-POINTS
   !     (THEY HAVE TO BE REMOVED FROM THEIR LIST) AND THE MEASURE VALUES
   !     JV OF SOME POINTS CHANGE DURING THE ALGORITHM. SUCH POINTS HAVE
   !     TO BE MOVED FROM ONE LIST TO ANOTHER ONE. IN ORDER TO KEEP TRACK
   !     OF THOSE POINTS WHICH HAVE TO BE RE-ARRANGED, A RESET-STACK IS
   !     USED WHICH CONTAINS ALL THESE POINTS I (I.E. THEIR "LOGICAL"
   !     NUMBERS II). THE POINTER IR IS USED TO POINT FROM ONE POINT IN
   !     THE STACK TO THE PREVIOUS ONE.
   
   !     - IR(J) --------- J=1,...,NPTS
   
   !     USED BELOW AS STACK-POINTER FOR POINTS TO BE RESET IN PERFORMING
   !     THE COLORING ALGORITHM. ITOP IS THE TOP-OF-STACK POINTER.
   !     THE STACK IS EMPTY IF ITOP=-1.
   
   !     THE GLOBAL PICTURE OF THE POINTERS IS SKETCHED IN THE FOLLOWING
   
   
   !                                          LIST END
   !                                       <-------------|
   !                                         LIST START  |
   !                                       <------------||
   !                                                    ||
   !                  |-- II=I+NPTS+1 ->|            IFG||ICG
   !                  |                 |               ||
   !                  |                 |               ||
   !       IMIN(K)          IMAX(K)           JVAL0             JVALX
   !          |       I        |        II      |       JV        |
   !          |=======*========|========*=======|========*========|
   !          |                |                |                 |
   !               GRID K          LOGICAL NUM      LIST ORIGINS
   !                                              (MEASURE VALUES)
   !                 ||                                  |
   !                 ||         IFG                      |
   !                 ||--------------------------------->|
   !                 |
   !                 |     |---> 1 (C)
   !                 |     |
   !                 | ICG |---> 0 (FF)
   !                 |-----|
   !                       |--->-1 (F)
   !                       |
   !                       |--->-2 (U)
   
   
   !          1                NPTS
   !          |        J        |
   !          |========*========|
   !          |                 |
   !           SHIFTED GRID K
   !                   |
   !                   | IR
   !                   |-----> RESET STACK (ITOP = TOP-OF-STACK)
   
   
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),jtr(*),ir(*),ifg(*),icg(*),imin(*), &
         imax(*),imaxw(*)
   external ctime
   save
   
   !===> PREPARATION
   
   call ctime(told)
   ilo = imin(k)
   ihi = imax(k)
   if (k /= 1) then
      iws = imaxw(k-1)+2-ilo
   else
      iws = 0
   end if
   ilo1 = ilo-1
   npts = imax(k)-imin(k)+1
   npts1 = npts+1
   ntrlim = 2*(iw(ihi+iws+1)-iw(ilo+iws))/npts
   nscnmx = 0
   jval0 = ihi+npts1+1
   jvalx = jval0+2*ntrlim+1
   mdicg = max(mdicg,jvalx+iias)
   mdir  = max(mdir,npts)
   if (jvalx > ndicg) goto 9901
   if (npts > ndir) goto 9902
   
   !===> PUT INITIAL "MEASURE" FOR EACH POINT I INTO IFG(I).
   
   do i=ilo,ihi
      nscn = iw(i+iws+1)-iw(i+iws)
      if (nscn <= ntrlim) then
         ifg(i) = jval0+nscn
      else
         ifg(i) = jvalx
      end if
   end do
   
   !===> SET CIRCULARLY LINKED LISTS AND RESET-STACK TO EMPTY
   
   itop = -1
   do j=jval0,jvalx
      icg(j) = j
      ifg(j) = j
   end do
   10 continue
   
   !===> PUT ALL U-POINTS OF GRID K INTO LISTS (I.E., NO FORCED F-POINTS).
   !     ADD POINTS ALWAYS TO THE END OF THEIR CORRESPONDING LIST.
   !     IN THE FOLLOWING, JCNBHI DENOTES THE ACTUAL HIGHEST MEASURE VALUE
   !     (AMONG U-POINTS).
   
   jcnbhi = 0
   do i=ilo,ihi
      if(ja(ia(i)) > ia(i)) goto 15
      icg(i) = 0
      goto 20
      15 icg(i) = -2
      ii = i+npts1
      ir(i-ilo1) = 0
      jv = ifg(i)
      if (jv > jcnbhi) jcnbhi = jv
      icg(ii) = icg(jv)
      ifg(ii) = jv
      icg(jv) = ii
      ifg(icg(ii)) = ii
      20 continue
   end do
   
   !     ****************
   !     * PRE-COLORING *
   !     ****************
   
   !     PICK A U-POINT IC WITH MAXIMAL MEASURE AS GIVEN BY JCNBHI.
   !     (TAKE THE FIRST FROM CORRESPONDING LIST: FIRST-IN/FIRST-OUT)
   !     MAKE THAT POINT A C-POINT AND REMOVE IT FROM LISTS. THEN MAKE
   !     ALL STRONG TRANSPOSE U-CONNECTIONS F-POINTS AND REMOVE THEM FROM
   !     THEIR LISTS. UPDATE THE MEASURE OF IMPORTANCE FOR U-POINTS TO
   !     BECOME C-POINTS AND ADD THESE POINTS TO THE RESET-STACK. FINALLY,
   !     RE-ARRANGE THE LISTS BY SWEEPING THROUGH THE RESET STACK. THEN
   !     PICK ANOTHER U-POINT IC.
   
   !     IF LIST CORRESPONDING TO JCNBHI IS EMPTY, GO TO NEXT LOWER VALUE.
   !     PRE-COLOURING IS FINISHED IF JCNBHI<=JVAL0. ALL U-POINTS LEFT AT
   !     THAT TIME, WILL BE REGARDED AS F-POINTS LATER.
   
   30 if(jcnbhi <= jval0) goto 100
   iic = ifg(jcnbhi)
   if (iic /= jcnbhi) goto 40
   jcnbhi = jcnbhi-1
   goto 30
   
   !===> CREATE C-POINT
   
   40 ic = iic-npts1
   icg(ic) = 1
   icg(ifg(iic)) = icg(iic)
   ifg(icg(iic)) = ifg(iic)
   
   !===> FOR POINT IC WITH ECCESSIVE NUMBER OF STRONG TRANSPOSE
   !     CONNECTIONS: MAKE IT A C-POINT BUT LET ITS CONNECTED POINTS
   !     REMAIN UNDECIDED.
   
   if (jcnbhi == jvalx) goto 78
   
   !===> CREATE F-POINTS AROUND ABOVE C-POINT
   
   do j=iw(ic+iws),iw(ic+iws+1)-1
      i = jtr(j)
      if (icg(i) /= -2) goto 77
      icg(i) = -1
      ii = i+npts1
      icg(ifg(ii)) = icg(ii)
      ifg(icg(ii)) = ifg(ii)
      
      !===>   INCREMENT MEASURE FOR ALL STRONG U-CONNECTIONS III OF I
      !       (IF NOT YET MEASURE=JVALX) AND PUT THEM ON RESET STACK (IF NOT
      !       YET THERE)
      
      do jj=ia(i)+1,ja(ia(i))
         iii = ja(jj)
         if (icg(iii) /= -2.or.ifg(iii) >= jvalx) goto 76
         ifg(iii) = ifg(iii)+1
         iiii = iii-ilo1
         if (ir(iiii) /= 0) goto 76
         ir(iiii) = itop
         itop = iiii
         76 continue
      end do
      77 continue
   end do
   
   !===> DECREMENT MEASURE FOR ALL STRONG U-CONNECTIONS I OF IC
   !     AND PUT THEM ON RESET-STACK (IF NOT YET THERE)
   
   78 do j=ia(ic)+1,ja(ia(ic))
      i = ja(j)
      if (icg(i) /= -2) goto 87
      ifg(i) = ifg(i)-1
      ii = i-ilo1
      if (ir(ii) /= 0) goto 87
      ir(ii) = itop
      itop = ii
      87 continue
   end do
   
   !===> REARRANGE THE LISTS BY SWEEPING THROUGH RESET-STACK.
   !     THEN GO BACK TO PICK ANOTHER U-POINT IC.
   
   in = itop
   itop = -1
   
   90 if (in <= 0) goto 30
   i = in+ilo1
   ii = i+npts1
   if (icg(i) /= -2) goto 95
   ifg(icg(ii)) = ifg(ii)
   icg(ifg(ii)) = icg(ii)
   jv = ifg(i)
   if (jv > jcnbhi) jcnbhi = jv
   icg(ii) = icg(jv)
   ifg(ii) = jv
   icg(jv) = ii
   ifg(icg(ii)) = ii
   95 iip = in
   in = ir(in)
   ir(iip) = 0
   goto 90
   
   100 call ctime(tnew)
   time(2) = time(2)+tnew-told
   return
   
   !===> ERROR MESSAGES
   
   9901 write (ium,9910)
   ierr = 6
   return
   
   9902 write (ium,9920)
   ierr = 4
   return
   
   9910 format (' *** ERROR IN PCOL: NDIG TOO SMALL ***')
   9920 format (' *** ERROR IN PCOL: NDU TOO SMALL ***')
end subroutine pcol 

!.......................................................................

!     PWINT                                             SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine pwint here]
subroutine pwint (k,ewt2,ichk,mmax,a,ia,ja,iw,imin,imax,iminw, &
      imaxw,ifg,icg,ncolor,time,ierr,irow0,ncolx, &
      nda,ndja,ndicg,ndia,ium,mda,mdja,iajas)
   
   !     SET UP FINAL COARSER GRID K+1 AND INTERPOLATION FORMULA FROM
   !     GRID K+1 TO GRID K. THIS IS THE VERSION AS DESCRIBED IN RUGE/
   !     STUEBEN (BRISTOL). PWINT ASSUMES THE GRID TO BE PRE-COLORED
   !     BY SUBROUTINE PCOL.
   
   !     ON EXIT, A, JA AND IW ARE SET TO CONTAIN THE INTERPOLATION
   !     WEIGHTS AND CORRESPONDING POINTERS AS REQUIRED IN THE SOLUTION
   !     PHASE OF AMG1R5. ALSO, ICG(I) (IMIN(K)<=I<=IMAX(K)) ARE SET TO
   !     THEIR FINAL VALUES, EXCEPT FOR THOSE I WITH ICG(I)<0.
   
   !     ============ COMMENTS ON INPUT ===================================
   
   !     - JA(IA(I)) ----- I=IMIN(K),...,IMAX(K)
   
   !     ASSUMED TO POINT TO THE LAST STRONG CONNECTION OF POINT I (OR TO
   !     IA(I) IF THERE IS NO SUCH CONNECTION). ON EXIT, RESET TO ORIGINAL
   !     VALUES (I.E. JA(IA(I))=I).
   
   !     - ICG(I) -------- I=IMIN(K),...,IMAX(K)
   
   !     ASSUMED TO CONTAIN INFORMATION ON PRE-COLORING:
   
   !       ICG(I)>0: I IS C-POINT
   !       ICG(I)=0: I IS FORCED F-POINT (I.E. I HAS NO STRONG CONNNECTION)
   !       ICG(I)<0: I IS F-POINT WITH (AT LEAST) ONE STRONG CONNECTION.
   
   !     ============== COMMENTS ON WORK SPACE USED =======================
   
   !     - IFG(I) -------- I=IMIN(K),...,IMAX(K)
   
   !     IS USED FOR SEVERAL PURPOSES. IN PARTICULAR, TO DISTINGUISH
   !     INTERPOLATORY AND NON-INTERPOLATORY POINTS.
   
   !     - NCOLOR(I) ----- I=1,...,#POINTS ON GRID K
   
   !     IS SET TO F-POINT-COLORS TO BE USED LATER IN SUBROUTINE OPDFN.
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),imaxw(*), &
         ifg(*),icg(*),ncolor(*)
   external ctime
   save
   
   call ctime(told)
   ndaja = min(nda,ndja)
   ncount = 0
   if (k == 1) then
      iminw(1) = 1
      iw(1) = ia(imax(1)+1)
   end if
   
   ilo = imin(k)
   ihi = imax(k)
   do i=ilo,ihi
      ifg(i) = 0
   end do
   
   !===> SWEEP OVER F-POINTS I WHICH HAVE AT LEAST ONE STRONG CONNECTION
   
   iblck = iminw(k)
   do i=ilo,ihi
      if (icg(i) >= 0) goto 400
      jlo = ia(i)+1
      jhi = ja(ia(i))
      ewt2i = ewt2/a(jlo)
      ncondc = 0
      
      !===>   INITIALIZE "BLOCK" OF INTERPOLATION WEIGHTS OF POINT I
      
      30 jw0 = iw(iblck)
      jwx = jw0
      if (jwx > ndaja) goto 2000
      do j=jlo,jhi
         ii = ja(j)
         if (icg(ii) <= 0) goto 20
         a(jwx)  = a(j)
         ja(jwx) = ii
         ifg(ii) = jwx
         jwx = jwx+1
         if (jwx > ndaja) goto 2000
      20 continue
      end do
      a(jwx)  = a(ia(i))
      ja(jwx) = i
      mda  = max(mda,jwx+iajas)
      mdja = max(mdja,jwx+iajas)
      ifg(i) = jwx
      
      !===>   SWEEP OVER STRONGLY CONNECTED F-POINTS. THESE MUST BE "COVERED"
      !       BY A TOTAL WEIGHT DEFINED BY EWT2. IF AN F-POINT HAS NO STRONG
      !       CONNECTIONS, REGARD IT TO BE COVERED, BUT DO NOT DISTRIBUTE
      !       THE CORRESPONDING WEIGHT (ERROR AT SUCH A POINT CAN BE ASSUMED
      !       TO BE VERY SMALL!).
      
      do j=jlo,jhi
         ii = ja(j)
         if (icg(ii) >= 0) goto 150
         
         !===>     COMPUTE DEPENDENCE ON SET OF INTERPOLATION POINTS
         
         s  = 0.0d0
         si = 0.0d0
         jjlo = ia(ii)+1
         jjhi = ia(ii+1)-1
         do jj=jjlo,jjhi
            if (ifg(ja(jj)) < jw0) goto 110
            if (ja(jj) == i) si = a(jj)
            s = s+a(jj)
         110 continue
         end do
         if (ichk == 2) goto 111
         if (s == 0.0d0) then
            a(jwx) = a(jwx)+a(j)
            goto 150
         else
            goto 135
         end if
         
         !===>     CHECK DEPENDENCE ON SET OF INTERPOLATION POINTS
         
         111 if (s-si <= ewt2i*a(j)*a(jjlo)) goto 135
         
         !===>     DEPENDENCE TOO SMALL: IF THERE IS NOT YET A CONDITIONAL
         !         C-POINT, MAKE II SUCH A POINT AND RESTART THE
         !         PROCESS FOR DEFINING INTERPOLATION WEIGHTS FOR POINT I.
         !         OTHERWISE MAKE I ITSELF A C-POINT AND LEAVE II AN F-POINT.
         
         if (ncondc == 0) then
            ncount = ncount+1
            ncondc = 1
            ip = ii
            icgp = icg(ii)
            icg(ii) = 1
            goto 30
         else
            icg(i) = 1
            icg(ip) = icgp
            do jj=jw0,jwx
               ifg(ja(jj)) = 0
            end do
            goto 400
         end if
         
         !===>     DISTRIBUTE THE WEIGHT OF POINT II
         
         135 ww = a(j)/s
         do jj=jjlo,jjhi
            if (ifg(ja(jj)) >= jw0) &
                  a(ifg(ja(jj))) = a(ifg(ja(jj)))+a(jj)*ww
         end do
      150 continue
      end do  ! j=jlo,jhi
      
      !===>   ALL NECESSARY POINTS ARE COVERED. NOW DISTRIBUTE WEIGHTS FROM
      !       WEAK CONNECTIONS OF POINT I (ANALOGOUS AS ABOVE)
      
      do j=jhi+1,ia(i+1)-1
         ii = ja(j)
         if (icg(ii) == 0) goto 190
         s = 0.0d0
         jjlo = ia(ii)+1
         jjhi = ia(ii+1)-1
         do jj=jjlo,jjhi
            if (ifg(ja(jj)) >= jw0) s = s+a(jj)
         end do
         if (s == 0.0d0) then
            a(jwx) = a(jwx)+a(j)
         else
            ww = a(j)/s
            do jj=jjlo,jjhi
               if (ifg(ja(jj)) >= jw0) &
                     a(ifg(ja(jj))) = a(ifg(ja(jj)))+a(jj)*ww
            end do
         end if
      190 continue
      end do
      
      icg(i) = -iblck
      iblck = iblck+1
      iw(iblck) = jwx+1
   400 continue
   end do  ! i=ilo,ihi
   
   !===> SET ICG; RESET JA(IA(I)); CHECK SIZE OF COARSEST GRID
   
   ic = ihi
   imin(k+1) = ic+1
   do i=ilo,ihi
      ja(ia(i)) = i
      if (icg(i) <= 0) goto 900
      ic = ic+1
      icg(i) = ic
      if (ic < ndicg) icg(ic) = 0
   900 continue
   end do
   imax(k+1) = ic
   
   npts = ihi-ilo+1
   nptsc = imax(k+1)-imin(k+1)+1
   if (nptsc == 1)                  mmax = k+1
   if (nptsc == 1.and.irow0 < 2)   mmax = k
   if (nptsc == npts.or.nptsc == 0) mmax = k
   if (k >= mmax) goto 1000
   if (ic >= ndia) goto 1700
   if (ic >= ndicg) goto 1800
   
   !===> RE-ARRANGE A
   
   1000 iblck1 = iminw(k)
   jwpos = iw(iblck1)
   do i=ilo,ihi
      if (icg(i) >= 0) goto 950
      iblck = -icg(i)
      j1 = iw(iblck)
      j2 = iw(iblck+1)-1
      if (j2 <= j1) then
         icg(i) = 0
      else
         icg(i) = -iblck1
         iw(iblck1) = jwpos
         scale = -1.0d0/a(j2)
         do j=j1,j2-1
            a(jwpos)  = a(j)*scale
            ja(jwpos) = icg(ja(j))
            jwpos = jwpos+1
         end do
         iblck1 = iblck1+1
      end if
   950 continue
   end do
   imaxw(k) = iblck1-1
   iw(iblck1) = jwpos
   
   !===> STORE TYPE OF POINTS (I.E. C, F OR FF)  ON VECTOR NCOLOR
   
   is = 1-ilo
   do i=ilo,ihi
      nc = icg(i)
      if (nc < 0) then
         ncolor(i+is) = 1
      else if (nc > 0) then
         ncolor(i+is) = 2
      else
         ncolor(i+is) = 3
      end if
   end do
   ncolx = 1
   
   !===> EXIT
   
!  write (ium,9000) k,ncount
   call ctime(tnew)
   time(4) = time(4)+tnew-told
   return
   
   !===> ERROR MESSAGES
   
   1700 write (ium,9030)
   ierr = 2
   return
   
   1800 write (ium,9040)
   ierr = 6
   return
   
   2000 if (nda <= ndja) then
      write (ium,9020)
      ierr = 1
   else
      write (ium,9050)
      ierr = 3
   end if
   return
   
   9000 format (' INTERPOLATION OPERATOR NO.',i3,' COMPLETED. C-POINTS', &
         ' ADDED IN PWINT:',i4)
   9020 format (' *** ERROR IN PWINT: NDA TOO SMALL ***')
   9030 format (' *** ERROR IN PWINT: NDIA TOO SMALL ***')
   9040 format (' *** ERROR IN PWINT: NDIG TOO SMALL ***')
   9050 format (' *** ERROR IN PWINT: NDJA TOO SMALL ***')
end subroutine pwint 

!.......................................................................

!     OPDFN                                             SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine opdfn here]
subroutine opdfn(k,ierr,mmax,a,ia,ja,iw,imin,imax,iminw,imaxw, &
      icg,ifg,ncolor,nstcol,ncolx,time,nda,ndja, &
      ium,mda,mdja,iajas)
   
   !     THIS SUBROUTINE CONSTRUCTS THE CG-OPERATOR A(K) (K>1).
   
   !     ONE ROW IS CONSTRUCTED AT A TIME, AND THE ROW STRUCTURE
   !     (I.E., WHICH CONNECTION GOES WHERE) MUST BE DETERMINED
   !     AT THE SAME TIME.  IN ORDER TO AVOID SEARCHES THROUGH
   !     THE CURRENT ROW TO DETERMINE IF A POSITION FOR A CONNECTION
   !     HAS ALREADY BEEN DEFINED, ICG (FOR LEVEL K) AND IFG (FOR LEVEL
   !     K-1) ARE USED AUXILIARILY. FURTHERMORE, TO SPEED UP COMPUTATION,
   !     THE TRANSPOSE OF INTERPOLATION IS TEMPORARILY STORED ON A (AT
   !     THE END OF THE AVAILABLE SPACE). IFG (FOR LEVEL K) IS USED
   !     AS POINTER TO THE CORRESPONDING ROWS. MORE DETAILS: SEE BELOW.
   
   !     ============ COMMENTS ON INPUT ===================================
   
   !     - IMIN(K), IMAX(K)
   
   !     FIRST/LAST NUMBER OF GRID POINTS ON GRID K.
   
   !     - IMINW(K-1), IMAXW(K-1)
   
   !     FIRST/LAST "BLOCK" NUMBER OF INTERPOLATION WEIGHTS CONTRIBUTING
   !     IN INTERPOLATION TO GRID K-1.
   
   !     - IW(IBL) ----- IBL=IMINW(K-1),IMAXW(K-1)+1
   !     - A(J) -------- J=JLO,JHI      (JLO=IW(IMINW(K-1)))
   !     - JA(J) ------- J=JLO,JHI      (JHI=IW(IMAXW(K-1)+1)-1)
   
   !     ARE ASSUMED TO CONTAIN THE WEIGHTS OF INTERPOLATION ALONG WITH
   !     THE NECESSARY POINTERS. IW(IBL) POINTS TO THE FIRST ENTRY OF
   !     "BLOCK" NUMBER IBL IN A (CF. BELOW).
   
   !     - ICG(IF) ----- IF=IMIN(K-1),IMAX(K-1)
   
   !     ARE ASSUMED TO BE SET TO THEIR FINAL VALUES:
   
   !       ICG(IF)>0: IF IS C-POINT OF GRID K-1 AND ICG(IF) POINTS JUST
   !                  TO THE CORRESPONDING POINT ON GRID K;
   !       ICG(IF)=0: IF IS F-POINT OF GRID K-1 WITHOUT ANY CONTRIBUTION
   !                  IN INTERPOLATION FROM GRID K;
   !       ICG(IF)<0: IF IS F-POINT OF GRID K-1; IBL=-ICG(IF) POINTS JUST
   !                  TO THE "BLOCK" OF INTERPOLATION WEIGHTS. THAT MEANS,
   !                  JW(J) (IW(IBL)<=J<=IW(IBL+1)-1) POINTS TO THE POINTS
   !                  (ON GRID K) WHICH CONTRIBUTE IN INTERPOLATION TO IF.
   !                  THE CORRESPONDING WEIGHTS ARE STORED IN A(J).
   
   !     - IFG(IF) ----- IF=IMIN(K-1),IMAX(K-1)
   
   !     WORK SPACE (SEE BELOW). IFG(IF) IS ASSUMED TO BE > -IMIN(K).
   
   !     - IFG(IC) ----- IC=IMIN(K),IMAX(K)+1
   
   !     WORK SPACE (SEE BELOW). THE CONTENTS OF IFG(IC) IS ARBITRARY.
   
   !     - ICG(IC) ----- IC=IMIN(K),IMAX(K)
   
   !     WORK SPACE (SEE BELOW). ICG(IC) IS ASSUMED TO BE ZERO.
   
   !     ============== COMMENTS ON WORK SPACE USED =======================
   
   !     - A(J) -------- J=JJLO,JJHI
   !     - JA(J) ------- J=JJLO,JJHI
   
   !     (JJLO=MIN(NDA,NDJA)-IW(IMAXW(K-1)+1)+IW(IMINW(K-1)+1),
   !      JJHI=MIN(NDA,NDJA), I.E. THE SPACE OF LENGTH OF INTERPOLATION
   !      (K-1) AT THE END OF A AND JA, RESPECTIVELY.)
   
   !     IS USED TO STORE THE TRANSPOSE OF INTERPOLATION W(K-1). THIS IS
   !     DONE IN ORDER TO SPEED UP THE COMPUTATION OF THE OPERATOR A(K).
   !     IF DURING ASSEMBLAGE OF A(K) THIS WORK SPACE IS REQUIRED BY A(K),
   !     PROCESSING CONTINUES USING THE NOT YET REDEFINED PART OF THE WORK
   !     SPACE AND RECALCULATING THE LOST ENTRIES EACH TIME THEY ARE
   !     NEEDED. THUS THE CALCULATION SLOWS DOWN.
   
   !     - IFG(IC) ----- IC=IMIN(K),IMAX(K)+1
   
   !     IS USED AS POINTER FOR THE TRANSPOSE OF INTERPOLATION: THE COARSE-
   !     GRID POINT IC CONTRIBUTES IN INTERPOLATION TO THE FINE-GRID POINTS
   !     IF=JA(J) (IFG(IC)<=J<=IFG(IC+1)-1) WITH WEIGHT A(J). THE
   !     CONTRIBUTION TO ITSELF (BY THE WEIGHT 1.0) IS NOT CONTAINED.
   
   !     - ICG(IC) ----- IC=IMIN(K),IMAX(K)
   
   !     IS USED FOR SEVERAL PURPOSES. IN ASSEMBLING THE CG-OPERATOR A(K),
   !     IT SERVES AS POINTER TO POSITIONS IN A(K) WHICH HAVE ALREADY BEEN
   !     DEFINED: IF THE CURRENT ROW CORRESPONDS TO POINT IC1, AND A
   !     CONNECTION TO ANOTHER CG-POINT IC2 HAS JUST BEEN FOUND, THEN IF
   !     ICG(IC2)<IA(IC1), THE CORRESPONDING ENTRY IN ROW IC1 OF A(K)
   !     HAS NOT YET BEEN DEFINED. OTHERWISE, ICG(IC2) POINTS TO THE
   !     LOCATION FOR THAT ENTRY. (ALSO SEE IFG BELOW.)
   
   !     - IFG(IF) ----- IF=IMIN(K-1),IMAX(K-1)
   
   !     IN ASSEMBLING THE CG-OPERATOR A(K), THIS VECTOR CONTAINS INFORMAT-
   !     ION ON WHETHER THE EXISTENCE OF ENTRIES IN A(K) HAS TO BE CHECKED
   !     OR NOT.
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),imaxw(*), &
         icg(*),ifg(*),ncolor(*),nstcol(*)
   external ctime
   save
   
   !===> EXTEND A, JA TO STORE TRANSPOSE OF INTERPOLATION
   
   call ctime(told)
   ndaja = min(nda,ndja)
   iminw(k) = imaxw(k-1)+1
   if (k > mmax) goto 8000
   jpos = iw(iminw(k))
   do j=iw(iminw(k-1)),iw(imaxw(k-1)+1)-1
      icg(ja(j)) = icg(ja(j))+1
   end do
   
   mda  = max(mda,jpos+jpos-iw(iminw(k-1))+iajas)
   mdja = max(mdja,jpos+jpos-iw(iminw(k-1))+iajas)
   ifg(imin(k)) = ndaja-jpos+iw(iminw(k-1))+1
   if (ifg(imin(k)) <= jpos) goto 9900
   do ic=imin(k),imax(k)
      ifg(ic+1) = ifg(ic)+icg(ic)
      icg(ic) = ifg(ic)
   end do
   
   do if=imin(k-1),imax(k-1)
      if (icg(if) >= 0) goto 250
      ibl = -icg(if)
      do j=iw(ibl),iw(ibl+1)-1
         ic = ja(j)
         a(icg(ic))=a(j)
         ja(icg(ic))=if
         icg(ic)=icg(ic)+1
      end do
   250 continue
   end do
   
   do ic=imin(k),imax(k)
      icg(ic) = 0
   end do
   
   !===> SWEEP OVER ALL CG-POINTS IC TO ASSEMBLE ROWS OF CG MATRIX
   
   istti = 1
   do if=imin(k-1),imax(k-1)
      if (icg(if) <= 0) goto 100
      ic = icg(if)
      if (jpos > ndaja) goto 9900
      ialow = jpos
      icg(ic) = jpos
      a(jpos)  = a(ia(if))
      ja(jpos) = ic
      jpos = jpos+1
      
      !       ----------------------------------------------
      !       | SEARCH FOR C-C-C-C AND C-C-F-C CONNECTIONS |
      !       ----------------------------------------------
      
      do jf1=ia(if)+1,ia(if+1)-1
         if1 = ja(jf1)
         ic1 = icg(if1)
         if (ic1) 11,25,20
         
         !===>     IF1 IS F-POINT: SWEEP OVER C-C-F-C CONNECTIONS
         
         11 ifg(if1) = -ic
         do jf2=iw(-ic1),iw(-ic1+1)-1
            ic2 = ja(jf2)
            if (icg(ic2) >= ialow) goto 10
            if (jpos > ndaja) goto 9900
            icg(ic2) = jpos
            a(jpos)  = a(jf1)*a(jf2)
            ja(jpos) = ic2
            jpos = jpos+1
            goto 15
            10 a(icg(ic2)) = a(icg(ic2))+a(jf1)*a(jf2)
         15 continue
         end do
         goto 25
         
         !===>     IF1 IS C-POINT: C-C-C-C CONNECTION
         
         20 if (icg(ic1) >= ialow) goto 23
         if (jpos > ndaja) goto 9900
         icg(ic1) = jpos
         a(jpos)  = a(jf1)
         ja(jpos) = ic1
         jpos = jpos+1
         goto 25
         23 a(icg(ic1)) = a(icg(ic1))+a(jf1)
      25 continue
      end do  ! jf1=ia(if)+1,ia(if+1)-1
      ist = imin(k-1)-1
      
      !        ----------------------------------------------
      !        | SEARCH FOR C-F-C-C AND C-F-F-C CONNECTIONS |
      !        ----------------------------------------------
      
      do jf1=ifg(ic),ifg(ic+1)-1
         if (jf1 >= jpos) then
            if1 = ja(jf1)
            wjf1 = a(jf1)
            ist = if1
         else
            istti = 0
            do jf2=ist+1,imax(k-1)
               if (icg(jf2) >= 0) goto 120
               ibl = -icg(jf2)
               do jb=iw(ibl),iw(ibl+1)-1
                  jc3 = ja(jb)
                  if(jc3 /= ic) goto 110
                  if1 = jf2
                  wjf1 = a(jb)
                  goto 130
               110 continue
               end do
            120 continue
            end do
            
            !===>       ERROR EXIT
            
            write (ium,9940) k-1
            ierr = 22
            return
            130 ist = if1
         end if
         do jf2=ia(if1),ia(if1+1)-1
            if2 = ja(jf2)
            ic2 = icg(if2)
            ww = wjf1*a(jf2)
            if (ic2) 35,90,50
            
            !===>       IF2 IS F-POINT: SWEEP OVER C-F-F-C CONNECTIONS
            
            35 if (ifg(if2) == -ic) goto 70
            ifg(if2) = -ic
            do jf3=iw(-ic2),iw(-ic2+1)-1
               ic3 = ja(jf3)
               if (icg(ic3) >= ialow) goto 30
               if (jpos > ndaja) goto 9900
               icg(ic3) = jpos
               a(jpos) = ww*a(jf3)
               ja(jpos) = ic3
               jpos = jpos+1
               goto 40
               30 a(icg(ic3)) = a(icg(ic3))+ww*a(jf3)
            40 continue
            end do
            goto 90
            
            !===>       IF2 HAS BEEN ENCOUNTERED BEFORE; DO NOT CHECK POSITIONS!
            
            70 do jf3=iw(-ic2),iw(-ic2+1)-1
               iadrs = icg(ja(jf3))
               a(iadrs) = a(iadrs)+ww*a(jf3)
            end do
            goto 90
            
            !===>       IF2 IS C-POINT: C-F-C-C CONNECTION
            
            50 if (icg(ic2) >= ialow) goto 60
            if (jpos > ndaja) goto 9900
            icg(ic2) = jpos
            a(jpos) = ww
            ja(jpos) = ic2
            jpos = jpos+1
            goto 90
            60 a(icg(ic2)) = a(icg(ic2))+ww
         90 continue
         end do  ! jf2=ia(if1),ia(if1+1)-1
      end do  ! jf1=ifg(ic),ifg(ic+1)-1
      ia(ic+1) = jpos
   100 continue
   end do  ! if=imin(k-1),imax(k-1)
   mda  = max(mda,ia(imax(k)+1)-1+iajas)
   mdja = max(mdja,ia(imax(k)+1)-1+iajas)
   
   !===> WARNING FOR STORAGE SHORTAGE
   
   if (istti /= 1) then
      k1 = k-1
      write (ium,9930) k1
      ierr = -1
   end if
!  write (ium,9000) k
   
   !===> SET UP LINKED LIST FOR RELAXATION ON GRID K-1
   
   8000 ia(imin(k)) = iw(iminw(k))
   ilo = imin(k-1)
   ihi = imax(k-1)
   ilo1 = ilo-1
   npts = ihi-ilo1
   ist = 100000000
   do icol=1,ncolx
      do i=npts,1,-1
         if (ncolor(i) /= icol) goto 8040
         icg(i+ilo1) = -ist
         ist=i+ilo1
      8040 continue
      end do
   end do
   nstcol(k-1)=ist
   
   !===> EXIT / ERROR MESSAGES
   
   call ctime(tnew)
   time(6) = time(6)+tnew-told
   return
   
   9900 if (nda <= ndja) then
      write (ium,9910)
      ierr = 1
   else
      write (ium,9920)
      ierr = 3
   end if
   return
   
   9000 format (' COARSE  GRID  OPERATOR NO.',i3,' COMPLETED')
   9910 format (' *** ERROR IN OPDFN: NDA TOO SMALL ***')
   9920 format (' *** ERROR IN OPDFN: NDJA TOO SMALL ***')
   9930 format (' --- WARNG: UNABLE TO STORE TRANSPOSE OF INTERPOLATION', &
         ' ON GRID ',i2,' DURING '/ &
         '            EXECUTION OF OPDFN, BECAUSE NDA OR NDJA ', &
         'TOO SMALL.'/ &
         '            SETUP COMPUTATION IS SLOWING DOWN.')
   9940 format (' *** ERROR IN OPDFN: INTERPOLATION ENTRY MISSING ON GRID' &
         ,i3)
end subroutine opdfn 

!.......................................................................

!     SETIFG                                              SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine setifg here]
subroutine setifg(imin,imax,icg,ifg,nstcol,levels,time)
   
   !     SET "INVERSE" POINTER IFG
   
   _REAL_ told,tnew,time(*)
   integer imin(*),imax(*),icg(*),ifg(*),nstcol(*)
   external ctime
   save
   
   call ctime(told)
   do i=imin(1),imax(levels-1)
      if (icg(i) > 0) ifg(icg(i))=i
   end do
   ib = 1
   do k=1,levels-1
      ist = nstcol(k)
      20 if (ist >= 100000000) goto 30
      ifg(ib) = ist
      ist = -icg(ist)
      ib = ib+1
      goto 20
   30 continue
   end do
   call ctime(tnew)
   time(8) = time(8)+tnew-told
   return
end subroutine setifg 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     AMG1R5 SOLUTION-SUBROUTINES

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!.......................................................................

!     SOLVE                                                SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine solve here]
subroutine solve(madapt,ncyc,nrd,nsolco,nru,iout,ierr,a,u,f,ia, &
      ja,iw,eps,imin,imax,iminw,imaxw,icg,ifg,nstcol, &
      iarr,time,ncyc0,irow0,levels,nda,ndja, &
      ndu,ndf,mda,mdja,mdu,mdf,iup,ium,resi,res0,res)
   
   !     SOLUTION PHASE OF AMG1R5
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*),resi(*)
   _REAL_ time(*)
   integer nrdtyp(10),nrutyp(10),ia(*),ja(*),iw(*),imin(*),imax(*), &
         iminw(*),imaxw(*),icg(*),ifg(*),nstcol(*),iarr(*)
   external cg,cyc,idec,resid,usave
   save
   
   !===> TEST OF AVAILABLE STORAGE
   
   if (ndu < imax(levels)) then
      write (ium,9050)
      ierr = 4
      return
   end if
   if (ndf < imax(levels)) then
      write (ium,9060)
      ierr = 5
      return
   end if
   
   m = levels
   ncyc0 = 0
   do n=11,20
      time(n) = 0.0
   end do
   if (eps /= 0.d0) then
      epsi = eps
   else
      epsi = 1.d-12
   end if
   
   !===> DECOMPOSE MADAPT
   
   if (madapt /= 0) then
      call idec(madapt,2,ndigit,iarr)
      msel = iarr(1)
      if (msel == 2) then
         if (iarr(2) /= 0) then
            fac = dble(iarr(2))
            do i=1,100
               fac = fac/10.d0
               if (fac <= 1.d0) goto 9
            end do
         else
            fac = 0.7d0
         end if
      end if
   else
      msel = 2
      fac = 0.7d0
   end if
   
   !===> DECOMPOSE NCYC
   
   9 if (ncyc /= 0) then
      call idec(abs(ncyc),4,ndigit,iarr)
      igam   = sign(iarr(1),ncyc)
      icgr   = iarr(2)
      iconv  = iarr(3)
      ncycle = iarr(4)
      if (ncycle == 0) return
   else
      igam   = 1
      icgr   = 0
      iconv  = 1
      ncycle = 10
   end if
   
   !===> SET EPSI ACCORDING TO CONVERGENCE CRITERION GIVEN BY ICONV
   
   if (iconv /= 3) then
      if (iconv == 4) then
         ama = 0.d0
         do i=1,imax(1)
            ama = max(ama,a(ia(i)))
         end do
         epsi = epsi*ama
      end if
   else
      fmax = 0.d0
      do i=1,imax(1)
         fmax = max(fmax,abs(f(i)))
      end do
      epsi = epsi*fmax
   end if
   
   !===> DECOMPOSE NRD
   
   if (nrd /= 0) then
      call idec(nrd,9,ndigit,nrdtyp)
      nrdx   = nrdtyp(2)
      nrdlen = ndigit-2
      do i=1,nrdlen
         nrdtyp(i) = nrdtyp(i+2)
      end do
      10 continue
   else
      nrdx      = 1
      nrdlen    = 2
      nrdtyp(1) = 3
      nrdtyp(2) = 1
   end if
   
   !===> DECOMPOSE NRU
   
   if (nru /= 0) then
      call idec(nru,9,ndigit,nrutyp)
      nrux   = nrutyp(2)
      nrulen = ndigit-2
      do i=1,nrulen
         nrutyp(i) = nrutyp(i+2)
      end do
      40 continue
   else
      nrux      = 1
      nrulen    = 2
      nrutyp(1) = 3
      nrutyp(2) = 1
   end if
   
   !===> DECOMPOSE NSOLCO
   
   if (nsolco /= 0) then
      call idec(nsolco,2,ndigit,iarr)
      nsc  = iarr(1)
      nrcx = iarr(2)
      
      !===> IN CASE OF YALE-SMP COARSE GRID SOLUTION, DON'T USE COARSEST
      !     GRID WITH LESS THAN 10 POINTS
      
      if (nsc == 2) then
         do i=m,1,-1
            l = i
            if (imax(i)-imin(i) >= 9) goto 60
         end do
         50 continue
         60 m = i
         levels = i
      end if
   else
      nsc  = 1
      nrcx = 0
   end if
   
   !===> CYCLING
   
   100 if (iout /= 0) then
      call resid(1,res0,a,u,f,ia,ja,iw,imin,imax,iminw)
      if (iout == 3) then
         write (iup,9005)
         write (iup,9000) res0
      end if
      resold = res0
   end if
   
   do iter=1,ncycle
      call usave(1,icgr,u,imin,imax,ndu,m,time,ierr,ium,mdu,ndf,mdf)
      call cyc(1,nrdx,nrdtyp,nrdlen,nrcx,nrux,nrutyp,nrulen, &
            igam,a,u,f,ia,ja,iw,imin,imax,iminw,imaxw, &
            ifg,icg,nstcol,iarr,time,irow0,m,ium,ierr,iter, &
            nsc,nda,ndja,mda,mdja,msel,fac,resi,levels)
      call cg(1,icgr,iter,a,u,f,ia,ja,iw,imin,imax,iminw,m,time, &
            ierr,ium)
      if (ierr > 0) return
      if (iter == 1) then
         mfirst = m
         if (iout == 3) write (iup,9040) m
      else if (iout == 3.and.m /= mfirst) then
         mfirst = m
         write (iup,9040) m
      end if
      if (iout == 3.or.iconv /= 1) call resid(1,res,a,u,f,ia,ja,iw, &
            imin,imax,iminw)
      ncyc0 = iter
      if (iout /= 3) goto 110
      cfac = res/(resold+1.0d-40)
      resold = res
      if (1 == m) goto 150
      call resid(2,rescg,a,u,f,ia,ja,iw,imin,imax,iminw)
      write (iup,9010) iter,rescg,res,cfac
      goto 110
      150 write (iup,9020) iter,res,cfac
      110 if (iconv == 1) goto 120
      epsil = epsi
      if (iconv /= 4) goto 115
      umax = 0.d0
      do i=imin(1),imax(1)
         umax = max(umax,abs(u(i)))
      end do
      epsil = epsi*umax
      115 if (res < epsil) goto 170
   120 continue
   end do  ! iter=1,ncycle
   170 if (iout /= 3.and.iout /= 0) call resid(1,res,a,u,f,ia,ja,iw, &
         imin,imax,iminw)
   return
   
   9000 format (' CYCLE  0:',3x,'RES=',d9.3)
   9005 format (/' ************* CYCLING..... *************'/)
   9010 format ( ' CYCLE ',i2,':   RESCG=',d9.3,'   RES=',d9.3, &
         '   CFAC=',d9.3)
   9020 format ( ' CYCLE ',i2,':   RES=',d9.3,'   CFAC=',d9.3)
   9030 format (' --------------------------------------------------', &
         '----------')
   9040 format (/' CYCLING BETWEEN GRIDS 1 AND',i3,':'/)
   9050 format (' *** ERROR IN SOLVE: NDU TOO SMALL ***')
   9060 format (' *** ERROR IN SOLVE: NDF TOO SMALL ***')
end subroutine solve 

!.......................................................................

!     CYC                                                    SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine cyc here]
subroutine cyc(l,nrdx,nrdtyp,nrdlen,nrcx,nrux,nrutyp,nrulen, &
      igam,a,u,f,ia,ja,iw,imin,imax,iminw,imaxw, &
      ifg,icg,nstcol,ng,time,irow0,m,ium,ierr,iter, &
      nsc,nda,ndja,mda,mdja,msel,fac,resi,levels)
   
   !     PERFORMS ONE AMG CYCLE WITH GRID L AS FINEST GRID
   
   
   _REAL_ a(*),u(*),f(*),resi(*),res,fac
   _REAL_ time(*),tnew,told
   integer nrdtyp(*),nrutyp(*),ia(*),ja(*),iw(*),imin(*),imax(*), &
         iminw(*),imaxw(*),ifg(*),icg(*),nstcol(*),ng(*)
   external coarse,ctime,inta,putz,relx,resid,vscale
   save
   
   !===> DURING FIRST CYCLE: INITIALIZE PARAMETERS CONTROLLING YALE-SMP
   !     FACTORIZATION AND ADAPTIVE DETERMINATION OF COARSEST GRID:
   !     IFAC=1: ON NEXT CALL OF YALE-SMP FACTORIZE MATRIX
   !     IFI =1: ON FIRST RETURN TO NEXT TO COARSEST GRID AFTER COARSE
   !             GRID SOLUTION COMPARE RESIDUAL WITH THE RESIDUAL ON THE
   !             SAME GRID BEFORE COARSE GRID SOLUTION. IF REDUCTION OF
   !             RESIDUAL NOT SATISFYING, REDUCE NUMBER OF GRIDS USED IN
   !             CYCLING BY ONE AND REPETE THE PROCESS WITH THE NOW
   !             COARSEST GRID.
   !     NPTSF:  NUMBER OF GRID POINTS ON FINEST GRID USED IN CYCLE,
   !             DIVIDED BY 10. ONLY GRIDS WITH LESS THEN NPTSF POINTS
   !             ARE ALLOWED TO BECOME COARSEST GRID DURING ADAPTIVE
   !             COARSE GRID DETERMINATION.
   
   if (iter == 1) ifac = 1
   if (msel /= 2) then
      ifi   = 0
      nptsf = 0
   else
      mink  = 1000000
      ifi   = 1
      nptsf = (imax(l)-imin(l)+1)/10
   end if
   if (l < m) goto 100
   
   !===> ONE GRID ONLY
   
   call coarse(m,ifac,nsc,nrcx,ium,ierr, &
         a,u,f,ia,ja,iw,imin,imax,iminw,icg,nstcol,time, &
         nda,ndja,mda,mdja,irow0)
   goto 1000
   
   !===> MORE THAN ONE GRID
   
   100 do k=l,m
      ng(k) = 0
   end do
   ivstar = 3-igam
   k = l
   
   !===> RELAX (DOWNWARDS)
   
   150 do n=1,nrdx
      do nl=1,nrdlen
         call relx(k,nrdtyp(nl),a,u,f,ia,ja,iw,imin,imax,iminw,icg, &
               nstcol,time)
      end do
   end do
   if (ifi /= 1.or.imax(k)-imin(k) >= nptsf) goto 190
   
   !===> MINK: LOWEST GRID NUMBER FOR WHICH RESIDUAL IS STORED DURING
   !           FIRST DOWNWARDS RELAXATION
   
   if (mink == 1000000) mink = k
   call ctime(told)
   call resid(k,resi(k),a,u,f,ia,ja,iw,imin,imax,iminw)
   call ctime(tnew)
   time(15) = time(15)+tnew-told
   190 ng(k) = ng(k)+1
   k = k+1
   195 call putz(k,u,imin,imax,time)
   call resc(k,a,u,f,ia,ja,iw,imin,imax,iminw,imaxw,ifg,time)
   if (k < m) goto 150
   
   !===> SOLVE ON COARSEST GRID
   
   call coarse(m,ifac,nsc,nrcx,ium,ierr, &
         a,u,f,ia,ja,iw,imin,imax,iminw,icg,nstcol,time, &
         nda,ndja,mda,mdja,irow0)
   
   !===> RELAX (UPWARDS)
   
   200 call vscale(k,ivstar,a,u,f,ia,ja,iw,imin,imax,iminw,time)
   k = k-1
   call inta(k,a,u,ia,ja,iw,imin,imax,iminw,imaxw,ifg,time)
   do n=1,nrux
      do nl=1,nrulen
         call relx(k,nrutyp(nl),a,u,f,ia,ja,iw,imin,imax,iminw,icg, &
               nstcol,time)
      end do
   end do
   if (ifi /= 1.or.k < mink) goto 219
   
   !===> ON FIRST RETURN TO NEXT TO COARSEST GRID COMPARE RESIDUAL WITH
   !     THE PREVIOUS ONE
   
   call ctime(told)
   call resid(k,res,a,u,f,ia,ja,iw,imin,imax,iminw)
   call ctime(tnew)
   time(15) = time(15)+tnew-told
   
   !===> IF RESIDUAL REDUCTION SATISFYING: COARSE GRID ADAPTION FINISHED
   
   if (res < resi(k)*fac) then
      ifi  = 0
   else
      if (nsc == 2) levels = k
      m    = k
      ifac = 1
      goto 195
   end if
   219 if (k == l) goto 1000
   
   !===> GRID SWITCHING CORRESPONDING TO IGAM
   
   if (igam >= 3) goto 220
   if (igam >= 0) goto 200
   if (k == l+1 .and. ng(k) < abs(igam)) goto 150
   goto 200
   220 if (ng(k) < 2) goto 150
   if (igam == 4)  ng(k) = 0
   goto 200
   
   !===> RETURN
   
   1000 call nrmu(l,u,imin,imax,irow0)
end subroutine cyc 

!.......................................................................

!     COARSE                                                SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine coarse here]
subroutine coarse(m,ifac,nsc,nrcx,ium,ierr,a,u,f,ia,ja, &
      iw,imin,imax,iminw,icg,nstcol,time,nda,ndja, &
      mda,mdja,irow0)
   
   !     SOLVES ON COARSEST GRID, EITHER WITH GAUSS-SEIDEL RELAXATION
   !     (NSC=1) OR WITH THE YALE-SMP DIRECT SOLVER NDRV (NSC=2)
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ time(*),told,tnew
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),icg(*), &
         nstcol(*),esp,path,flag
   external ctime,ndrv,relx,resid
   
   !===> CONV: IF COARSE GRID SOLUTION IS DONE WITH GS-RELAXATION AND
   !     NRCX=0, AS MANY GS-SWEEPS ARE PERFORMED AS ARE NECESSARY TO RE-
   !     DUCE THE RESIDUAL BY THE FACTOR CONV
   
   parameter (conv=1.d-2)
   save
   
   if (nsc /= 2) goto 400
   
   !===> SOLUTION WITH YALE-SMP
   
   call ctime(told)
   ilo = imin(m)
   jlo = ia(ilo)
   if (ifac == 1) then
      
      !===> FIRST CALL ON GRID M, FIRST FACTORIZE MATRIX
      
      ihi = imax(m)
      jhi = iw(iminw(m))-1
      np = ihi-ilo+1
      is = ilo-1
      js = jlo-1
      
      !===>   TEST OF AVAILABLE WORK SPACE
      
      if (jhi+3*np > ndja) then
         nsc = 1
         write (ium,9020)
         ierr = -3
         goto 400
      end if
      
      !===>   INITIALISATION OF YALE-SMP POINTER VECTORS
      
      do i=1,np
         ja(jhi+i) = i
         ja(jhi+np+i) = i
         ja(jhi+2*np+i)= i
      end do
      if (irow0 /= 1) then
         
         !===>   COARSE GRID OPERATOR REGULAR, SHIFT CONTENTS OF POINTER
         !       VECTORS IA AND JA
         
         do i=ilo,ihi
            ia(i) = ia(i)-js
         end do
         iaux = ia(ihi+1)
         ia(ihi+1) = iw(iminw(m))-js
         do i=jlo,jhi
            ja(i) = ja(i)-is
         end do
         npoint = np
      else
         
         !===>   COARSE GRID OPERATOR HAS ROWSUM ZERO, ELIMINATE LAST SOLUTION
         !       COMPONENT BY SETTING U(IHI) TO ZERO AND CANCELLING THE COR-
         !       RESPONDING ENTRIES IN A AND JA, RESPECTIVELY. THE CANCELLED
         !       ENTRIES ARE STORED BEFORE THE LAST ROW IN A AND JA.
         
         iaux = ia(ihi)
         
         !===>     JPOS: POINTER TO POSITION IN A AND JA TO CONTAIN NEXT
         !               ELIMINATED ENTRY
         
         jpos = iaux-1
         j = ia(ilo)
         do i=ilo,ihi-1
            ia(i) = j-js
            ja(j) = i-is
            j = j+1
            13 if (j == ia(i+1)) goto 20
            if (ja(j) /= ihi) then
               ja(j) = ja(j)-is
               j = j+1
            else
               aaux = a(j)
               do jj=j,jpos-1
                  a(jj) = a(jj+1)
                  ja(jj) = ja(jj+1)
               end do
               do ii=i+1,ihi
                  ia(ii) = ia(ii)-1
               end do
               a(jpos) = aaux
               ja(jpos) = i
               jpos = jpos-1
            end if
            goto 13
         20 continue
         end do
         ia(ihi) = ia(ihi)-js
         
         !===>     DECREASE NUMBER OF POINTS BY ONE AND SET LAST SOLUTION
         !         COMPONENT TO ZERO
         
         npoint = np-1
         u(ihi) = 0.d0
      end if  ! (irow0 /= 1)
      nsp = nda-jhi
      path = 1
      call ndrv(npoint,ja(jhi+1),ja(jhi+np+1),ja(jhi+2*np+1),ia(ilo), &
            ja(jlo),a(jlo),f(ilo),u(ilo),nsp,a(jhi+1),a(jhi+1), &
            esp,path,flag)
      
      !===>   RESTORE PREVIOUS VALUES FOR IA, JA AND A, IN PARTICULAR PUT
      !       BACK ELIMINATED MATRIX ENTRIES, IF ROWSUM ZERO
      
      do i=ilo,ihi
         ia(i) = ia(i)+js
      end do
      if (irow0 /= 1) then
         ia(ihi+1) = iaux
         do i=jlo,jhi
            ja(i) = ja(i)+is
         end do
      else
         do i=jlo,jpos
            ja(i)=ja(i)+is
         end do
         do j=jpos+1,iaux-1
            aaux = a(j)
            i = ja(j)
            do ii=i+1,ihi
               ia(ii) = ia(ii)+1
            end do
            do jj=j,ia(i+1),-1
               a(jj) = a(jj-1)
               ja(jj) = ja(jj-1)
            end do
            a(ia(i+1)-1) = aaux
            ja(ia(i+1)-1) = ihi
         end do
      end if
      
      !===>   IF AN ERROR OCCURED DURING EXECUTION OF NDRV, SOLVE WITH
      !       GAUSS-SEIDEL RELAXATION
      
      if (flag /= 0) then
         nsc = 1
         if (esp < 0) then
            write (ium,9030)
            ierr = -1
         else
            write (ium,9040) flag
            ierr = -32
         end if
         goto 400
      else
         
         !===>     FACTORIZATION SUCCESSFULL, UPDATE LIMITS OF USED STORAGE
         
         mda = max(mda,nda-esp)
         mdja = max(mdja,jhi+3*np)
         ifac = 0
      end if
   else
      
      !===>   FACTORIZATION ALLREADY DONE
      
      if (irow0 /= 1) then
         npoint = np
      else
         npoint = np-1
         u(ihi) = 0.d0
      end if
      path = 3
      call ndrv(npoint,ja(jhi+1),ja(jhi+np+1),ja(jhi+2*np+1),ia(ilo), &
            ja(jlo),a(jlo),f(ilo),u(ilo),nsp,a(jhi+1),a(jhi+1), &
            esp,path,flag)
   end if  ! (ifac == 1)
   
   !===> UPDATE TIME COUNTER
   
   call ctime(tnew)
   time(17) = time(17)+tnew-told
   goto 190
   
   !===> SOLUTION WITH GAUSS-SEIDEL RELAXATION
   
   400 if (nrcx /= 0) then
      do iter=1,nrcx
         call relx(m,2,a,u,f,ia,ja,iw,imin,imax,iminw,icg,nstcol,time)
      end do
   else
      
      !===> IF NRCX=0: REDUCE RESIDUAL ON COARSEST GRID BY FACTOR CONV
      !                (IF NOT YET IN THE RANGE OF THE TRUNCATION ERROR)
      
      call ctime(told)
      
      !===>   CALCULATE SUPREMUM NORM OF RIGHT HAND SIDE
      
      fmax = 0.d0
      do i=imin(m),imax(m)
         fmax = max(fmax,abs(f(i)))
      end do
      call resid(m,resold,a,u,f,ia,ja,iw,imin,imax,iminw)
      call ctime(tnew)
      time(15) = time(15)+tnew-told
      resold = max(resold*conv,fmax*1.d-12)
      do i=1,10
         do j=1,10
            call relx(m,2,a,u,f,ia,ja,iw,imin,imax,iminw,icg,nstcol, &
                  time)
         end do
         call ctime(told)
         call resid(m,resnew,a,u,f,ia,ja,iw,imin,imax,iminw)
         call ctime(tnew)
         time(15) = time(15)+tnew-told
         if (resnew <= resold) goto 190
      end do
   end if  ! (nrcx /= 0)
   
   !===> COMPUTE RESIDUAL AFTER SOLUTION ON GRID M
   
   190 return
   
   9020 format (' --- WARNG IN COARSE: NO YALE-SMP BECAUSE NDJA TOO ', &
         'SMALL')
   9030 format (' --- WARNG IN COARSE: NO YALE-SMP BECAUSE NDA TOO SMALL')
   9040 format (' --- WARNG IN COARSE: NO YALE-SMP BECAUSE OF ERROR IN ', &
         'FACTORIZATION,'/'     CODE=',i8)
end subroutine coarse 

!.......................................................................

!     RELX                                                  SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine relx here]
subroutine relx(k,irel,a,u,f,ia,ja,iw,imin,imax,iminw,icg,nstcol, &
      time)
   
   !     PERFORMS ONE (PARTIAL) GAUSS-SEIDEL SWEEP ON GRIK K:
   
   !     IREL = 1 :   PARTIAL GAUSS-SEIDEL SWEEP (ONLY F-POINTS)
   !          = 2 :   FULL GAUSS-SEIDEL SWEEP (ALL POINTS)
   !          = 3 :   PARTIAL GAUSS-SEIDEL SWEEP (ONLY C-POINTS)
   !          = 4 :   FULL SWEEP: FF -- C -- COLORS (HIGHEST FIRST)
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),icg(*), &
         nstcol(*)
   external ctime
   save
   
   call ctime(told)
   iaux = ia(imax(k)+1)
   ia(imax(k)+1) = iw(iminw(k))
   goto (100,200,300,400), irel
   goto 200
   
   !===> F-RELAXATION
   
   100 do i=imin(k),imax(k)
      if (icg(i) > 0) goto 120
      s = f(i)
      do j=ia(i)+1,ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      u(i) = s/a(ia(i))
   120 continue
   end do
   goto 1000
   
   !===> FULL GS RELAXATION
   
   200 do i=imin(k),imax(k)
      s = f(i)
      do j=ia(i)+1,ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      u(i) = s/a(ia(i))
   end do
   goto 1000
   
   !===> C-RELAXATION
   
   300 do i=imin(k),imax(k)
      if (icg(i) <= 0) goto 320
      s = f(i)
      do j=ia(i)+1,ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      u(i) = s/a(ia(i))
   320 continue
   end do
   goto 1000
   
   !===> FF-RELAXATION
   
   400 do i=imin(k),imax(k)
      if (icg(i) /= 0) goto 420
      s = f(i)
      do j=ia(i)+1,ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      u(i) = s/a(ia(i))
   420 continue
   end do
   
   !===> C-RELAXATION
   
   do i=imin(k),imax(k)
      if (icg(i) <= 0) goto 440
      s = f(i)
      do j=ia(i)+1,ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      u(i) = s/a(ia(i))
   440 continue
   end do
   
   !===> MORE-COLOR RELAXATION
   
   i = nstcol(k)
   470 if (i >= 100000000) goto 1000
   s = f(i)
   do j=ia(i)+1,ia(i+1)-1
      s = s-a(j)*u(ja(j))
   end do
   u(i) = s/a(ia(i))
   i = -icg(i)
   goto 470
   
   1000 call ctime(tnew)
   time(13) = time(13)+tnew-told
   ia(imax(k)+1) = iaux
   return
end subroutine relx 

!.......................................................................

!     VSCALE                                                SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine vscale here]
subroutine vscale(k,ivstar,a,u,f,ia,ja,iw,imin,imax,iminw,time)
   
   !     SCALES ACTUAL APPROXIMATE SOLUTION ON GRID K (V*-CYCLE); SCALING
   !     IS DONE SUCH THAT ENERGY NORM BECOMES MINIMAL
   
   !     NOTE: THIS SCALING MAKES SENSE ONLY FOR SYMMETRIC PROBLEMS
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*)
   external ctime
   save
   
   !===> COMPUTATION OF SCALING FACTOR "FAC"
   
   if (ivstar /= 1) return
   call ctime(told)
   iaux = ia(imax(k)+1)
   ia(imax(k)+1) = iw(iminw(k))
   s1 = 0.0d0
   s2 = 0.0d0
   do i=imin(k),imax(k)
      sa = 0.0d0
      do j=ia(i),ia(i+1)-1
         sa = sa+a(j)*u(ja(j))
      end do
      20 continue
      s1 = s1+u(i)*f(i)
      s2 = s2+u(i)*sa
   end do
   fac = 1.0d0
   if (s2 /= 0.0d0) fac = s1/s2
   
   !===> SCALING
   
   do i=imin(k),imax(k)
      u(i) = u(i)*fac
   end do
   ia(imax(k)+1) = iaux
   call ctime(tnew)
   time(14) = time(14)+tnew-told
   return
end subroutine vscale 

!.......................................................................

!     INTA                                                  SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine inta here]
subroutine inta(kf,a,u,ia,ja,iw,imin,imax,iminw,imaxw,ifg,time)
   
   !     INTERPOLATES CORRECTION FROM GRID KF+1 TO GRID KF
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),imaxw(*),ifg(*)
   external ctime
   save
   
   !===> C->C CONTRIBUTIONS
   
   call ctime(told)
   do ic=imin(kf+1),imax(kf+1)
      if = ifg(ic)
      u(if) = u(if)+u(ic)
   end do
   
   !===> C->F CONTRIBUTIONS
   
   iaux = iw(imaxw(kf)+1)
   iw(imaxw(kf)+1) = ia(imin(kf+1))
   do i=iminw(kf),imaxw(kf)
      if = ifg(i)
      do j=iw(i),iw(i+1)-1
         u(if) = u(if)+a(j)*u(ja(j))
      end do
   end do
   iw(imaxw(kf)+1) = iaux
   call ctime(tnew)
   time(11) = time(11)+tnew-told
   return
end subroutine inta 

!.......................................................................

!     RESC                                                  SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine resc here]
subroutine resc(kc,a,u,f,ia,ja,iw,imin,imax,iminw,imaxw,ifg,time)
   
   !     RESTRICTS RESIDUALS FROM GRID KC-1 TO GRID KC
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*),imaxw(*),ifg(*)
   external ctime
   save
   
   !===> TRANSFER OF C-POINT DEFECTS
   
   call ctime(told)
   iaux = ia(imax(kc-1)+1)
   ia(imax(kc-1)+1) = iw(iminw(kc-1))
   iaux1 = iw(imaxw(kc-1)+1)
   iw(imaxw(kc-1)+1) = iaux
   do ic=imin(kc),imax(kc)
      if = ifg(ic)
      d = f(if)
      do j=ia(if),ia(if+1)-1
         d = d-a(j)*u(ja(j))
      end do
      f(ic) = d
   end do
   
   !===> TRANSFER OF F-POINT DEFECTS
   
   do i=iminw(kc-1),imaxw(kc-1)
      if = ifg(i)
      d = f(if)
      do j=ia(if),ia(if+1)-1
         d = d-a(j)*u(ja(j))
      end do
      do j=iw(i),iw(i+1)-1
         f(ja(j)) = f(ja(j))+a(j)*d
      end do
   end do
   ia(imax(kc-1)+1) = iaux
   iw(imaxw(kc-1)+1) = iaux1
   call ctime(tnew)
   time(12) = time(12)+tnew-told
   return
end subroutine resc 

!.......................................................................

!     PUTZ                                                   SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine putz here]
subroutine putz(k,u,imin,imax,time)
   
   !     PUTS ZERO TO U-VALUES OF GRID K
   
   _REAL_ u(*)
   _REAL_ told,tnew,time(*)
   integer imin(*),imax(*)
   external ctime
   save
   
   call ctime(told)
   do i=imin(k),imax(k)
      u(i) = 0.0d0
   end do
   call ctime(tnew)
   time(15) = time(15)+tnew-told
   return
end subroutine putz 

!.......................................................................

!     FIRST                                                 SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine first here]
subroutine first(ifirst,u,imin,imax,iarr,irow0)
   
   !     PUTS A FIRST APPROXIMATION TO FINEST GRID
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ u(*)
   integer imin(*),imax(*),iarr(*)
   external idec,random
   save
   
   if (ifirst /= 0) then
      call idec(ifirst,3,ndigit,iarr)
      ifrst = iarr(2)
   else
      ifrst = 3
   end if
   
   goto (100,200,300), ifrst
   return
   
   100 do i=imin(1),imax(1)
      u(i) = 0.0d0
   end do
   return
   
   200 do i=imin(1),imax(1)
      u(i) = 1.0d0
   end do
   if (irow0 < 2) u(imax(1)) = 0.0d0
   return
   
   300 if (iarr(3)*ifirst == 0) goto 350
   s = dble(iarr(3))
   do i=1,10
      s = s*0.1d0
      if (s < 1.0d0) goto 370
   end do
   350 s = 0.72815d0
   370 do i=imin(1),imax(1)
      u(i) = random(s)
   end do
   if (irow0 < 2) u(imax(1)) = 0.0d0
   return
end subroutine first 

!.......................................................................

!     INJF                                                   SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine injf here]
subroutine injf(kc,f,imin,imax,ifg)
   
   !     INJECTS F-VALUES FROM GRID KC-1 TO GRID KC
   
   _REAL_ f(*)
   integer imin(*),imax(*),ifg(*)
   save
   do ic=imin(kc),imax(kc)
      f(ic) = f(ifg(ic))
   end do
   return
end subroutine injf 

!.......................................................................

!     INJU                                                   SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine inju here]
subroutine inju(kc,u,imin,imax,ifg)
   
   !     INJECTS U-VALUES FROM GRID KC-1 TO GRID KC
   
   _REAL_ u(*)
   integer imin(*),imax(*),ifg(*)
   save
   do ic=imin(kc),imax(kc)
      u(ic) = u(ifg(ic))
   end do
   return
end subroutine inju 

!.......................................................................

!     NRMU                                                SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine nrmu here]
subroutine nrmu(k,u,imin,imax,irow0)
   
   !     NORMALIZES U ON GRID K IF ROWSUM=0 (LAST COMPONENT =0)
   
   _REAL_ u(*),fac
   integer imin(*),imax(*)
   save
   
   if (irow0 > 1) return
   fac = u(imax(k))
   do i=imin(k),imax(k)
      u(i) = u(i)-fac
   end do
   return
end subroutine nrmu 

!.......................................................................

!     RESID                                                 SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine resid here]
subroutine resid(k,res,a,u,f,ia,ja,iw,imin,imax,iminw)
   
   !     COMPUTES L2-NORM OF RESIDUAL ON GRID K
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*)
   save
   
   res = 0.0d0
   iaux = ia(imax(k)+1)
   ia(imax(k)+1) = iw(iminw(k))
   do i=imin(k),imax(k)
      s = f(i)
      do j=ia(i),ia(i+1)-1
         s = s-a(j)*u(ja(j))
      end do
      res = res+s*s
   end do
   ia(imax(k)+1) = iaux
   res = sqrt(res)
   return
end subroutine resid 

!.......................................................................

!     CG                                                 SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine cg here]
subroutine cg(k,icgr,iter,a,u,f,ia,ja,iw,imin,imax,iminw,m,time, &
      ierr,ium)
   
   !     PERFORMS ONE STEP OF PRECONDITIONED CONJUGATE GRADIENT
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   _REAL_ told,tnew,time(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*)
   external cgalf,cgeps,ctime
   save
   
   if (icgr == 0) return
   
   !===> COMPUTE MOST RECENT MG CORRECTION
   
   call ctime(told)
   nnu = imax(k)-imin(k)+1
   ishift = imax(m)+1-imin(k)
   do i=imin(k),imax(k)
      u(i) = u(i)-u(i+ishift)
   end do
   if (icgr == 2 .and. iter > 1) goto 100
   
   !===> FIRST CG STEP
   
   do i=imin(k),imax(k)
      f(i+ishift) = u(i)
   end do
   goto 200
   
   !===> NEXT CG STEPS (IF ICGR=2 ONLY)
   
   100 alf = cgalf(k,s2,a,u,f,ia,ja,iw,imin,imax,iminw,m)
   do i=imin(k),imax(k)
      f(i+ishift) = u(i)+alf*f(i+ishift)
   end do
   200 eps = cgeps(k,s2,a,u,f,ia,ja,iw,imin,imax,iminw,m,ierr,ium)
   if (ierr > 0) return
   do i=imin(k),imax(k)
      u(i) = u(i+ishift)+eps*f(i+ishift)
   end do
   call ctime(tnew)
   time(16) = time(16)+tnew-told
   return
end subroutine cg 

!.......................................................................

!     USAVE                                            SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine usave here]
subroutine usave(k,icgr,u,imin,imax,ndu,m,time,ierr,ium,mdu,ndf, &
      mdf)
   
   !     MAKES A BACK-UP OF THE CURRENT APPROXIMATION ON LEVEL K
   !     IF ICGR.NE.0.
   
   _REAL_ u(*)
   _REAL_ time(*)
   integer imin(*),imax(*)
   external ctime
   save
   
   if (icgr == 0) return
   ishift = imax(m)+1-imin(k)
   mdu = max(mdu,imax(k)+ishift)
   mdf = max(mdf,imax(k)+ishift)
   if (mdu <= ndu.and.mdf <= ndf) goto 10
   write (ium,9000) mdu,mdf
   ierr = -4
   icgr = 0
   return
   
   10 call ctime(told)
   do i=imin(k),imax(k)
      u(i+ishift) = u(i)
   end do
   call ctime(tnew)
   time(16) = time(16)+tnew-told
   return
   
   9000 format (' --- WARNG IN USAVE: NO CG BECAUSE OF STORAGE ---'/ &
         '     REQUIRED: NDU =',    i9                      / &
         '               NDF =',    i9                      )
end subroutine usave 

!.......................................................................

!     CGEPS                                            FUNCTION

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of real function cgeps here]
real function cgeps(k,s2,a,u,f,ia,ja,iw, &
      imin,imax,iminw,m,ierr,ium)
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   integer imin(*),imax(*),iminw(*),ia(*),iw(*),ja(*)
   save
   
   ishift = imax(m)+1-imin(k)
   s1 = 0.0d0
   s2 = 0.0d0
   iaux = ia(imax(k)+1)
   ia(imax(k)+1) = iw(iminw(k))
   do i=imin(k),imax(k)
      sr = f(i)
      sp = 0.0d0
      do j=ia(i),ia(i+1)-1
         sr = sr-a(j)*u(ja(j)+ishift)
         sp = sp+a(j)*f(ja(j)+ishift)
      end do
      40 continue
      s1 = s1+sr*f(i+ishift)
      s2 = s2+sp*f(i+ishift)
   end do
   ia(imax(k)+1) = iaux
   if (s2 == 0.0d0) goto 100
   cgeps = s1/s2
   return
   
   !===> ERROR EXIT
   
   100 write (ium,9000)
   ierr = 31
   return
   9000 format (' *** ERROR IN CGEPS: CG CORRECTION NOT DEFINED ***')
end function cgeps 

!.......................................................................

!     CGALF                                            FUNCTION

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of real function cgalf here]
real function cgalf(k,s2,a,u,f,ia,ja,iw, &
      imin,imax,iminw,m)
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   integer ia(*),ja(*),iw(*),imin(*),imax(*),iminw(*)
   save
   
   ishift=imax(m)+1-imin(k)
   s1 = 0.0d0
   iaux = ia(imax(k)+1)
   ia(imax(k)+1) = iw(iminw(k))
   do i=imin(k),imax(k)
      sr = 0.0d0
      do j=ia(i),ia(i+1)-1
         sr = sr+a(j)*u(ja(j))
      end do
      s1 = s1+sr*f(i+ishift)
   end do
   cgalf = -s1/s2
   ia(imax(k)+1) = iaux
   return
end function cgalf 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     OUTPUT OF STATISTICAL INFORMATION ON CP-TIMES AND DIMENSIONING

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!.......................................................................

!     WRKCNT                                           SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine wrkcnt here]
subroutine wrkcnt(iout,ia,iw,imin,imax,iminw,levels,time,ncyc0, &
      iup,mda,mdia,mdja,mdu,mdf,mdig,res0,res)
   
   !     RESIDUALS / CP-TIMES / COMPLEXITY / DIMENSIONING
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ time(*),t(10),sum1,sum2
   integer ia(*),iw(*),imin(*),imax(*),iminw(*)
   save
   
   if (iout < 1) return
   
   !===> RESIDUALS / CONVERGENCE
   
   if (ncyc0 > 0) then
      cfac = res/(res0+1.d-40)
!     write (iup,9110) res0,res,cfac
      cfpc = cfac**(1.d0/dble(ncyc0))
!     write (iup,9120) cfpc
   end if
   if (iout <= 1) return
!  write (iup,9000)
   nnu = imax(1)
   
   !===> COMPUTING TIMES
   
   sum1 = 0.0
   sum2 = 0.0
   do i=1,10
      t(i) = 0.0
      if (ncyc0 > 0) t(i) = time(i+10)/real(ncyc0)
      sum1 = sum1+time(i)
      sum2 = sum2+t(i)
   end do
   
!  write (iup,9020)
!  write (iup,9030)
!  write (iup,9040) (time(i),t(i),i=1,8)
!  write (iup,9030)
!  write (iup,9050) sum1,sum2
!  write (iup,9030)
   
   !===> SPACE OCCUPIED BY OPERATORS A(1) - A(LEVELS)
   
   idima = 0
   do k=1,levels
      idima = idima+iw(iminw(k))-ia(imin(k))
   end do
   
   !===> THEORETICAL MINIMAL SPACE REQUIREMENTS
   
   if (levels < 2) return
   mdta  = iw(iminw(levels))-1
   mdtja = mdta
   mdtia = imax(levels)+1
   mdtu  = imax(levels)
   mdtf  = mdtu
   mdtig = 2*mdtu+nnu
!  write (iup,9100) mda,mdta,mdja,mdtja,mdia,mdtia,mdu,mdtu, &
!        mdf,mdtf,mdig,mdtig
   
   !===> COMPLEXITIES
   
   scmplx = dble(2*(mda+mdu+mdf)+mdja+mdia+mdig)/ &
         dble(1+5*nnu+3*(iw(iminw(1))-ia(imin(1))))
   tcmplx = dble(2*(mdta+mdtu+mdtf)+mdtja+mdtia+mdtig)/ &
         dble(1+5*nnu+3*(iw(iminw(1))-ia(imin(1))))
   
   acmplx = dble(idima)/dble(iw(1)-1)
   ocmplx = dble(mdtu )/dble(nnu)
!  write (iup,9080) acmplx,ocmplx,scmplx,tcmplx
   return
   
   9000 format (//' ************** WORK COUNT ***************'/)
   9020 format (  '   PREP       SEC       SOL      SEC/CYCLE')
   9030 format (  ' -----------------------------------------')
   9040 format (  ' 1 RWSRT   ',f7.2,'   11 INTADD   ',f7.2/ &
         ' 2 PRE-COL ',f7.2,'   12 RESCAL   ',f7.2/ &
         ' 3 CHK-COL ',f7.2,'   13 RELAX    ',f7.2/ &
         ' 4 INTERPOL',f7.2,'   14 V-*      ',f7.2/ &
         ' 5 RESTRICT',f7.2,'   15 OTHERS   ',f7.2/ &
         ' 6 OPDFN   ',f7.2,'   16 CONJ-GRAD',f7.2/ &
         ' 7 TRUNC   ',f7.2,'   17 YALE-SMP ',f7.2/ &
         ' 8 OTHERS  ',f7.2,'   18 ------   ',f7.2)
   9050 format (  '   SUM     ',f7.2,'      SUM      ',f7.2)
   9080 format (/' ******************* COMPLEXITIES ********************'/ &
         ' SPACE OCCUPIED BY ALL OPERATORS / SPACE OF OPERATOR  '/ &
         ' ON THE FINEST GRID   = ',f8.2,'   (A-COMPLEXITY)     '/ &
         ' TOTAL NUMBER OF GRID POINTS / NUMBER OF POINTS IN    '/ &
         ' THE  FINEST  GRID    = ',f8.2,'   (O-COMPLEXITY)     '/ &
         ' TOTAL SPACE USED BY AMG1R5 / SPACE OCCUPIED BY USER- '/ &
         ' DEFINED  PROBLEM     = ',f8.2,'   (S-COMPLEXITY)     '/ &
         ' SPACE USED DURING SOLUTION PHASE / SPACE OCCUPIED BY '/ &
         ' USER-DEFINED PROBLEM = ',f8.2                         / &
         ' *****************************************************')
   9100 format (//' ********* SPACE REQUIREMENTS *********'// &
         ' VECTOR      NEEDED      THEOR. MINIMUM'/ &
         ' --------------------------------------'/ &
         '   A ',        i14   ,2x,     i16       / &
         '   JA',        i14   ,2x,     i16       / &
         '   IA',        i14   ,2x,     i16       / &
         '   U ',        i14   ,2x,     i16       / &
         '   F ',        i14   ,2x,     i16       / &
         '   IG',        i14   ,2x,     i16       / &
         ' --------------------------------------'/)
   9110 format (//' **************** CONVERGENCE *****************'/ &
         ' L2-NORM OF RESIDUAL BEFORE CYCLING =',d10.3/ &
         ' L2-NORM OF RESIDUAL AFTER  CYCLING =',d10.3/ &
         ' CONVERGENCE FACTOR                 =',d10.3)
   9120 format (  ' CONVERGENCE FACTOR PER CYCLE       =',d10.3)
end subroutine wrkcnt 
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     AMG1R5 AUXILIARY PROGRAMS

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!.......................................................................

!     IDEC                                                  SUBROUTINE

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine idec here]
subroutine idec(into,nnum,ndigit,iarr)
   
   !     DECOMPOSE NON-NEGATIVE INTEGER INTO INTO NNUM INTEGERS
   
   !     INPUT:  INTO   - INTEGER (0.LE. INTO .LE.999999999)
   !             NNUM   - INTEGER (1.LE. NNUM .LE.9); NUMBER OF INTEGERS
   !                      TO BE RETURNED ON ARRAY IARR (SEE BELOW)
   
   !     OUTPUT: NDIGIT - INTEGER; NUMBER OF DIGITS OF INTO
   !             IARR   - INTEGER-ARRAY OF LENGTH 10:
   !                      IARR(1)        = FIRST      DIGIT OF INTO,
   !                      IARR(2)        = SECOND     DIGIT OF INTO, .....
   !                      IARR(NNUM-1)   = (NNUM-1)ST DIGIT OF INTO,
   !                      IARR(NNUM)     = REST OF INTO
   !                      IF NNUM > NDIGIT, THE CORRESPONDING COMPONENTS
   !                      OF IARR ARE PUT TO ZERO.
   
   !     WARNING: BE SURE THAT YOUT COMPUTER CAN STORE NNUM DIGITS ON AN
   !              INTEGER VARIABLE.
   
   implicit _REAL_ (a-h,o-z)
   dimension iarr(*)
   data eps /0.5d0/
   save
   
   nrest = into
   ndigit = 1+int(log10(eps+dble(into)))
   if (nnum >= ndigit) goto 20
   iq = 10**(ndigit-nnum+1)
   iarr(nnum) = nrest-(nrest/iq)*iq
   nrest = nrest/iq
   do i=nnum-1,1,-1
      iarr(i) = nrest-(nrest/10)*10
      nrest = nrest/10
   end do
   return
   
   20 do i=ndigit,1,-1
      iarr(i) = nrest-(nrest/10)*10
      nrest = nrest/10
   end do
   do i=ndigit+1,nnum
      iarr(i) = 0
   end do
   return
end subroutine idec 

!.......................................................................

!     RANDOM                                                FUNCTION

!.......................................................................


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of real function random here]
real function random(s)
   
   !     FUNCTION TO CREATE "RANDOM" SEQUENCE OF NUMBERS BETWEEN 0 AND 0.1
   
   !     INPUT:   S      - NUMBER BETWEEN 0 AND 0.1
   
   !     OUTPUT:  RANDOM - NUMBER BETWEEN 0 AND 0.1
   !              S      - S=RANDOM
   
   implicit _REAL_ (a-h,o-z)
   save
   random=100.0d0*exp(s)
   random=random-dble(int(random))
   s=random
   return
end function random 
!***********************************************************************
! $LARGE
! $NOFLOATCALLS
!                            APPENDIX 1                          1/15/81

!        SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS
!        OF LINEAR EQUATIONS  (UNCOMPRESSED POINTER STORAGE)

!        REAL*8 VERSION. NOTE: THE ORIGINAL SUBROUTINES

!            NDRV, NSF, NNF, NNS AND NNT

!        HAVE BEEN RENAMED TO

!            YALE8, NSF8, NNF8, NNS8 AND NNT8, RESPECTIVELY.

!*** SUBROUTINE YALE8 (OLD NAME: NDRV)
!*** DRIVER FOR SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS OF
!       LINEAR EQUATIONS (UNCOMPRESSED POINTER STORAGE)

!       SUBROUTINE  NDRV  (= OLD NAME)
!       SUBROUTINE  YALE8 (= NEW NAME)

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine ndrv here]
subroutine  ndrv &
      (n, r,c,ic, ia,ja,a, b, z, nsp,isp,rsp,esp, path, flag)
   
   !    PARAMETERS
   !    CLASS ABBREVIATIONS ARE --
   !       N - INTEGER VARIABLE
   !       F - REAL VARIABLE
   !       V - SUPPLIES A VALUE TO THE DRIVER
   !       R - RETURNS A RESULT FROM THE DRIVER
   !       I - USED INTERNALLY BY THE DRIVER
   !       A - ARRAY
   
   ! CLASS   PARAMETER
   ! ------+----------
   
   !         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
   !    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
   !    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
   !    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
   !    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
   !    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
   !    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
   !    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
   !    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
   !    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
   !    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
   !    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
   !    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
   !    CONSECUTIVELY IN
   !            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
   !    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
   !            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
   !    FOR EXAMPLE, THE 5 BY 5 MATRIX
   !                ( 1. 0. 2. 0. 0.)
   !                ( 0. 3. 0. 0. 0.)
   !            M = ( 0. 4. 5. 6. 0.)
   !                ( 0. 0. 0. 7. 0.)
   !                ( 0. 0. 0. 8. 9.)
   !    WOULD BE STORED AS
   !                 1  2  3  4  5  6  7  8  9
   !            ---+--------------------------
   !            IA   1  3  4  7  8 10
   !            JA   1  3  2  2  3  4  4  4  5
   !             A   1. 2. 3. 4. 5. 6. 7. 8. 9.         .
   
   ! NV      N     - NUMBER OF VARIABLES/EQUATIONS.
   ! FVA     A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
   !                   BY ROWS.
   !                   SIZE = NUMBER OF NONZERO ENTRIES IN M.
   ! NVA     IA    - POINTERS TO DELIMIT THE ROWS IN A.
   !                   SIZE = N+1.
   ! NVA     JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
   !                   SIZE = SIZE OF A.
   ! FVA     B     - RIGHT-HAND SIDE B;  B AND Z CAN THE SAME ARRAY.
   !                   SIZE = N.
   ! FRA     Z     - SOLUTION X;  B AND Z CAN BE THE SAME ARRAY.
   !                   SIZE = N.
   
   !         THE ROWS AND COLUMNS OF THE ORIGINAL MATRIX M CAN BE
   !    REORDERED (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY)
   !    BEFORE CALLING THE DRIVER.  IF NO REORDERING IS DONE, THEN SET
   !    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE SOLUTION Z IS RETURNED
   !    IN THE ORIGINAL ORDER.
   
   ! NVA     R     - ORDERING OF THE ROWS OF M.
   !                   SIZE = N.
   ! NVA     C     - ORDERING OF THE COLUMNS OF M.
   !                   SIZE = N.
   ! NVA     IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
   !                   IC(C(I)) = I  FOR I=1,...,N.
   !                   SIZE = N.
   
   !         THE SOLUTION OF THE SYSTEM OF LINEAR EQUATIONS IS DIVIDED INTO
   !    THREE STAGES --
   !      NSF -- THE MATRIX M IS PROCESSED SYMBOLICALLY TO DETERMINE WHERE
   !              FILLIN WILL OCCUR DURING THE NUMERIC FACTORIZATION.
   !      NNF -- THE MATRIX M IS FACTORED NUMERICALLY INTO THE PRODUCT LDU
   !              OF A UNIT LOWER TRIANGULAR MATRIX L, A DIAGONAL MATRIX D,
   !              AND A UNIT UPPER TRIANGULAR MATRIX U, AND THE SYSTEM
   !              MX = B  IS SOLVED.
   !      NNS -- THE LINEAR SYSTEM  MX = B  IS SOLVED USING THE LDU
   !  OR          FACTORIZATION FROM NNF.
   !      NNT -- THE TRANSPOSED LINEAR SYSTEM  MT X = B  IS SOLVED USING
   !              THE LDU FACTORIZATION FROM NNF.
   !    FOR SEVERAL SYSTEMS WHOSE COEFFICIENT MATRICES HAVE THE SAME
   !    NONZERO STRUCTURE, NSF NEED BE DONE ONLY ONCE (FOR THE FIRST
   !    SYSTEM);  THEN NNF IS DONE ONCE FOR EACH ADDITIONAL SYSTEM.  FOR
   !    SEVERAL SYSTEMS WITH THE SAME COEFFICIENT MATRIX, NSF AND NNF NEED
   !    BE DONE ONLY ONCE (FOR THE FIRST SYSTEM);  THEN NNS OR NNT IS DONE
   !    ONCE FOR EACH ADDITIONAL RIGHT-HAND SIDE.
   
   ! NV      PATH  - PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE --
   !                   1  PERFORM NSF AND NNF.
   !                   2  PERFORM NNF ONLY  (NSF IS ASSUMED TO HAVE BEEN
   !                       DONE IN A MANNER COMPATIBLE WITH THE STORAGE
   !                       ALLOCATION USED IN THE DRIVER).
   !                   3  PERFORM NNS ONLY  (NSF AND NNF ARE ASSUMED TO
   !                       HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
   !                       STORAGE ALLOCATION USED IN THE DRIVER).
   !                   4  PERFORM NNT ONLY  (NSF AND NNF ARE ASSUMED TO
   !                       HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
   !                       STORAGE ALLOCATION USED IN THE DRIVER).
   !                   5  PERFORM NSF ONLY.
   
   !         VARIOUS ERRORS ARE DETECTED BY THE DRIVER AND THE INDIVIDUAL
   !    SUBROUTINES.
   
   ! NR      FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
   !                     0     NO ERRORS DETECTED
   !                     N+K   NULL ROW IN A  --  ROW = K
   !                    2N+K   DUPLICATE ENTRY IN A  --  ROW = K
   !                    3N+K   INSUFFICIENT STORAGE IN NSF  --  ROW = K
   !                    4N+1   INSUFFICIENT STORAGE IN NNF
   !                    5N+K   NULL PIVOT  --  ROW = K
   !                    6N+K   INSUFFICIENT STORAGE IN NSF  --  ROW = K
   !                    7N+1   INSUFFICIENT STORAGE IN NNF
   !                    8N+K   ZERO PIVOT  --  ROW = K
   !                   10N+1   INSUFFICIENT STORAGE IN NDRV
   !                   11N+1   ILLEGAL PATH SPECIFICATION
   
   !         WORKING STORAGE IS NEEDED FOR THE FACTORED FORM OF THE MATRIX
   !    M PLUS VARIOUS TEMPORARY VECTORS.  THE ARRAYS ISP AND RSP SHOULD BE
   !    EQUIVALENCED;  INTEGER STORAGE IS ALLOCATED FROM THE BEGINNING OF
   !    ISP AND REAL STORAGE FROM THE END OF RSP.
   
   ! NV      NSP   - DECLARED DIMENSION OF RSP;  NSP GENERALLY MUST
   !                   BE LARGER THAN  5N+3 + 2K  (WHERE  K = (NUMBER OF
   !                   NONZERO ENTRIES IN M)).
   ! NVIRA   ISP   - INTEGER WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
   !                   NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
   !                   EQUIVALENCED.
   !                   SIZE = LRATIO*NSP
   ! FVIRA   RSP   - REAL WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
   !                   NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
   !                   EQUIVALENCED.
   !                   SIZE = NSP.
   ! NR      ESP   - IF SUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
   !                   SYMBOLIC FACTORIZATION (NSF), THEN ESP IS SET TO THE
   !                   AMOUNT OF EXCESS STORAGE PROVIDED (NEGATIVE IF
   !                   INSUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
   !                   NUMERIC FACTORIZATION (NNF)).
   
   
   !  CONVERSION TO REAL
   
   !    TO CONVERT THESE ROUTINES FOR REAL ARRAYS, SIMPLY USE
   !    THE REAL DECLARATIONS IN PLACE OF THE REAL DECLARATIONS
   !    IN EACH SUBPROGRAM;  IN ADDITION, THE DATA VALUE OF THE INTEGER
   !    VARIABLE LRATIO MUST BE SET AS INDICATED IN SUBROUTINE NDRV
   
   integer  r(1), c(1), ic(1),  ia(1), ja(1),  isp(1), esp, &
         path, flag,  q, im, d, u, row, tmp,  umax
   !       REAL  A(1),  B(1),  Z(1),  RSP(1)
   _REAL_  a(1),  b(1),  z(1),  rsp(1)
   
   !  SET LRATIO EQUAL TO THE RATIO BETWEEN THE LENGTH OF FLOATING POINT
   !  AND INTEGER ARRAY DATA;  E. G., LRATIO = 1 FOR (REAL, INTEGER),
   !  LRATIO = 2 FOR (REAL, INTEGER)
   
   data lratio/2/
   
   if (path < 1 .or. 5 < path)  goto 111
   !  ******  INITIALIZE AND DIVIDE UP TEMPORARY STORAGE  *****************
   il = 1
   iu = il + n+1
   jl = iu + n+1
   
   !  ******  CALL NSF IF FLAG IS SET  ************************************
   if ((path-1) * (path-5) /= 0)  goto 2
   max = (lratio*nsp + 1 - jl) - (n+1) - n
   jlmax = max/2
   q     = jl  + jlmax
   im    = q   + (n+1)
   jutmp = im  +   n
   jumax = lratio*nsp + 1 - jutmp
   esp = max/lratio
   if (jlmax <= 0 .or. jumax <= 0)  goto 110
   call  nsf8 &
         (n,  r, ic,  ia, ja, &
         isp(il), isp(jl), jlmax,  isp(iu), isp(jutmp), jumax, &
         isp(q),  isp(im),  flag)
   if (flag /= 0)  goto 100
   !  ******  MOVE JU NEXT TO JL  *****************************************
   jlmax = isp(il+n)-1
   ju    = jl + jlmax
   jumax = isp(iu+n)-1
   if (jumax <= 0)  goto 2
   do j=1,jumax
      isp(ju+j-1) = isp(jutmp+j-1)
   end do
   
   !  ******  CALL REMAINING SUBROUTINES  *********************************
   2 jlmax = isp(il+n)-1
   ju    = jl  + jlmax
   jumax = isp(iu+n)-1
   l     = (ju + jumax - 2 + lratio)  /  lratio    +    1
   lmax  = jlmax
   d     = l   + lmax
   u     = d   + n
   row   = nsp + 1 - n
   tmp   = row - n
   umax  = tmp - u
   esp = umax - jumax
   
   if ((path-1) * (path-2) /= 0)  goto 3
   if (umax <= 0)  goto 110
   call  nnf8 &
         (n,  r, c, ic,  ia, ja, a,  z,  b, &
         isp(il), isp(jl), rsp(l), lmax,   rsp(d), &
         isp(iu), isp(ju), rsp(u), umax, &
         rsp(row),  rsp(tmp),  flag)
   if (flag /= 0)  goto 100
   return
   
   3 if ((path-3) /= 0)  goto 4
   call  nns8 &
         (n,  r, c, &
         isp(il), isp(jl), rsp(l),  rsp(d), &
         isp(iu), isp(ju), rsp(u), &
         z,  b,  rsp(tmp))
   
   4 if ((path-4) /= 0)  goto 5
   call  nnt8 &
         (n,  r, c, &
         isp(il), isp(jl), rsp(l),  rsp(d), &
         isp(iu), isp(ju), rsp(u), &
         z,  b,  rsp(tmp))
   5 return
   
   ! ** ERROR:  ERROR DETECTED IN NSF, NNF, NNS, OR NNT
   100 return
   ! ** ERROR:  INSUFFICIENT STORAGE
   110 flag = 10*n + 1
   return
   ! ** ERROR:  ILLEGAL PATH SPECIFICATION
   111 flag = 11*n + 1
   return
end subroutine ndrv 

!       ----------------------------------------------------------------

!               YALE SPARSE MATRIX PACKAGE - NONSYMMETRIC CODES
!                    SOLVING THE SYSTEM OF EQUATIONS MX = B
!                        (UNCOMPRESSED POINTER STORAGE)

!    I.   CALLING SEQUENCES
!         THE COEFFICIENT MATRIX CAN BE PROCESSED BY AN ORDERING ROUTINE
!    (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY) BEFORE USING
!    THE REMAINING SUBROUTINES.  IF NO REORDERING IS DONE, THEN SET
!    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE CALLING SEQUENCE IS --
!        (      (MATRIX ORDERING))
!         NSF   (SYMBOLIC FACTORIZATION TO DETERMINE WHERE FILLIN WILL
!                 OCCUR DURING NUMERIC FACTORIZATION)
!         NNF   (NUMERIC FACTORIZATION INTO PRODUCT LDU OF UNIT LOWER
!                 TRIANGULAR MATRIX L, DIAGONAL MATRIX D, AND UNIT UPPER
!                 TRIANGULAR MATRIX U, AND SOLUTION OF LINEAR SYSTEM)
!         NNS   (SOLUTION OF LINEAR SYSTEM FOR ADDITIONAL RIGHT-HAND
!     OR          SIDE USING LDU FACTORIZATION FROM NNF)
!         NNT   (SOLUTION OF TRANSPOSED LINEAR SYSTEM FOR ADDITIONAL
!                 RIGHT-HAND SIDE USING LDU FACTORIZATION FROM NNF)

!    II.  STORAGE OF SPARSE MATRICES
!         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!    CONSECUTIVELY IN
!            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!    FOR EXAMPLE, THE 5 BY 5 MATRIX
!                ( 1. 0. 2. 0. 0.)
!                ( 0. 3. 0. 0. 0.)
!            M = ( 0. 4. 5. 6. 0.)
!                ( 0. 0. 0. 7. 0.)
!                ( 0. 0. 0. 8. 9.)
!    WOULD BE STORED AS
!                 1  2  3  4  5  6  7  8  9
!            ---+--------------------------
!            IA   1  3  4  7  8 10
!            JA   1  3  2  2  3  4  4  4  5
!             A   1. 2. 3. 4. 5. 6. 7. 8. 9.         .

!         THE STRICT TRIANGULAR PORTIONS OF THE MATRICES L AND U ARE
!    STORED IN THE SAME FASHION USING THE ARRAYS  IL, JL, L  AND
!    IU, JU, U  RESPECTIVELY.  THE DIAGONAL ENTRIES OF L AND U ARE
!    ASSUMED TO BE EQUAL TO ONE AND ARE NOT STORED.  THE ARRAY D
!    CONTAINS THE RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.

!    III. ADDITIONAL STORAGE SAVINGS
!         IN NSF, R AND IC CAN BE THE SAME ARRAY IN THE CALLING
!    SEQUENCE IF NO REORDERING OF THE COEFFICIENT MATRIX HAS BEEN DONE.
!         IN NNF, R, C AND IC CAN ALL BE THE SAME ARRAY IF NO REORDERING
!    HAS BEEN DONE.  IF ONLY THE ROWS HAVE BEEN REORDERED, THEN C AND IC
!    CAN BE THE SAME ARRAY.  IF THE ROW AND COLUMN ORDERINGS ARE THE
!    SAME, THEN R AND C CAN BE THE SAME ARRAY.  Z AND ROW CAN BE THE
!    SAME ARRAY.
!         IN NNS OR NNT, R AND C CAN BE THE SAME ARRAY IF NO REORDERING
!    HAS BEEN DONE OR IF THE ROW AND COLUMN ORDERINGS ARE THE SAME.  Z
!    AND B CAN BE THE SAME ARRAY;  HOWEVER, THEN B WILL BE DESTROYED.

!    IV.  PARAMETERS
!         FOLLOWING IS A LIST OF PARAMETERS TO THE PROGRAMS.  NAMES ARE
!    UNIFORM AMONG THE VARIOUS SUBROUTINES.  CLASS ABBREVIATIONS ARE --
!       N - INTEGER VARIABLE
!       F - REAL VARIABLE
!       V - SUPPLIES A VALUE TO A SUBROUTINE
!       R - RETURNS A RESULT FROM A SUBROUTINE
!       I - USED INTERNALLY BY A SUBROUTINE
!       A - ARRAY

! CLASS   PARAMETER
! ------+----------
! FVA     A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!                   BY ROWS.
!                   SIZE = NUMBER OF NONZERO ENTRIES IN M.
! FVA     B     - RIGHT-HAND SIDE B.
!                   SIZE = N.
! NVA     C     - ORDERING OF THE COLUMNS OF M.
!                   SIZE = N.
! FVRA    D     - RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!                   SIZE = N.
! NR      FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!                    0     NO ERRORS DETECTED
!                    N+K   NULL ROW IN A  --  ROW = K
!                   2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!                   3N+K   INSUFFICIENT STORAGE FOR JL  --  ROW = K
!                   4N+1   INSUFFICIENT STORAGE FOR L
!                   5N+K   NULL PIVOT  --  ROW = K
!                   6N+K   INSUFFICIENT STORAGE FOR JU  --  ROW = K
!                   7N+1   INSUFFICIENT STORAGE FOR U
!                   8N+K   ZERO PIVOT  --  ROW = K
! NVA     IA    - POINTERS TO DELIMIT THE ROWS IN A.
!                   SIZE = N+1.
! NVA     IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!                   IC(C(I) = I  FOR I=1,...N.
!                   SIZE = N.
! NVRA    IL    - POINTERS TO DELIMIT THE ROWS IN L.
!                   SIZE = N+1.
! NVRA    IU    - POINTERS TO DELIMIT THE ROWS IN U.
!                   SIZE = N+1.
! NVA     JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!                   SIZE = SIZE OF A.
! NVRA    JL    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF L.
!                   SIZE = JLMAX.
! NV      JLMAX - DECLARED DIMENSION OF JL;  JLMAX MUST BE LARGER THAN
!                   THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!                   TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER NSF).
! NVRA    JU    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF U.
!                   SIZE = JUMAX.
! NV      JUMAX - DECLARED DIMENSION OF JU;  JUMAX MUST BE LARGER THAN
!                   THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!                   TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER NSF).
! FVRA    L     - NONZERO ENTRIES IN THE STRICT LOWER TRIANGULAR PORTION
!                   OF THE MATRIX L, STORED BY ROWS.
!                   SIZE = LMAX
! NV      LMAX  - DECLARED DIMENSION OF L;  LMAX MUST BE LARGER THAN
!                   THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!                   TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER NSF).
! NV      N     - NUMBER OF VARIABLES/EQUATIONS.
! NVA     R     - ORDERING OF THE ROWS OF M.
!                   SIZE = N.
! FVRA    U     - NONZERO ENTRIES IN THE STRICT UPPER TRIANGULAR PORTION
!                   OF THE MATRIX U, STORED BY ROWS.
!                   SIZE = UMAX.
! NV      UMAX  - DECLARED DIMENSION OF U;  UMAX MUST BE LARGER THAN
!                   THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!                   TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER NSF).
! FRA     Z     - SOLUTION X.
!                   SIZE = N.


!       ----------------------------------------------------------------

!*** SUBROUTINE NSF
!*** SYMBOLIC LDU-FACTORIZATION OF A NONSYMMETRIC SPARSE MATRIX
!      (UNCOMPRESSED POINTER STORAGE)


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine nsf8 here]
subroutine  nsf8 &
      (n, r,ic, ia,ja, il,jl,jlmax, iu,ju,jumax, q, im, flag)
   
   !       INPUT VARIABLES:   N, R,IC, IA,JA, JLMAX, JUMAX.
   !       OUTPUT VARIABLES:  IL,JL, IU,JU, FLAG.
   
   !       PARAMETERS USED INTERNALLY:
   ! NIA     Q     - SUPPOSE M' IS THE RESULT OF REORDERING M;  IF
   !                   PROCESSING OF THE KTH ROW OF M' (HENCE THE KTH ROWS
   !                   OF L AND U) IS BEING DONE, THEN Q(J) IS INITIALLY
   !                   NONZERO IF M'(K,J) IS NONZERO;  SINCE VALUES NEED
   !                   NOT BE STORED, EACH ENTRY POINTS TO THE NEXT
   !                   NONZERO;  FOR EXAMPLE, IF  N=9  AND THE 5TH ROW OF
   !                   M' IS
   !                           0 X X 0 X 0 0 X 0,
   !                   THEN Q WILL INITIALLY BE
   !                           A 3 5 A 8 A A 10 A 2        (A - ARBITRARY);
   !                   Q(N+1) POINTS TO THE FIRST NONZERO IN THE ROW AND
   !                   THE LAST NONZERO POINTS TO  N+1;  AS THE ALGORITHM
   !                   PROCEEDS, OTHER ELEMENTS OF Q ARE INSERTED IN THE
   !                   LIST BECAUSE OF FILLIN.
   !                   SIZE = N+1.
   ! NIA     IM    - AT EACH STEP IN THE FACTORIZATION, IM(I) IS THE LAST
   !                   ELEMENT IN THE ITH ROW OF U WHICH NEEDS TO BE
   !                   CONSIDERED IN COMPUTING FILLIN.
   !                   SIZE = N.
   
   !  INTERNAL VARIABLES--
   !    JLPTR - POINTS TO THE LAST POSITION USED IN  JL.
   !    JUPTR - POINTS TO THE LAST POSITION USED IN  JU.
   
   integer  r(1), ic(1),  ia(1), ja(1),  il(1), jl(1), &
         iu(1), ju(1),  q(1),  im(1),  flag,  qm, vj
   
   !  ******  INITIALIZE POINTERS  ****************************************
   jlptr = 0
   il(1) = 1
   juptr = 0
   iu(1) = 1
   
   !  ******  FOR EACH ROW OF L AND U  ************************************
   do k=1,n
      !  ******  SET Q TO THE REORDERED ROW OF A  ****************************
      q(n+1) = n+1
      jmin = ia(r(k))
      jmax = ia(r(k)+1) - 1
      if (jmin > jmax)  goto 101
      do j=jmin,jmax
         vj = ic(ja(j))
         qm = n+1
         1 m = qm
         qm = q(m)
         if (qm < vj)  goto 1
         if (qm == vj)  goto 102
         q(m) = vj
         q(vj) = qm
      end do
      
      !  ******  FOR EACH ENTRY IN THE LOWER TRIANGLE  ***********************
      i = n+1
      3 i = q(i)
      if (i >= k)  goto 7
      !  ******  L(K,I) WILL BE NONZERO, SO ADD IT TO JL  ********************
      jlptr = jlptr+1
      if (jlptr > jlmax)  goto 103
      jl(jlptr) = i
      qm = i
      !  ******  INSPECT ITH ROW FOR FILLIN, ADJUST IM IF POSSIBLE  **********
      jmin = iu(i)
      jmax = im(i)
      if (jmin > jmax)  goto 6
      do j=jmin,jmax
         vj = ju(j)
         if (vj == k)  im(i) = j
         4 m = qm
         qm = q(m)
         if (qm < vj)  goto 4
         if (qm == vj)  goto 5
         q(m) = vj
         q(vj) = qm
         qm = vj
      5 continue
      end do
      6 goto 3
      
      !  ******  CHECK FOR NULL PIVOT  ***************************************
      7 if (i /= k)  goto 105
      !  ******  REMAINING ELEMENTS OF Q DEFINE STRUCTURE OF U(K, )  *********
      8 i = q(i)
      if (i > n)  goto 9
      juptr = juptr+1
      if (juptr > jumax)  goto 106
      ju(juptr) = i
      goto 8
      !  ******  GET READY FOR NEXT ROW  *************************************
      9 im(k) = juptr
      il(k+1) = jlptr+1
      iu(k+1) = juptr+1
   end do  ! k=1,n
   
   flag = 0
   return
   
   ! ** ERROR:  NULL ROW IN A
   101 flag = n + r(k)
   return
   ! ** ERROR:  DUPLICATE ENTRY IN A
   102 flag = 2*n + r(k)
   return
   ! ** ERROR:  INSUFFICIENT STORAGE FOR JL
   103 flag = 3*n + k
   return
   ! ** ERROR:  NULL PIVOT
   105 flag = 5*n + k
   return
   ! ** ERROR:  INSUFFICIENT STORAGE FOR JU
   106 flag = 6*n + k
   return
end subroutine nsf8 

!       ----------------------------------------------------------------

!*** SUBROUTINE NNF
!*** NUMERIC LDU-FACTORIZATION OF SPARSE NONSYMMETRIC MATRIX AND
!      SOLUTION OF SYSTEM OF LINEAR EQUATIONS (UNCOMPRESSED POINTER
!      STORAGE)


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine nnf8 here]
subroutine  nnf8 &
      (n, r,c,ic, ia,ja,a, z, b, il,jl,l,lmax, d, iu,ju,u,umax, &
      row, tmp, flag)
   
   !       INPUT VARIABLES:   N, R,C,IC, IA,JA,A, B, IL,JL,LMAX, IU,JU,UMAX
   !       OUTPUT VARIABLES:  Z, L,D,U, FLAG
   
   !       PARAMETERS USED INTERNALLY:
   ! FIA     ROW   - HOLDS INTERMEDIATE VALUES IN CALCULATION OF L, D, U.
   !                   SIZE = N.
   ! FIA     TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
   !                   EQUATION  UX = B'.
   !                   SIZE = N.
   
   integer  r(1), c(1), ic(1),  ia(1), ja(1), &
         il(1), jl(1), lmax,  iu(1), ju(1), umax,  flag
   !       REAL  A(1), Z(1), B(1),  L(1), D(1), U(1),
   !    *     ROW(1), TMP(1),  LI, SUM, DK
   _REAL_  a(1), z(1), b(1),  l(1), d(1), u(1), &
         row(1), tmp(1),  li, sum, dk
   
   !  ******  CHECK STORAGE  **********************************************
   if (il(n+1)-1 > lmax)  goto 104
   if (iu(n+1)-1 > umax)  goto 107
   
   !  ******  FOR EACH ROW  ***********************************************
   do k=1,n
      !  ******  SET THE INITIAL STRUCTURE OF ROW  ***************************
      jmin = il(k)
      jmax = il(k+1) - 1
      if (jmin > jmax)  goto 2
      !  ******  IF L(K,M) .NE. 0, ROW(M)=0  *********************************
      do j=jmin,jmax
         row(jl(j)) = 0
      end do
      2 row(k) = 0
      jmin = iu(k)
      jmax = iu(k+1) - 1
      if (jmin > jmax)  goto 4
      !  ******  IF U(K,M) .NE. 0, ROW(M)=0  *********************************
      do j=jmin,jmax
         row(ju(j)) = 0
      end do
      4 jmin = ia(r(k))
      jmax = ia(r(k)+1) - 1
      !  ******  SET ROW TO KTH ROW OF REORDERED A  **************************
      do j=jmin,jmax
         row(ic(ja(j))) = a(j)
      end do
      !  ******  INITIALIZE SUM  *********************************************
      sum = b(r(k))
      
      !  ******  ASSIGN THE KTH ROW OF L AND ADJUST ROW, SUM  ****************
      imin = il(k)
      imax = il(k+1) - 1
      if (imin > imax)  goto 8
      do i=imin,imax
         li = - row(jl(i))
         !  ******  IF L IS NOT REQUIRED, THEN COMMENT OUT THE FOLLOWING LINE  **
         l(i) = - li
         sum = sum + li * tmp(jl(i))
         jmin = iu(jl(i))
         jmax = iu(jl(i)+1) - 1
         if (jmin > jmax)  goto 7
         do j=jmin,jmax
            row(ju(j)) = row(ju(j)) + li * u(j)
         end do
      7 continue
      end do
      
      !  ******  ASSIGN DIAGONAL D AND KTH ROW OF U, SET TMP(K)  *************
      8 if (row(k) == 0)  goto 108
      dk = 1 / row(k)
      d(k) = dk
      tmp(k) = sum * dk
      jmin = iu(k)
      jmax = iu(k+1) - 1
      if (jmin > jmax)  goto 10
      do j=jmin,jmax
         u(j) = row(ju(j)) * dk
      end do
   10 continue
   end do  ! k=1,n
   
   !  ******  SOLVE  UX = TMP  BY BACK SUBSTITUTION  **********************
   k = n
   do i=1,n
      sum = tmp(k)
      jmin = iu(k)
      jmax = iu(k+1) - 1
      if (jmin > jmax)  goto 12
      do j=jmin,jmax
      end do
      11 sum = sum - u(j) * tmp(ju(j))
      12 tmp(k) = sum
      z(c(k)) = sum
      k = k-1
   end do
   
   flag = 0
   return
   
   ! ** ERROR:  INSUFFICIENT STORAGE FOR L
   104 flag = 4*n + 1
   return
   ! ** ERROR:  INSUFFICIENT STORAGE FOR U
   107 flag = 7*n + 1
   return
   ! ** ERROR:  ZERO PIVOT
   108 flag = 8*n + k
   return
end subroutine nnf8 

!       ----------------------------------------------------------------

!*** SUBROUTINE NNS
!*** NUMERIC SOLUTION OF A SPARSE NONSYMMETRIC SYSTEM OF LINEAR
!      EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER STORAGE)


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine nns8 here]
subroutine  nns8 &
      (n, r,c, il,jl,l, d, iu,ju,u, z, b, tmp)
   
   !       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
   !       OUTPUT VARIABLES:  Z
   
   !       PARAMETERS USED INTERNALLY:
   ! FIA     TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
   !                   EQUATION UX = B'.
   !                   SIZE = N.
   
   integer  r(1), c(1),  il(1), jl(1),  iu(1), ju(1)
   !       REAL  L(1), D(1), U(1),  Z(1), B(1),  TMP(1), SUM
   _REAL_  l(1), d(1), u(1),  z(1), b(1),  tmp(1), sum
   
   !  ******  SOLVE LDY = B  BY FORWARD SUBSTITUTION  *********************
   do k=1,n
      sum = b(r(k))
      jmin = il(k)
      jmax = il(k+1) - 1
      if (jmin > jmax)  goto 2
      do j=jmin,jmax
         sum = sum - l(j) * tmp(jl(j))
      end do
      2 tmp(k) = sum * d(k)
   end do
   
   !  ******  SOLVE  UX = Y  BY BACK SUBSTITUTION  ************************
   k = n
   do i=1,n
      sum = tmp(k)
      jmin = iu(k)
      jmax = iu(k+1) - 1
      if (jmin > jmax)  goto 4
      do j=jmin,jmax
         sum = sum - u(j) * tmp(ju(j))
      end do
      4 tmp(k) = sum
      z(c(k)) = sum
      k = k-1
   end do
   return
end subroutine nns8 
!       ----------------------------------------------------------------

!*** SUBROUTINE NNT
!*** NUMERIC SOLUTION OF THE TRANSPOSE OF A SPARSE NONSYMMETRIC SYSTEM
!      OF LINEAR EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER
!      STORAGE)


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine nnt8 here]
subroutine  nnt8 &
      (n, r,c, il,jl,l, d, iu,ju,u, z, b, tmp)
   
   !       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
   !       OUTPUT VARIABLES:  Z
   
   !       PARAMETERS USED INTERNALLY:
   ! FIA     TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
   !                   EQUATION LX = B'.
   !                   SIZE = N.
   
   integer  r(1), c(1),  il(1), jl(1),  iu(1), ju(1)
   !       REAL  L(1), D(1), U(1),  Z(1), B(1),  TMP(1), TMPK
   _REAL_  l(1), d(1), u(1),  z(1), b(1),  tmp(1), tmpk
   
   !  ******  SOLVE  UT Y = B  BY FORWARD SUBSTITUTION  *******************
   do k=1,n
      tmp(k) = b(c(k))
   end do
   do k=1,n
      tmpk = - tmp(k)
      jmin = iu(k)
      jmax = iu(k+1) - 1
      if (jmin > jmax)  goto 3
      do j=jmin,jmax
      tmp(ju(j)) = tmp(ju(j)) + u(j) * tmpk
      end do
   3 continue
   end do
   
   !  ******  SOLVE  D LT X = Y  BY BACK SUBSTITUTION  ********************
   k = n
   do i=1,n
      tmpk = - tmp(k) * d(k)
      jmin = il(k)
      jmax = il(k+1) - 1
      if (jmin > jmax)  goto 5
      do j=jmin,jmax
         tmp(jl(j)) = tmp(jl(j)) + l(j) * tmpk
      end do
      5 z(r(k)) = - tmpk
      6 k = k-1
   end do
   return
end subroutine nnt8 
