#define _REAL_ double precision
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!     AUX1R5

!     INTERFACE ROUTINE FOR SIMPLIFIED CALL OF AMG1R5

!     RELEASE 1.4, SEPTEMBER 1986

!     CHANGES AGAINST VERSION 1.1, JULY 1985:

! 1.  A BUG WAS DETECTED WHICH UNDER CERTAIN CIRCUMSTANCES INFLUENCED
!     SLIGHTLY THE CONVERGENCE RATE OF AMG1R1. FOR THAT REASON, THE
!     FOLLOWING LINE IN SUBROUTINE RESC:
!     IW(IMAXW(KC-1)+1) = IA(IMIN(KC))
!     HAS BEEN CHANGED TO:
!     IW(IMAXW(KC-1)+1) = IAUX

! 2.  A BUG WAS DETECTED IN SUBROUTINE PWINT. UNDER CERTAIN CIRCUM-
!     STANCES AN UNDEFINED VARIABLE WAS USED. ALTHOUGH THIS DID NOT
!     AFFECT THE NUMERICAL RESULTS, PROBLEMS CAN OCCUR IF CHECKING
!     FOR UNDEFINED VARIABLES IS USED. TO FIX THIS ERROR, IN PWINT
!     THE LABEL 1000 WAS MOVED TO THE STATEMENT
!     IBLCK1 = IMINW(K).

! 3.  A PARAMETER LRATIO HAS BEEN INTRODUCED, DENOTING THE RATIO
!     OF SPACE OCCUPIED BY A DOUBLE PRECISION REAL VARIABLE AND
!     THAT OF AN INTEGER. FOR THE IBM-VERSION LRATIO HAS BEEN SET
!     TO 2. CHANGE THIS VALUE IF NECESSARY. (IF, FOR EXAMPLE, YOU
!     WANT TO CHANGE THE DOUBLE PRECISION VECTORS TO SINGLE PRE-
!     CISION, LRATIO HAS TO BE SET TO 1. IN THE YALE SMP - ROUTINE
!     NDRV THERE IS A PARAMETER LRATIO, TOO.

! 4.  TYPE DECLARATIONS REAL*4 AND REAL*8 HAVE BEEN CHANGED TO THE
!     STANDARD-CONFORMING KEYWORDS REAL AND DOUBLE PRECISION, RESPEC-
!     TIVELY.

! 5.  CALLS TO THE FOLLOWING INTRINSIC FUNCTIONS HAVE BEEN REPLACED BY
!     CALLS USING GENERIC NAMES: DSQRT, MIN0, MAX0, IABS, DABS, FLOAT,
!     DFLOAT, DMAX1, ISIGN, IDINT, DLOG10.

! 6.  A SAVE STATEMENT HAS BEEN INSERTED IN ALL SUBROUTINES.

! 7.  EXTERNAL DECLARATION STATEMENTS HAVE BEEN INSERTED IN ALL SUB-
!     ROUTINES FOR ALL EXTERNAL REFERENCES.

!-----------------------------------------------------------------------

!     CHANGE AGAINST VERSION 1.3, APRIL 1986:

! 1.  A BUG IN SUBROUTINE CHECK HAS BEEN REMOVED. IF THE ORIGINAL MATRIX
!     WAS STORED IN AN UNSYMMETRIC WAY, THE SYMMETRIZATION BY AMG1R3
!     COULD FAIL UNDER CERTAIN CIRCUMSTANCES. FOR A FIX, THE FOLLOWING
!     STATEMENTS IN SUBROUTINE CHECK HAVE BEEN CHANGED:

!     DO 450 J=IA(I)+1,IA(I+1)-1 WAS CHANGED TO
!     DO 450 J=IA(I)+1,ICG(I)-1

!     DO 430 J1=IA(I1)+1,IA(I1+1)-1 WAS CHANGED TO
!     DO 430 J1=IA(I1)+1,ICG(I1)-1

!     DO 550 J=IA(I)+1,IA(I+1)-1 WAS CHANGED TO
!     DO 550 J=IA(I)+1,ICG(I)-1

!     DO 530 J1=IA(I1)+1,IA(I1+1)-1 WAS CHANGED TO
!     DO 530 J1=IA(I1)+1,ICG(I1)-1

! 2.  THE EXPLANATORY PART IN SUBROUTINE AMG1R5 HAS BEEN ENLARGED TO
!     AVOID MISUNDERSTANDINGS IN THE DEFINITION OF THE ARGUMENT LIST.


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+ [Enter a one-line description of subroutine aux1r5 here]
subroutine aux1r5(a,ia,ja,u,f,ig, &
      nda,ndia,ndja,ndu,ndf,ndig,nnu,matrix, &
      eps,ifirst,iswtch,iout,iprint, &
      ierr)
   
   !      ------------------------------------------------------------
   !      | INTERFACE TO AMG-MODULE FOR SOLVING LINEAR SYSTEMS L*U=F |
   !      ------------------------------------------------------------
   
   !         --------------------------------------------------------------
   
   !     ASSUMPTIONS ON L:
   
   !         THE PROGRAM REQUIRES:
   
   !             - DIAGONAL ENTRIES ARE ALWAYS POSITIVE (ON ALL GRIDS);
   !             - L IS A SQUARE MATRIX WHICH IS EITHER REGULAR OR SINGULAR
   !               WITH ROWSUMS=0.
   
   !         FOR THEORETICAL REASONS THE FOLLOWING SHOULD HOLD:
   
   !             - L POSITIVE DEFINITE (OR SEMI-DEFINITE WITH ROWSUM=0)
   !             - L "ESSENTIALLY" POSITIVE TYPE, I.E.,
   
   !                  -- DIAGONAL ENTRIES MUST BE > 0 ;
   !                  -- MOST OF THE OFF-DIAGONAL ENTRIES <= 0 ;
   !                  -- ROWSUMS SHOULD BE >= 0 .
   
   
   !     THE USER HAS TO PROVIDE THE MATRIX L, THE RIGHT HAND SIDE F AND
   !     CERTAIN POINTER VECTORS IA AND JA.
   
   !         --------------------------------------------------------------
   
   !     STORAGE OF L:
   
   !         THE NON-ZERO ENTRIES OF THE MATRIX L ARE STORED IN
   !         "COMPRESSED" SKY-LINE FASHION IN A 1-D VECTOR A, I.E., ROW
   !         AFTER ROW, EACH ROW STARTING WITH ITS DIAGONAL ELEMENT. THE
   !         OTHER NON-ZERO ROW ENTRIES FOLLOW THEIR DIAGONAL ENTRY IN ANY
   !         ORDER.
   
   !         IN ORDER TO IDENTIFY EACH ELEMENT IN A, THE USER HAS TO
   !         PROVIDE TWO POINTER ARRAYS IA AND JA. IF NNU DENOTES THE TOTAL
   !         NUMBER OF UNKNOWNS, THE NON-ZERO ENTRIES OF ANY ROW I OF L
   !         (1.LE.I.LE.NNU) ARE STORED IN A(J) WHERE THE RANGE OF J
   !         IS GIVEN BY
   
   !                     IA(I) .LE. J .LE. IA(I+1)-1.
   
   !         THUS, IA(I) POINTS TO THE POSITION OF THE DIAGONAL ENTRY OF
   !         ROW I WITHIN THE VECTOR A. IN PARTICULAR,
   
   !                     IA(1) = 1 ,  IA(NNU+1) = 1 + NNA
   
   !         WHERE NNA DENOTES THE TOTAL NUMBER OF MATRIX ENTRIES STORED.
   !         THE POINTER VECTOR JA HAS TO BE DEFINED SUCH THAT
   !         ANY ENTRY A(J) CORRESPONDS TO THE UNKNOWN U(JA(J)), I.E.,
   !         JA(J) POINTS TO THE COLUMN INDEX OF A(J).
   !         IN PARTICULAR, A(IA(I)) IS THE DIAGONAL ENTRY OF ROW I
   !         AND CORRESPONDS TO THE UNKNOWN U(I): JA(IA(I))=I.
   
   !         IN THIS TERMINOLOGY, THE I-TH EQUATION READS AS FOLLOWS
   !         (FOR ANY I WITH  1.LE.I.LE.NNU):
   
   !                  F(I) =        SUM      A(J) * U(JA(J))
   !                           J1.LE.J.LE.J2
   
   !         WHERE F(I) DENOTES THE I-TH COMPONENT OF THE RIGHT HAND
   !         SIDE AND
   
   !                     J1 = IA(I) ,  J2 = IA(I+1)-1.
   
   !         NOTES: THE ENTRY IA(NNU+1) HAS TO POINT TO THE FIRST FREE
   !                ENTRY IN VECTORS A AND JA, RESPECTIVELY. OTHERWISE,
   !                AMG CANNOT KNOW THE LENGTH OF THE LAST MATRIX ROW.
   
   !                THE INPUT VECTORS A, IA AND JA ARE CHANGED BY AMG1R5.
   !                SO, AFTER RETURN FROM AMG1R5, THE PACKAGE MUST NOT
   !                BE CALLED A SECOND TIME WITHOUT HAVING NEWLY DEFINED
   !                THE INPUT VECTORS AND USING ISWTCH=4. OTHERWISE, THE
   !                SETUP PHASE WILL FAIL.
   !                  ON THE OTHER HAND, RUNNING AMG A SECOND TIME ON THE
   !                SAME INPUT DATA WITH ISWTCH=4 HAS NO SENSE, BECAUSE
   !                THE RESULTS OF THE FIRST SETUP PHASE ARE STILL STORED
   !                AND THUS THIS PHASE CAN BE SKIPPED IN A SECOND CALL.
   !                IN ORDER TO DO THIS, SET ISWTCH TO 1, 2 OR 3.
   
   
   !-----------------------------------------------------------------------
   
   !         THE FORM OF THE CALLING PROGRAM HAS TO BE AS FOLLOWS:
   
   !               PROGRAM DRIVER
   !         C
   !               REAL*8 A(#NDA),U(#NDU),F(#NDF)
   !               INTEGER IA(#NDIA),JA(#NDJA),IG(#NDIG)
   !         C
   !               NDA  = #NDA
   !               NDU  = #NDU
   !               NDF  = #NDF
   !               NDIA = #NDIA
   !               NDJA = #NDJA
   !               NDIG = #NDIG
   !         C
   !         C     SET UP A, F, IA, JA AND SPECIFY NECESSARY PARAMETERS
   !         C
   !               ....
   !               ....
   !         C
   !               CALL AUX1R5(A,IA,JA,U,F,IG,
   !        +                  NDA,NDIA,NDJA,NDU,NDF,NDIG,NNU,MATRIX,
   !        +                  EPS,IFIRST,ISWTCH,IOUT,IPRINT,
   !        +                  IERR)
   !         C
   !               ....
   !               ....
   !         C
   !               STOP
   !               END
   
   !-----------------------------------------------------------------------
   
   !     INPUT VIA ARRAYS (SEE ABOVE):
   
   !     A        -   MATRIX L
   
   !     IA       -   POINTER VECTOR
   
   !     JA       -   POINTER VECTOR
   
   !     U        -   FIRST APPROXIMATION TO SOLUTION
   
   !     F        -   RIGHT HAND SIDE
   
   
   !-----------------------------------------------------------------------
   
   
   !     SCALAR INPUT PARAMETERS OF AUX1R5:
   
   !     THE INPUT PARAMETERS OF AUX1R5 IN THE LIST BELOW ARE ARRANGED
   !     ACCORDING TO THEIR IMPORTANCE TO THE GENERAL USER. THE PARAMETERS
   !     PRECEEDED BY A * MUST BE SPECIFIED EXPLICITELY (DEFINITION OF THE
   !     USER-DEFINED PROBLEM AND DIMENSIONING OF VECTORS IN THE CALLING
   !     PROGRAM). THE OTHER PARAMETERS ARE SET TO STANDARD VALUES IF ZERO
   !     ON INPUT.
   
   
   !  *  NDA      -   DIMENSIONING OF VECTOR A IN CALLING PROGRAM
   
   !  *  NDIA     -   DIMENSIONING OF VECTOR IA IN CALLING PROGRAM
   
   !  *  NDJA     -   DIMENSIONING OF VECTOR JA IN CALLING PROGRAM
   
   !  *  NDU      -   DIMENSIONING OF VECTOR U IN CALLING PROGRAM
   
   !  *  NDF      -   DIMENSIONING OF VECTOR F IN CALLING PROGRAM
   
   !  *  NDIG     -   DIMENSIONING OF VECTOR IG IN CALLING PROGRAM
   
   !  *  NNU      -   NUMBER OF UNKNOWNS
   
   !  *  MATRIX   -   INTEGER VALUE CONTAINING INFO ABOUT THE MATRIX L.
   
   !                  1ST DIGIT OF MATRIX  --  ISYM:
   !                    =1: L IS SYMMETRIC;
   !                    =2: L IS NOT SYMMETRIC.
   
   !                  2ND DIGIT OF MATRIX  --  IROW0:
   !                    =1: L HAS ROWSUM ZERO;
   !                    =2: L DOES NOT HAVE ROWSUM ZERO.
   
   
   !  *  EPS      -   CONVERGENCE CRITERION FOR SOLUTION PROCESS. STOP, IF
   !                  L2-NORM OF THE RESIDUAL OF THE USER-DEFINED PROBLEM
   !                  IS LESS THAN EPS.
   
   !     IFIRST   -   PARAMETER FOR FIRST APPROXIMATION.
   
   !                  1ST DIGIT OF IFIRST: NOT USED; HAS TO BE NON-ZERO.
   
   !                  2ND DIGIT OF IFIRST  --  ITYPU:
   !                    =0: NO SETTING OF FIRST APPROXIMATION,
   !                    =1: FIRST APPROXIMATION CONSTANT TO ZERO,
   !                    =2: FIRST APPROXIMATION CONSTANT TO ONE,
   !                    =3: FIRST APPROXIMATION IS RANDOM FUNCTION WITH
   !                        THE CONCRETE RANDOM SEQUENCE BEING DETERMINED
   !                        BY THE FOLLWING DIGITS.
   
   !                  REST OF IFIRST  --  RNDU:
   !                    DETERMINES THE CONCRETE RANDOM SEQUENCE USED IN
   !                    THE CASE ITYPU=3. (IFIRST=13 IS EQUIVALENT TO
   !                    IFIRST=1372815)
   
   !     ISWTCH   -   PARAMETER CONTROLLING WHICH MODULES OF AMG1R5 ARE TO
   !                  BE USED.
   !                    =1:   CALL FOR -----, -----, -----, WRKCNT.
   !                    =2:   CALL FOR -----, -----, SOLVE, WRKCNT.
   !                    =3:   CALL FOR -----, FIRST, SOLVE, WRKCNT.
   !                    =4:   CALL FOR SETUP, FIRST, SOLVE, WRKCNT.
   !                  SETUP DEFINES THE OPERATORS NEEDED IN THE SOLUTION
   !                         PHASE.
   !                  FIRST INITIALIZES THE SOLUTION VECTOR (SEE PARAMETER
   !                         IFIRST).
   !                  SOLVE COMPUTES THE SOLUTION BY AMG CYCLING (SEE
   !                         AMG1R5-PARAMETER NCYC)
   !                  WRKCNT PROVIDES THE USER WITH INFORMATION ABOUT
   !                         RESIDUALS, STORAGE REQUIREMENTS AND CP-TIMES
   !                         (SEE PARAMETER IOUT).
   !                  IF AUX1R5 IS CALLED THE FIRST TIME, ISWTCH HAS TO
   !                  BE =4. INDEPENDENT OF ISWTCH, SINGLE MODULES CAN BE
   !                  BYPASSED BY A PROPER CHOICE OF THE CORRESPONDING
   !                  PARAMETER.
   
   !     IOUT     -   PARAMETER CONTROLLING THE AMOUNT OF OUTPUT DURING
   !                  SOLUTION PHASE:
   
   !                  1ST DIGIT: NOT USED; HAS TO BE NON-ZERO.
   
   !                  2ND DIGIT:
   !                    =0: NO OUTPUT (EXCEPT FOR MESSAGES)
   !                    =1: RESIDUAL BEFORE AND AFTER SOLUTION PROCESS
   !                    =2: ADD.: STATISTICS ON CP-TIMES AND STORAGE REQUI-
   !                        REMENTS
   !                    =3: ADD.: RESIDUAL AFTER EACH AMG-CYCLE
   
   !     IPRINT   -   PARAMETER SPECIFYING THE FORTRAN UNIT NUMBERS FOR
   !                  OUTPUT:
   
   !                  1ST DIGIT: NOT USED; HAS TO BE NON-ZERO
   
   !                  2ND AND 3RD DIGIT  --  IUP: UNIT NUMBER FOR RESULTS
   
   !                  4TH AND 5TH DIGIT  --  IUM: UNIT NUMBER FOR MESSAGES
   
   
   !-----------------------------------------------------------------------
   
   !     OUTPUT:
   
   !     U        -   CONTAINS THE COMPUTED SOLUTION
   
   
   !     IERR     -   ERROR PARAMETER:
   
   !                    >0: FATAL ERROR (ABNORMAL TERMINATION)
   !                    <0: NON-FATAL ERROR (EXECUTION CONTINUES)
   
   !                  ERROR CODES IN DETAIL:
   
   !                  1. DIMENSIONING TOO SMALL FOR VECTOR
   !                        A      (IERR = 1)
   !                        IA     (IERR = 2)
   !                        JA     (IERR = 3)
   !                        U      (IERR = 4)
   !                        F      (IERR = 5)
   !                        IG     (IERR = 6)
   
   !                     NO YALE-SMP BECAUSE OF STORAGE (NDA TOO SMALL):
   !                               (IERR = -1)
   !                     NO YALE-SMP BECAUSE OF STORAGE (NDJA TOO SMALL):
   !                               (IERR = -3)
   !                     NO CG BECAUSE OF STORAGE (NDU TOO SMALL):
   !                               (IERR = -4)
   !                     NO SPACE FOR TRANSPOSE OF INTERPOLATION (NDA OR
   !                                                     NDJA TOO SMALL):
   !                               (IERR = -1)
   
   !                  2. INPUT DATA ERRONEOUS:
   
   !                     A-ENTRY MISSING, ISYM = 1:           (IERR = -11)
   !                     PARAMETER MATRIX MAY BE ERRONEOUS:   (IERR = -12)
   !                     DIAGONAL ELEMENT NOT STORED FIRST:   (IERR =  13)
   !                     DIAGONAL ELEMENT NOT POSITIV:        (IERR =  14)
   !                     POINTER IA ERRONEOUS:                (IERR =  15)
   !                     POINTER JA ERRONEOUS:                (IERR =  16)
   !                     PARAMETER ISWTCH ERRONEOUS:          (IERR =  17)
   
   !                  3. ERRORS OF THE AMG1R5-SYSTEM (SHOULD NOT OCCUR):
   
   !                     TRANSPOSE A-ENTRY MISSING:           (IERR =  21)
   !                     INTERPOLATION ENTRY MISSING:         (IERR =  22)
   
   !                  4. ALGORITHMIC ERRORS:
   
   !                     CG-CORRECTION NOT DEFINED:           (IERR =  31)
   !                     NO YALE-SMP BECAUSE OF ERROR IN
   !                     FACTORIZATION:                       (IERR = -32)
   
   !-----------------------------------------------------------------------
   
   !     WORK SPACE:
   
   !     THE INTEGER VECTOR IG HAS TO BE PASSED TO AUX1R5 AS WORK SPACE.
   
   !-----------------------------------------------------------------------
   
   !     DIMENSIONING OF INPUT VECTORS AND WORK SPACE:
   
   !     IT'S IMPOSSIBLE TO TELL IN ADVANCE THE EXACT STORAGE REQUIREMENTS
   !     OF AMG. THE FOLLOWING FORMULAS THUS GIVE ONLY REASONABLE GUESSES
   !     FOR THE REQUIRED VECTOR LENGTHS AS DECLARED IN THE CALLING PRO-
   !     GRAM. IN THESE FORMULAS NNA DENOTES THE NUMBER OF NON-ZERO ENTRIES
   !     IN THE INPUT-MATRIX L AND NNU IS THE NUMBER OF UNKNOWNS.
   
   !     VECTOR         NEEDED LENGTH (GUESS)
   !       A               3*NNA + 5*NNU
   !       JA              3*NNA + 5*NNU
   !       IA              2.2*NNU
   !       U               2.2*NNU
   !       F               2.2*NNU
   !       IG              5.4*NNU
   
   !-----------------------------------------------------------------------
   
   
   !     STANDARD CHOICES OF PARAMETERS:
   
   !          IFIRST = 13
   !          ISWTCH = 4
   !          IOUT   = 12
   !          IPRINT = 10606
   
   !     IF ANY ONE OF THESE PARAMETERS IS 0 ON INPUT, ITS CORRESPONDING
   !     STANDARD VALUE IS USED BY AUX1R5.
   
   !-----------------------------------------------------------------------
   
   !     PORTABILITY RESTRICTIONS:
   
   !     1. ROUTINE CTIME IS MACHINE DEPENDENT AND HAS TO BE ADAPTED TO
   !        YOUR COMPUTER INSTALLATION OR REPLACED BY A DUMMY ROUTINE.
   
   !     2. THE AMG1R5 SYSTEM USES INTEGER PARAMETERS OF UP TO FIVE DIGITS.
   !        BE SURE THAT YOUR COMPUTER CAN STORE FIVE DIGITS ON AN INTEGER
   !        VARIABLE.
   
   !     3. APART FROM FORTRAN INTRINSIC FUNCTIONS AND SERVICE ROUTINES,
   !        THERE IS ONLY ONE EXTERNAL REFERENCE TO A PROGRAM NOT CONTAINED
   !        IN THE AMG1R5 SYSTEM, I.E. THE LINEAR SYSTEM SOLVER NDRV OF
   !        THE YALE SPARSE MATRIX PACKAGE. IF YOU HAVN'T ACCESS TO THIS
   !        PACKAGE, ENTER A DUMMY ROUTINE NDRV AND REPLACE THE LINE
   !             NSOLCO = 2
   !        BY
   !             NSOLCO = 1
   !        IN AUX1R5. THEN NDRV ISN'T CALLED BY AUX1R5. IN THIS CASE,
   !        HOWEVER, INDEFINITE PROBLEMS WILL NOT BE SOLVABLE.
   !          THE YALE SPARSE MATRIX PACKAGE IS FREELY AVAILABLE FOR NON-
   !        PROFIT PURPOSES. CONTACT THE DEPARTMENT OF COMPUTER SCIENCE,
   !        YALE UNITVERSITY.
   
   !-----------------------------------------------------------------------
   
   !     AUTHORS:
   
   !          JOHN RUGE, FORT COLLINS (USA),
   !              INSTITUTE FOR COMPUTATIONAL STUDIES AT CSU;
   
   !          KLAUS STUEBEN, D-5205 ST. AUGUSTIN (W.-GERMANY),
   !              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG (GMD).
   
   !          ROLF HEMPEL, D-5205 ST. AUGUSTIN (W.-GERMANY),
   !              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG (GMD).
   
   !-----------------------------------------------------------------------
   
   implicit _REAL_ (a-h,o-z)
   _REAL_ a(*),u(*),f(*)
   integer ia(*),ja(*),ig(*)
   
   !===> FOR ZERO-PARAMETERS STANDARD VALUES ARE USED BY AMG1R5
   
   levelx = 0
   ncyc   = 1025000
   madapt = 0
   nrd    = 0
   nsolco = 1
   nru    = 0
   ecg1   = 0.d0
   ecg2   = 0.25d0
   ewt2   = 0.35d0
   nwt    = 2
   ntr    = 0
   
   call amg1r5(a,ia,ja,u,f,ig, &
         nda,ndia,ndja,ndu,ndf,ndig,nnu,matrix, &
         iswtch,iout,iprint, &
         levelx,ifirst,ncyc,eps,madapt,nrd,nsolco,nru, &
         ecg1,ecg2,ewt2,nwt,ntr, &
         ierr)
end subroutine aux1r5 
