#include "../include/dprec.fh"

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Reads in the header data from a Amber style Xvv file.  Determines the version
!!!and reads in POINTER information that determines the sizes of array in the
!!!rest of the file
!!!IN:
!!!   nf    : Fortran unit number for an open file
!!!   natom : total number of solvent sites
!!!   nr    : number of grid points in Xvv
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rdxvv1(nf,natom,nr)
  use rism3d_solv_c
  implicit none
#include "rism_nxtsec.h"      
  integer, intent(in) :: nf
  integer, intent(out) :: natom, nr
  integer :: i, iok
  character(len=80) fmt
  character(len=80) fmtin,ifmt,afmt,rfmt,type
  
  !nspecies :: local variable to read the number of species
  integer :: nspecies
  
  ifmt = '(10I8)'
  afmt = '(20A4)'
  rfmt = '(5E16.8)'
  initprmtop = .true.
  !     ----- READ THE SYMBOLS AND THE CHARGES AND THE MASSES -----

  fmtin = ifmt
  type = 'POINTERS'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) NR,NATOM,NSPECIES

end subroutine rdxvv1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Reads the data from a Amber style Xvv file into a solvent object.  Does 
!!!necessary unit conversion. Number of sites and grid size should be set in the
!!!solvent object based on previous call to rdxvv1().
!!!IN:
!!!   solv : solvent object
!!!   nf   : Fortran unit number of an open file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rdxvv2(solv,nf)
  use rism3d_solv_c
  use constants, only : COULOMB_CONST_E, KB, BOLTZMANN, AVOGADRO
  implicit none
  type(rism3d_solv), intent(inout) :: solv
  integer, intent(in) :: nf
  character(len=255) aa

  integer :: i, iok
  character(len=80) fmt
  character(len=80) fmtin,ifmt,afmt,rfmt,type
  
  integer itab, itab1, ig,iv,imv, iv1, iv2, iiga

  _REAL_ :: err

  ifmt = '(10I8)'
  afmt = '(20A4)'
  rfmt = '(5E16.8)'
  !     ----- READ THE SYMBOLS AND THE CHARGES AND THE MASSES -----
  
  fmtin = rfmt
  type = 'THERMO'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) solv%temperature,solv%dielconst,solv%xappa,solv%xikt,solv%dr,solv%smear

  fmtin = ifmt
  type = 'MTV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%mult(i),i = 1,solv%natom)

  fmtin = afmt
  type = 'ATOM_NAME'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%atomname(i),i = 1,solv%natom)

  fmtin = rfmt
  type = 'RHOV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%rho(i),i = 1,solv%natom)

  fmtin = rfmt
  type = 'QSPV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%charge_sp(i),i = 1,solv%natom)
  !UNIT CONVERSION 
  !from [e] to [sqrt(kT A)]
  solv%charge_sp = solv%charge_sp&
       *sqrt(COULOMB_CONST_E/(KB *solv%temperature))

  fmtin = rfmt
  type = 'QV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%charge(i),i = 1,solv%natom)
  !UNIT CONVERSION 
  !from [e] to [sqrt(kT A)]
  solv%charge = solv%charge&
       *sqrt(COULOMB_CONST_E/(KB *solv%temperature))

  fmtin = rfmt
  type = 'EPSV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%eps(i),i = 1,solv%natom)
  !UNIT CONVERSION
  !from [J/MOL] to [kT]
!  solv%eps=solv%eps/JPKC
  solv%eps=solv%eps/ (BOLTZMANN * AVOGADRO *solv%temperature) 
  

  fmtin = rfmt
  type = 'SIGV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%sig_s(i),i = 1,solv%natom)

  fmtin = rfmt
  type = 'DELHV0'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (solv%delhv0(i),i = 1,solv%natom)
  !UNIT CONVERSION
  !from [e] to [sqrt(kT A)]
  solv%delhv0 = solv%delhv0/sqrt(COULOMB_CONST_E/KB/solv%temperature)

  fmtin = rfmt
  type = 'XVV'
  call rism_nxtsec(nf,  6,  0,fmtin,  type,  fmt,  iok)
  read(nf,fmt) (((solv%xvv(itab,iv1,iv2),itab = 1,solv%nr),iv1=1,solv%natom),iv2=1,solv%natom)

end subroutine rdxvv2
