#!/usr/bin/env python

from pyleap import addions
from pymort import ionee_i, dbevec, strvec, charge
from chimera import Mol2io, defaultMol2ioHelper, Element
from chimera.molEdit import addAtom
from sys import argv, exit


def is_solvent(r):
    for a in r.atoms:
        for n in a.neighbors:
            if n.residue !=r:
                return False

    return True

def get_solute_nresd(m):

    i=len(m.residues)-1

    while i>=0 and is_solvent(m.residues[i]):
        i -= 1

    return i+1
 
def get_solute_natom(m,solute_nresd=-1):
    if solute_nresd==-1:
        solute_nresd = get_solute_nresd(m)

    sum=0
    for i in xrange(solute_nresd):
        sum += len(m.residues[i].atoms)

    return sum

class chimera_ionee( ionee_i ):
	def __init__(self, m):
		from pyleap import load_amber_params
		load_amber_params( "parm99.dat" )

		from pymort import get_vdwr
		ionee_i.__init__(self)
		self.m_pmol = m
		self.m_natom = len(m.atoms)
		self.m_nresd = len(m.residues)

		self.m_solute_nresd = get_solute_nresd(m)
		self.m_solute_natom = get_solute_natom(m, self.m_solute_nresd)

		self.m_cord = dbevec(self.m_natom*3)
		self.m_chrg = dbevec(self.m_natom)
		self.m_vdwr = dbevec(self.m_natom)
		self.m_type = strvec(self.m_natom)

		iatom = 0
		iresd = 0
		for r in m.residues:
			iresd += 1
			for a in r.atoms:
				crd = a.coord()
				self.m_cord[3*iatom  ] = crd[0]
				self.m_cord[3*iatom+1] = crd[1]
				self.m_cord[3*iatom+2] = crd[2]
				self.m_chrg[iatom] = a.charge
				
				if hasattr( a, "gaffType" ):
					self.m_type[iatom] = a.gaffType
				else:
					self.m_type[iatom] = a.mol2type

				self.m_vdwr[iatom] = get_vdwr(self.m_type[iatom])
				iatom += 1

		if iatom != self.m_natom:
			raise RuntimeError( "Error: there are atoms not belong to any residue" )

		self.m_rids = [None]*self.m_nresd
		for i in xrange(self.m_nresd):
			self.m_rids[i] = i

	def natom(self):
		return self.m_natom

	def nresd(self):
		return self.m_nresd

	def solute_natom(self):
		return self.m_solute_natom

	def solute_nresd(self):
		return self.m_solute_nresd

	def getcord(self):
		return self.m_cord

	def getchrg(self):
		return self.m_chrg

	def getvdwr(self):
		return self.m_vdwr

	def gettype(self):
		return self.m_type

	def insert_resd( self, ion, sft, rid ):
		from chimera import Element, Point, MolResId
		from chimera.molEdit import addAtom

		if( len(ion.atoms) != 1 ):
			raise RuntimeError, "Error: polyatomic ion not supported yet"
		
		resd = self.m_pmol.newResidue(ion.name, MolResId(rid) )
		asrc = ion.atoms[0]

		npos = asrc.position
		npos[0] += sft[0]
		npos[1] += sft[1]
		npos[2] += sft[2]
		adst = addAtom(asrc.name, Element(asrc.element), resd, Point(npos[0], npos[1], npos[2]))
		adst.charge = asrc.pchg
		adst.gaffType = asrc.type

	def remove_resd( self, rid ):
		idx = self.m_rids.index(rid)
		#print "Deleting residue: ", self.m_pmol.residues[idx].atoms[2].coords
		self.m_pmol.deleteResidue( self.m_pmol.residues[idx] )	
		del self.m_rids[idx]


if len(argv)!=4:
	print "addions.py infile ioname otfile"
	exit(-1)


infile = argv[1]
ioname = argv[2]
otfile = argv[3]

mol2io = Mol2io(defaultMol2ioHelper)
mols = mol2io.readMol2file(infile)
if not mol2io.ok():
	print mol2io.error()
	exit(-1)

if not mols:
	print "No molecule in ", infile


m = mols[0]
ionee =  chimera_ionee(m)
addions( ionee, ioname )


from WriteMol2 import writeMol2
writeMol2( [m], otfile )

	
