#!/usr/bin/env python

from pymort import solute_i, dbevec
from pyleap import solvatebox

from chimera import Mol2io, defaultMol2ioHelper, Element
from chimera.molEdit import addAtom
from sys import argv, exit

class chimera_solute( solute_i ):
	def __init__(self, m):
		from pyleap import load_amber_params
		load_amber_params( "parm99.dat" )

		from pymort import get_vdwr
		solute_i.__init__(self)
		self.m_pmol = m
		self.m_natom = len(m.atoms)

		self.m_cord = dbevec(self.m_natom*3)
		self.m_vdwr = dbevec(self.m_natom)

		iatom = 0
		for a in m.atoms:
			crd = a.coord()
			self.m_cord[3*iatom  ] = crd[0]
			self.m_cord[3*iatom+1] = crd[1]
			self.m_cord[3*iatom+2] = crd[2]
				
			if hasattr( a, "gaffType" ):
				type = a.gaffType
			else:
				type = a.mol2type

			self.m_vdwr[iatom] = get_vdwr(type)
			iatom += 1

	def natom(self):
		return self.m_natom

	def getcord(self):
		return self.m_cord

	def getvdwr(self):
		return self.m_vdwr

	def insert_resd( self, svt, sft, rid ):
		from chimera import Element, Point, MolResId
		from chimera.molEdit import addAtom
		from pymort import get_vdwr

		resd = self.m_pmol.newResidue(svt.name, ' ', 1, ' ')

		atmap = {}
		for a in svt.atoms:
			npos = a.position
			npos[0] += sft[0]
			npos[1] += sft[1]
			npos[2] += sft[2]
			adst = addAtom(a.name, Element(a.element), resd, Point(npos[0], npos[1], npos[2]), self.m_natom)
			adst.charge = a.pchg
			adst.gaffType = a.type

			atmap[a.absid]=adst
			self.m_cord.push_back( npos[0] )
			self.m_cord.push_back( npos[1] )
			self.m_cord.push_back( npos[2] )
			self.m_vdwr.push_back( get_vdwr(a.type) )
			self.m_natom += 1
			assert self.m_cord.size()==self.m_natom*3
			assert self.m_vdwr.size()==self.m_natom

		for b in svt.bonds:
			a1 = atmap[ b.atoms[0].absid ]
			a2 = atmap[ b.atoms[1].absid ]
			self.m_pmol.newBond( a1, a2 )

	def finish( self ):
		from chimera import Point
		assert self.m_cord.size() == self.m_natom*3
		assert self.m_vdwr.size() == self.m_natom

		ia = 0
		for a in self.m_pmol.atoms:
			assert ia < self.m_natom
			p = Point( self.m_cord[3*ia], self.m_cord[3*ia+1], self.m_cord[3*ia+2] )
			a.setCoord( p )
			ia += 1

		assert ia==self.m_natom

if len(argv)!=4:
	print "solvatebox.py infile.mol2 solvent otfile.mol2"
	exit(-1)

infile = argv[1]
solvent = argv[2]
otfile = argv[3]

mol2io = Mol2io(defaultMol2ioHelper)
mols = mol2io.readMol2file(infile)
if not mol2io.ok():
	print mol2io.error()
	exit(-1)

if not mols:
	print "No molecule in ", infile


m = mols[0]
solute =  chimera_solute(m)
solvatebox( solute, solvent, 2.0 )


from WriteMol2 import writeMol2
writeMol2( [m], otfile )

